/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESSS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.jini;

import java.io.IOException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.util.LinkedList;
import java.util.List;

import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceRegistrar;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.discovery.DiscoveryEvent;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.DiscoveryListener;
import net.jini.discovery.LookupDiscovery;
import net.jini.lookup.entry.Name;

/**
 * The LookupManager class can be used to interrogate a 
 * lookup service in order to obtain a proxy for a service.
 * LookupManager is only a convenience class, which 
 * invokes the discovery protocol in order to obtain a
 * proxy for the lookup service and then interrogate it
 * opportunely. The Jgroup programmer can use different
 * techniques to perform the lookup, as illustrated in
 * the Jini specification.
 * <p>
 * In order to search for a group which has joined the
 * Jini federation, a GroupEntry entry containing the
 * group name must be included in the template.
 *
 * @author  Alberto Montresor
 * @author  Rohnny Moland
 * @since   Jgroup 1.1
 */
public class LookupManager
  implements DiscoveryListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Singleton manager */
  private static LookupManager manager; 
  

  /**
   * Perform a lookup in the lookup service, searching for the
   * specified object with the specified attributes.
   * @param template template of the service to be searched
   * @return a proxy for the service, or null is no service can be
   *  found with the specified template
   * @throws IOException an error occurred in doing lookup
   */
  public static Object lookup(ServiceTemplate template)
    throws IOException
  {
    return lookup(template, 5000);
  }

  /**
   * Perform a lookup in the lookup service, searching for the
   * specified object with the specified attributes.
   * @param template template of the service to be searched
   * @param timeout maximum time allowed for the lookup operation
   * @return a proxy for the service, or null is no service can be
   *  found with the specified template
   * @throws IOException an error occurred in doing lookup
   */
  public static Object lookup(ServiceTemplate template, int timeout)
    throws IOException
  {
    if (manager == null) 
      manager = new LookupManager();
    return manager.searchDiscovered(template, timeout);
  }

  /**
   * Perform a lookup in the lookup service, searching for the given
   * service name.
   * 
   * @param serviceName the service name to lookup
   * @return the proxy for the service
   * @throws IOException
   */
  public static Object lookup(String serviceName)
    throws IOException
  {
    Entry[] attr = new Entry[] { new GroupEntry(serviceName) };
    ServiceTemplate template = new ServiceTemplate(null, null, attr);
    return lookup(template);
  }

  
 /**
  * Perform a lookup for the given <code>Name</code>.
  * 
  * @param name attribute name to lookup
  * @return the proxy for the service
  */ 
 public static Object lookup(Name name)
   throws IOException
 {
   Entry[] attr = new Entry[]{name};
   ServiceTemplate template = new ServiceTemplate(null, null, attr);
   return lookup(template);
 }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private List<ServiceRegistrar> registrars = new LinkedList<ServiceRegistrar>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Private constructor for singleton use
   * @throws IOException an error occurred in starting discovery
   */
  private LookupManager()
    throws IOException
  {
    if (System.getSecurityManager() == null)
      System.setSecurityManager(new RMISecurityManager());

    // Create lookup discovery object and have it notify us
    LookupDiscovery ld = new LookupDiscovery(DiscoveryGroupManagement.ALL_GROUPS);
    ld.addDiscoveryListener(this);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DiscoveryListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public synchronized void discovered(DiscoveryEvent ev)
  {
    ServiceRegistrar[] regs = ev.getRegistrars();
    for (int i=0; i < regs.length; i++)
      registrars.add(regs[i]);
    notifyAll();
  }

  public synchronized void discarded(DiscoveryEvent ev)
  {
    ServiceRegistrar[] regs = ev.getRegistrars();
    for (int i=0; i < regs.length; i++)
      registrars.remove(regs[i]);
    notifyAll();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Method
  ////////////////////////////////////////////////////////////////////////////////////////////

  private synchronized Object searchDiscovered(ServiceTemplate template, int timeout)
    throws RemoteException
  {
    long end = System.currentTimeMillis() + timeout;
    long timeLeft = end - System.currentTimeMillis();
    while (timeLeft > 0 && registrars.isEmpty()) {
      try { wait(timeLeft); } catch (InterruptedException e) { }
      timeLeft = end - System.currentTimeMillis();
    }
    if (!registrars.isEmpty()) {
      ServiceRegistrar reg = (ServiceRegistrar) registrars.get(0);
      return reg.lookup(template);
    } else {
      return null;
    }
  }
  
} // END LookupManager
