/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gm;

import java.io.IOException;
import java.rmi.Remote;
import java.rmi.server.ExportException;

import jgroup.core.ExternalGMIService;
import jgroup.jini.GroupEntry;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.gmi.JeriGroupExporter;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.LookupDiscovery;
import net.jini.export.Exporter;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;

import org.apache.log4j.Logger;

/**
 * This layer is used to make the joining process into
 * greg transparent, when creating an application.
 * 
 * @author Rohnny Moland
 * @since Jgroup 2.2
 */
public class GregRegistryLayer 
  implements GregRegistryService, ServiceIDListener
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GregRegistryLayer.class);
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The service proxy */  
  private ExternalGMIService egmis;
  
  /** The server object using this layer. */
  private Object server;

  /** The remote proxy */
  private Remote proxy;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  private GregRegistryLayer(ExternalGMIService egmis)
  {
    this.egmis = egmis;
  }

  public static GregRegistryLayer getLayer(ExternalGMIService egmis)
  {
    return new GregRegistryLayer(egmis);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Layer
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.Layer#addListener(java.lang.Object)
   */
  public void addListener(Object listener) 
  {
    if (listener == null)
      throw new NullPointerException("No replica specified for the registry layer.");
    this.server = listener;

    /** Get the application object for this replica */
    AppConfig thisApp = AppConfig.getApplication(listener);
    
    String[] groups = new String[] {thisApp.getParam("lookup.group", "jgroup.sf.net")};
    if (!thisApp.getBooleanParam("GregRegistry.auto"))
      return;
    
    /** Joins the app into greg */
    join(thisApp.getRegistryName(), groups);
    if (log.isDebugEnabled())
      log.debug("Successfully joined: " + thisApp.getRegistryName());    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from GregRegistryService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.GregRegistryService#join(java.lang.String)
   */
  public void join(String serviceName, String[] groups)
  {
    Exporter exporter = new JeriGroupExporter(serviceName);
    try {
      proxy = exporter.export((Remote) server);
      Entry[] attrs = new Entry[] { new GroupEntry(serviceName) };
      // FIXME: Register the lease time also in greg.   
      LookupDiscovery ld = new LookupDiscovery(groups);
      JoinManager joinMgr = new JoinManager(proxy, attrs, this, ld, null);
    } catch (ExportException e) {
      log.error(e.getMessage());      
    } catch (IOException e) {
      log.error(e.getMessage());
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ServiceIDListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.lookup.ServiceIDListener#serviceIDNotify(net.jini.core.lookup.ServiceID)
   */
  public void serviceIDNotify(ServiceID serviceID) 
  {
    if (server instanceof ServiceIDListener) {
      ((ServiceIDListener) server).serviceIDNotify(serviceID);
    }
    if (log.isDebugEnabled())
      log.debug("Service ID (from layer): " + serviceID);
  }

} // END GregRegistryLayer
