/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.gm;

import java.io.InputStream;
import java.io.OutputStream;
import java.rmi.RemoteException;

import jgroup.core.GroupManager;
import jgroup.core.InternalGMIService;
import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.gm.DispatcherService;
import jgroup.relacs.rmi.SynchAckListener;

import org.apache.log4j.Logger;


/**
 *  Test layer for remote interaction styles; for the purpose of
 *  testing the IGMI and Mcast layer performance.
 *
 *  @author Hein Meling
 */
public class RemoteTestLayer
  implements RemoteTestService, MulticastListener, InternalRemoteTest
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(RemoteTestLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Values for the test runs */
  private static final int NUM_CALLS = 2000;
  private static final int INIT_CALLS = 2000;
  private static final int INIT_SIZE = 0;
  private static final int INC_SIZE = 5000;
  private static final int MAX_SIZE = 20000;

  /** Protocol name used to distinguish messages from other protocols. */
  private static final String PROTOCOL_NAME = "RTL";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The group manager instance that is in the process of being constructed */
  private GroupManager gm;

  /** The multicast service */
  private MulticastService multicastService;

  /** The dispatcher service */
  private DispatcherService dispatcherService;

  /** The internal remote test interface */
  private InternalRemoteTest irt;

  /** Configurable number of calls to perform. */
  private int numCalls = NUM_CALLS;

  /** Configurable number of calls to perform as part of the initialization. */
  private int initCalls = INIT_CALLS;

  /** Configurable size for the first run. */
  private int initSize = INIT_SIZE;

  /** Configurable size to increment for each run. */
  private int incSize = INC_SIZE;

  /** Configurable size for which we should not exceed. */
  private int maxSize = MAX_SIZE;

  /** The interaction style to use between the RemoteTestLayer instances. */
  private String interactionStyle;

  /** Set to true if the iteractions should halt. */
  private volatile boolean halted = false;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private RemoteTestLayer(GroupManager gm, DispatcherService dispatcher, MulticastService multicastService)
    throws JgroupException, RemoteException
  {
    this.gm = gm;
    this.dispatcherService = dispatcher;
    this.multicastService = multicastService;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Get an instance of this layer; note that this layer needs the group manager
   *  object and this must be passed to the layer as the first argument to the
   *  getLayer method.
   */
  public static RemoteTestLayer getLayer(GroupManager gm, DispatcherService dispatcher,
                                         InternalGMIService igmis, MulticastService mcast)
    throws JgroupException, RemoteException
  {
    /*
     * We ignore the IGMI service, since we don't need to invoke 
     * addListener() on the igmis object.
     */
    return new RemoteTestLayer(gm, dispatcher, mcast);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a server or layer that is listening for remote test layer events.
   *  Actually, there is no events, so this is used for bootstrapping the
   *  layer.  That is to read the configuration parameter.
   */
  public void addListener(Object listener)
  {
    AppConfig app = AppConfig.getApplication(listener);
    interactionStyle = app.getParam("RemoteTest.interactionStyle");
    numCalls = app.getIntParam("RemoteTest.numCalls", NUM_CALLS);
    initCalls = app.getIntParam("RemoteTest.initCalls", INIT_CALLS);
    initSize = app.getIntParam("RemoteTest.initSize", INIT_SIZE);
    incSize = app.getIntParam("RemoteTest.incSize", INC_SIZE);
    maxSize = app.getIntParam("RemoteTest.maxSize", MAX_SIZE);
    if (interactionStyle.length() == 0) {
      /* IGMI is the default */
      interactionStyle = "igmi";
    } else {
      interactionStyle = interactionStyle.toLowerCase();
      if (!(interactionStyle.equals("igmi")
         || interactionStyle.equals("mcast-obj")
         || interactionStyle.equals("mcast-stream"))) {
        throw new IllegalArgumentException("Unsupported interaction style: "
          + interactionStyle);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from RemoteTestService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @see jgroup.test.gm.RemoteTestService#activate()
   */
  public void activate()
  {
    irt = (InternalRemoteTest) gm.getService(InternalRemoteTest.class);
    System.out.println("\n\n-- RemoteTest: " + interactionStyle + " ---------------------------------------");
    int calls = numCalls + initCalls;
    long start = 0;
    long[] timeVals = new long[calls];
    byte[] buffer = new byte[0];
    for (int size = initSize; size <= maxSize; size += incSize) {
      buffer = new byte[size];
      for (int i = 0; !halted && i < calls; i++) {
        try {
          if (interactionStyle.equals("igmi")) {
//            start = System.nanoTime();
            irt.rtest(buffer);
//            timeVals[i] = System.nanoTime() - stop;
          } else if (interactionStyle.equals("mcast-obj")) {
//            start = System.nanoTime();
            SynchAckListener listener = new SynchAckListener(dispatcherService);
            multicastService.mcast(PROTOCOL_NAME, "Some object", listener.getRemoteAckListener());
            Object result = listener.getResults();
//            timeVals[i] = System.nanoTime() - stop;
          } else if (interactionStyle.equals("mcast-stream")) {
//            start = System.nanoTime();
            OutputStream msg = multicastService.getMessage(PROTOCOL_NAME);
            msg.write(buffer);
            SynchAckListener listener = new SynchAckListener(dispatcherService);
            multicastService.mcast(msg, listener.getRemoteAckListener());
            Object result = listener.getResults();
//            timeVals[i] = System.nanoTime() - stop;
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
      int begin = 0;
      if (calls != numCalls) {
        System.out.println("JIT Compiling and optimization phase (initCalls):");
        System.out.println("InitCalls phase - Size=" + size + "--------");
        for (int i = 0; i < initCalls; i++) {
          System.out.println(timeVals[i]);
        }
        begin = initCalls;
      }
      System.out.println("\n\n-- RemoteTest: " + interactionStyle + " ---------------------------------------");
      System.out.println("Normal run phase - Size=" + size + "--------");
      for (int i = begin; i < calls; i++) {
        System.out.println(timeVals[i]);
      }
      calls = numCalls;
    }
    System.out.println("------------------ COMPLETED ------------------");
  }


  public void halt()
  {
    halted = true;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalRemoteTest
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @see jgroup.test.gm.InternalRemoteTest#rtest()
   */
  public void rtest(int[] array)
    throws RemoteException
  {
    return;
  }

  /**
   * @see jgroup.test.gm.InternalRemoteTest#rtest(int)
   */
  public Object rtest(byte[] var)
    throws RemoteException
  {
    return var;
  }

  /**
   * @see jgroup.test.gm.InternalRemoteTest#rtest(int)
   */
  public Object rtest(int var)
    throws RemoteException
  {
    return new Integer(var);
  }

  /**
   * @see jgroup.test.gm.InternalRemoteTest#rtest(Object)
   */
  public Object rtest(Object obj)
    throws RemoteException
  {
    return obj;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * @see jgroup.core.multicast.MulticastListener#deliverObject(Object)
   */
  public Object deliverObject(Object msg, MemberId sender, int seqNo)
  {
    log.debug("Received: " + msg);
    return msg;
  }

  /**
   * @see jgroup.core.multicast.MulticastListener#deliverStream(InputStream)
   */
  public Object deliverStream(InputStream msg, MemberId sender, int seqNo)
  {
//    try {
//      log.debug("Received: " + msg.read());
//    } catch (java.io.IOException e) {
//    }
    return null;
  }

  /**
   * @see jgroup.core.multicast.MulticastListener#getProtocolName()
   */
  public String getProtocolName()
  {
    return PROTOCOL_NAME;
  }

} // End of RemoteTestLayer
