/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.hello;

import java.rmi.RemoteException;

import jgroup.core.ExternalGMIService;
import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.protocols.Atomic;
import jgroup.core.registry.RegistryService;
import jgroup.util.Network;

/**
 *  Simple Hello server aimed at demonstrating the capabilities of atomic
 *  external method invocations.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.0
 */
public class AtomicHelloServer
  implements Hello, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Stores the answer for replying to external invocations */
  private Answer answer;

  /** Reference to the partitionable group membership service */
  private MembershipService membershipService;

  /** Reference to the external group method invocation service */
  private ExternalGMIService externalGMIService;

  /** The binding identifier for the servers stub in the dependable registry */
  private IID bindId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method (initialize the HelloServer object)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String argv[])
    throws Exception
  {
    AtomicHelloServer obj = new AtomicHelloServer();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor for the HelloServer
  ////////////////////////////////////////////////////////////////////////////////////////////

  public AtomicHelloServer()
    throws Exception
  {
    System.out.println("Server starting...");
    /* Prepare the answer for replying to external invocations. */
    answer = new Answer("Hello from " + Network.getLocalHostName());
    System.out.println("Server is connected to a network...");

    /* Obtain the group manager */
    GroupManager gm = GroupManager.getGroupManager(this);
    System.out.println("Server obtained group manager");

    /* Obtain proxies for the services required by the HelloServer */
    membershipService = (MembershipService) gm.getService(MembershipService.class);
    externalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    RegistryService registryService = (RegistryService) gm.getService(RegistryService.class);

    /* Join the group and bind the server in the dependable registry */
    membershipService.join(11);
    bindId = registryService.bind(this);
    System.out.println("Server joined group 11");
    MemberId thisMember = membershipService.getMyIdentifier();
    System.out.println("My memberId: " + thisMember + ", " + thisMember.getLocalId().getMemberNo());
    System.out.println("Server ready and bound to the dependable registry");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Hello interface (External Group Method Invocation)
  ////////////////////////////////////////////////////////////////////////////////////////////

  @Atomic public Answer sayHello()
    throws RemoteException
  {
    System.out.println("sayHello: returning " + answer);
    System.out.println("sayHello identifier: " + externalGMIService.getIdentifier());
    return answer;
  }

  @Atomic public double doCompute()
    throws RemoteException
  {
    System.out.println("Starting computatations ...");
    long start = System.currentTimeMillis();
    long limit = 100000l;
    double dd = 0.0d;
    double ret = 0.0d;
    for (long l = -limit; l < limit; l++) {
      dd = Math.sin(3.141576d*l)*l;
      ret += dd*2.00d;
    }
    long end = System.currentTimeMillis();
    System.out.println("Computations took: " + (end-start) + " msec");
    return ret;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  @AllowDuplicateViews public void viewChange(View view)
  {
    System.out.println("  ** ATOMIC HelloServer **" + view);
  }

  public void prepareChange()
  {
    System.out.println("The current view is invalid; please await a new view...");
  }

  public void hasLeft()
  {
    System.out.println("I have left the group");
  }

} // END AtomicHelloServer
