/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.hello;

import java.rmi.RemoteException;

import jgroup.core.ExternalGMIService;
import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.registry.RegistryService;
import jgroup.util.Network;

/**
 *  Simple Hello server aimed at demonstrating the capabilities of both
 *  internal and external invocations.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.0
 */
public class ExperimentalHelloServer
  implements Hello, InternalHello, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Stores the answer for replying to external invocations */
  private Answer answer;

  /** Reference to the partitionable group membership service */
  private MembershipService membershipService;

  /** Reference to the external group method invocation service */
  private ExternalGMIService externalGMIService;

  /** Reference to the internal hello group proxy (for group internal invocations) */
  private InternalHello internalHello;

  /** The binding identifier for the servers stub in the dependable registry */
  private IID bindId;

  /** This member's member identifier */
  private MemberId thisMember;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method (initialize the HelloServer object)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String argv[])
    throws Exception
  {
    ExperimentalHelloServer obj = new ExperimentalHelloServer();
    /*
     * Note that, at this point we do not exit since the dependable
     * registry reference holds a strong reference to this
     * <code>HelloServer</code> object, and thus prevents it from being
     * garbage collected.  This again prevents the server from exiting
     * at this point.  So, to stop the server from running we must
     * either, use CTRL-C or unbind the helloserver reference in the
     * dependable registry.
     */
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor for the HelloServer
  ////////////////////////////////////////////////////////////////////////////////////////////

  public ExperimentalHelloServer()
    throws Exception
  {
    System.out.println("Server starting...");
    /* Prepare the answer for replying to external invocations. */
    answer = new Answer("Hello from " + Network.getLocalHostName());
    System.out.println("Server is connected to a network...");

    /* Obtain the group manager */
    GroupManager gm = GroupManager.getGroupManager(this);
    System.out.println("Server obtained group manager");

    /* Obtain proxies for the services required by the HelloServer */
    membershipService = (MembershipService) gm.getService(MembershipService.class);
    externalGMIService = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    RegistryService registryService = (RegistryService) gm.getService(RegistryService.class);
    internalHello = (InternalHello) gm.getService(InternalHello.class);

    /* Join the group and bind the server in the dependable registry */
    membershipService.join(12);
    System.out.println("Server joined group 12");
    thisMember = membershipService.getMyIdentifier();
    System.out.println("My memberId: " + thisMember + ", " + thisMember.getLocalId().getMemberNo());
    bindId = registryService.bind(this);
    System.out.println("Server ready and bound to the dependable registry");

    /*
     * Code to test the membership.leave() method; invoked after
     * being member of the group for some 15 seconds.
     */
    Thread.sleep(15000);
    System.out.println("I'm leaving the group");
    membershipService.leave();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Hello interface (External Group Method Invocation)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Answer sayHello()
    throws RemoteException
  {
    System.out.println("sayHello: returning " + answer);
    System.out.println("sayHello indentifier: " + externalGMIService.getIdentifier());
    return answer;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the InternalHello interface (Internal Group Method Invocation)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Object time()
    throws RemoteException
  {
    long tmp = System.currentTimeMillis();
    System.out.println("Generating time: " + tmp);
    return new Long(tmp);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View view)
  {
    System.out.println("--- HelloServer.viewChange: ---" + view);
  }

  public void prepareChange()
  {
    System.out.println("The current view is invalid; please await a new view...");
  }

  public void hasLeft()
  {
    System.out.println("I have left the group");
  }


  public double doCompute() throws RemoteException
  {
    return 0;
  }

} // END ExperimentalHelloServer
