/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.hello;

import java.rmi.RemoteException;

import jgroup.core.GroupManager;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.View;
import jgroup.core.protocols.Multicast;
import jgroup.util.Network;

/**
 *  Simple Hello server aimed at demonstrating the capabilities of both
 *  internal and external invocations.  This server is modified to be used
 *  with the automatic join and bind of the PGMS and RegistryLayer, and to
 *  test things like the LeaseLayer and NotifyLayer.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.0
 */
public class LeaseServer
  implements Hello, InternalHello, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Stores the answer for replying to external invocations */
  private Answer answer;

  /** The group manager reference */
  private GroupManager gm = null;

  /** Reference to the internal hello group proxy (for group internal invocations) */
  private InternalHello internalHello;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main method (initialize the LeaseServer object)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String argv[])
    throws Exception
  {
    LeaseServer obj = new LeaseServer();
    /*
     * Note that, at this point we do not exit since the dependable
     * registry reference holds a strong reference to this
     * <code>LeaseServer</code> object, and thus prevents it from being
     * garbage collected.  This again prevents the server from exiting
     * at this point.  So, to stop the server from running we must
     * either, use CTRL-C or unbind the helloserver reference in the
     * dependable registry.
     */
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor for the LeaseServer
  ////////////////////////////////////////////////////////////////////////////////////////////

  public LeaseServer()
    throws Exception
  {
    System.out.println("Server starting...");
    /* Prepare the answer for replying to external invocations. */
    answer = new Answer("Hello from " + Network.getLocalHostName());

    /* Obtain the group manager */
    gm = GroupManager.getGroupManager(this);
    System.out.println("Server obtained group manager");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Hello interface (External Group Method Invocation)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Answer sayHello()
    throws RemoteException
  {
    System.out.println("sayHello: returning " + answer);
    return answer;
  }

  @Multicast public double doCompute() throws RemoteException
  {
    return 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the InternalHello interface (Internal Group Method Invocation)
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Object time()
    throws RemoteException
  {
    long tmp = System.currentTimeMillis();
    System.out.println("Generating time: " + tmp);
    return new Long(tmp);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View view)
  {
    if (internalHello == null) {
      /* Obtain the reference for this groups group manager */
      gm = GroupManager.getGroupManager(view.getGid());
      /* Obtain proxy for the internal hello service required by the LeaseServer */
      internalHello = (InternalHello) gm.getService(InternalHello.class);
    }

    System.out.println("View id: " + view.getVid());
    MemberId[] members = view.getMembers();
    for (int i = 0; i < members.length; i++)
      System.out.println("Member[" + i + "]: " + members[i]);

    try {
      /*
       * The time() method is defined in the InternalHello interface and
       * is marked as a group internal method.  By definition, all group
       * internal methods will return an array of values instead of a
       * single value as with standard remote (or external group) method
       * calls.
       */
      Object[] objs = (Object[]) internalHello.time();
      for (int i = 0; i < objs.length; i++)
        System.out.println("Time: " + objs[i]);
      answer.setTime(objs);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void prepareChange()
  {
  }

  public void hasLeft()
  {
  }

} // END LeaseServer
