package jgroup.test.performance;

import java.io.InputStream;
import java.rmi.RemoteException;

import jgroup.core.ConfigManager;
import jgroup.core.GroupManager;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MergingListener;
import jgroup.core.View;
import jgroup.core.multicast.MulticastListener;
import jgroup.relacs.gm.MembershipLayer;
import jgroup.relacs.gm.MergingLayer;
import jgroup.relacs.gm.MulticastLayer;
import jgroup.relacs.gm.NewDispatcherLayer;
import jgroup.relacs.gmi.JeriEGMILayer;
import jgroup.relacs.rmi.IntGroupHandler;
import jgroup.test.hello.Answer;
import jgroup.test.hello.Hello;
import jgroup.test.hello.InternalHello;

/**
 *  Simple class to test the performance of the dynamic
 *  <code>GroupManager</code> composition framework.  To compare
 *  with static composition.
 *
 *  @author Hein Meling <meling@acm.org>
 */
public final class PerfGroupManager
  implements Hello, InternalHello, MembershipListener, MulticastListener, MergingListener
{

  private static final int NUM_ROUNDS = 10;

  /**
   * Constructor for PerfGroupManager.
   */
  public PerfGroupManager() { }

  public void staticComposition()
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, emax = 0, eavg = 0, avg = 0, start = 0, time = 0;

    start = System.currentTimeMillis();
    for (int i=0; i < NUM_ROUNDS; i++) {

      // Perform test and measure times
      start = System.currentTimeMillis();

      NewDispatcherLayer disp = NewDispatcherLayer.getLayer();
      MembershipLayer pgms = MembershipLayer.getLayer(disp);
      MulticastLayer mcast = MulticastLayer.getLayer(disp, pgms);
      //FIXME gm must be set
      JeriEGMILayer egmi = JeriEGMILayer.getLayer(null, mcast, pgms);
      IntGroupHandler igmi = IntGroupHandler.getLayer(mcast);
      MergingLayer sms = MergingLayer.getLayer(pgms, mcast);
      
      disp.addListener(pgms);
      pgms.addListener(this);
      pgms.addListener(sms);
      mcast.addListener(sms);
      mcast.addListener(egmi);
      mcast.addListener(igmi);
      egmi.addListener(this);
      igmi.addListener(this);
      sms.addListener(this);

      // Compute statics
      time = System.currentTimeMillis() - start;
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;
      if (time != max) {
        eavg += time;
        if (time > emax)
          emax = time;
      }
    }
    // Print statistics
    System.out.println((avg/NUM_ROUNDS) + " ; " + (eavg/(NUM_ROUNDS-1)) + " ; " + emax + " ; " + max + " ; " + min);
  }

  public void dynamicComposition(String[] serviceSet)
    throws Exception
  {
    /* Min, max and average time; initialization */
    long min = Long.MAX_VALUE, max = 0, emax = 0, eavg = 0, avg = 0, start = 0, time = 0;

    start = System.currentTimeMillis();
    for (int i=0; i < NUM_ROUNDS; i++) {

      // Perform test and measure times
      start = System.currentTimeMillis();

      GroupManager gm = GroupManager.getGroupManager(99, serviceSet, this);

      // Compute statics
      time = System.currentTimeMillis() - start;
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;
      if (time != max) {
        eavg += time;
        if (time > emax)
          emax = time;
      }
    }
    // Print statistics
    System.out.println((avg/NUM_ROUNDS) + " ; " + (eavg/(NUM_ROUNDS-1)) + " ; " + emax + " ; " + max + " ; " + min);
  }


  public static void main(String[] args)
    throws Exception
  {
    ConfigManager.init();
    PerfGroupManager perfObj = new PerfGroupManager();

    if (args[0].equals("--static")) {
      perfObj.staticComposition();
    } else if (args[0].equals("--dynamic")) {
      String[] serviceSet = new String[] { "Dispatcher", "PGMS", "Mcast", "EGMI", "IGMI", "SMS" };
      perfObj.dynamicComposition(serviceSet);
    }

    System.exit(0);
  }

  /**
   *  The external interface.
   */
  public Answer sayHello()
    throws RemoteException
  {
    return new Answer("Test");
  }


  /**
   *  The internal interface.
   */
  public Object time()
    throws RemoteException
  {
    return new Integer(2);
  }

  /**
   *  The <code>MembershipListener</code> interface.
   */
  public void viewChange(View v){}

  public void hasLeft(){}

  public void prepareChange(){}

  /**
   *  The <code>MulticastListener</code> interface.
   */
  public String getProtocolName()
  {
    return "Test";
  }

  public Object deliverStream(InputStream msg, MemberId sender, int seqNo)
  {
    throw new UnsupportedOperationException();
  }

  public Object deliverObject(Object msg, MemberId sender, int seqNo)
  {
    throw new UnsupportedOperationException();
  }
  
  /**
   *  The <code>MergingListener</code> interface.
   */
  public Object getState(MemberId[] dests)
  {
    return this;
  }

  public void putState(Object status, MemberId[] sources) {}

  public double doCompute() throws RemoteException
  {
    return 0;
  }
}
