/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.test.performance;

import java.rmi.RemoteException;
import java.util.Timer;
import java.util.TimerTask;

import jgroup.core.GroupUnreachableException;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.RegistryFactory;
import jgroup.util.Network;
import jgroup.util.log.Eventlogger;

/**
 *  Test class for measuring the performance of GMI invocations.  This client
 *  can run tests with several different servers, but only one at a time.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class PerformanceClient
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The possible 'measurement.type' property values */
  private static final String DELAY_MEASUREMENT      = "delay";
  private static final String THROUGHPUT_MEASUREMENT = "throughput";

  /**
   * Supported properties: Note that if additional properties are to
   * be added here, recall to also add them to the dev-build.xml file;
   * the 'perfclient' target.
   */
  public static final String[] properties = new String[] {
      "calibration.period",
      "request.load.duration",
      "measurement.type",
      "requests.per.second",
      "invocation.type",
      "buffer.size",
  };

  /**
   *  Set of servers supported for running tests through
   *  the <code>SpeedTest</code> interface.
   */
  private static final String[] supportedServers = new String[] {
    "Jgroup/PerformanceServer",
    "ARM/ReplicatedServer",
    "Jgroup/SpeedTest",
    "UpgradeServer"
  };

  /** Number of milliseconds in a second */
  private static final int TIME_UNIT = 1000;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The speed test server reference */
  private static SpeedTest test = null;

  /** The delay between each request performed by this client */
  private static int requestPeriod;

  /** The invocation type to be used in the measurements */
  private static String invocationType;

  /** The buffer used with buffer invocations */
  private static byte[] buffer;

  /** Used to indicate that the calibration period has completed */
  private static volatile boolean trafficGenerationCompleted = false;

  /**
   * Used to indicate that logging has completed,
   * and that the next generation sequence can commence.
   */
  private static volatile boolean loggingCompleted = false;

  /**
   *  True when the invocation delays are to be logged, false otherwise.
   *  Typically, we only log the invocation delay after calibration and
   *  stop logging as soon as the actual measurement period is complete,
   *  even though we perform additional invocations for a small period
   *  to maintain a constant load on the servers.
   */
  private static volatile boolean logDelays = false;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws Exception
  {
    String localHost = Network.getLocalMachineName();
    System.out.println("Running PerformanceClient: " + localHost);
    /*
     * Undefine/clear properties for which ant did not receive any input;
     * there were no value to assign to the property on the input line.
     */
    for (int i = 0; i < properties.length; i++) {
      String prop = System.getProperty(properties[i]);
      if (prop != null && prop.startsWith("${")) {
        System.clearProperty(properties[i]);
      }
    }
    DependableRegistry reg = RegistryFactory.getRegistry();
    String[] servers = reg.list();
    for (int i = 0; test == null && i < servers.length; i++) {
      for (int j = 0; test == null && j < supportedServers.length; j++) {
        if (servers[i].equals(supportedServers[j])) {
          test = (SpeedTest) reg.lookup(servers[i]);
          System.out.println("Found supported server: " + servers[i]);
        }
      }
    }
    if (test == null) {
      System.out.println("No supported servers found");
      System.exit(0);
    }

    /*
     * Retreive system property values for use by the performance client
     */
    int calibrationPeriod = Integer.getInteger("calibration.period", 20).intValue();
    int requestDuration = Integer.getInteger("request.load.duration", 60).intValue();
    String measurementType = System.getProperty("measurement.type", DELAY_MEASUREMENT);
    if (measurementType.equalsIgnoreCase(DELAY_MEASUREMENT)) {
      int requestRate = Integer.getInteger("requests.per.second", 2).intValue();
      requestPeriod = TIME_UNIT / requestRate;
    } else if (measurementType.equalsIgnoreCase(THROUGHPUT_MEASUREMENT)) {
      requestPeriod = -1;
    }
    invocationType = System.getProperty("invocation.type", "anycast");
    String invType = invocationType;
    if (invocationType.endsWith("-buffer")) {
      int bufSize = Integer.getInteger("buffer.size", 1000).intValue();
      buffer = new byte[bufSize];
      invType = invocationType + "(" + bufSize + ")";
    }

    /*
     * Generate traffic to calibrate the clients.  That is, give all clients
     * the chance to start up and begin to perform request invocations, and
     * reduce the impact of JVM initialization, such as class loading etc.
     */
    System.out.println("Calibration: " + localHost);
    generateTraffic(calibrationPeriod);
    // Calibration is complete; perform the actual traffic generation
    logDelays = true;
    System.out.println("Begin logging: " + localHost);
    Eventlogger.logEvent("BeginTrafficGeneration: " + invType);
    generateTraffic(requestDuration);
    Eventlogger.logEvent("EndTrafficGeneration: " + invType);
    System.out.println("End logging: " + localHost);
    logDelays = false;
    Eventlogger.close();
    /*
     * Traffic generation complete; generate some addtional traffic
     * to avoid that some clients experience a less load at the
     * end of the experiment.
     */
    generateTraffic(5);
    System.out.println("Exiting: " + localHost);
  }


  /**
   * Generate traffic at a <i>fixed-rate</i> for the given duration.
   * @param duration
   *   The duration to generate traffic at a fixed-rate.
   */
  public static void generateTraffic(final int duration)
  {
    final Timer timer = new Timer("GenerateTraffic", true);
    if (requestPeriod > 0) {
      TimerTask timerTask = getDelayTimerTask();
      timer.scheduleAtFixedRate(timerTask, 0, requestPeriod);
    } else {
      TimerTask timerTask = getThroughputTimerTask();
      timer.schedule(timerTask, 0);
    }
    new Thread("StopTrafficGeneration") {
      public void run() {
        try { sleep(duration*TIME_UNIT);
        } catch (InterruptedException e) { }
        // Indicate the end of the current traffic generation period
        trafficGenerationCompleted = true;
      }
    }.start();
    /*
     * Await the completion of the current traffic generation period.
     */
    while (!trafficGenerationCompleted || !loggingCompleted) {
      Thread.yield();
    }
    // Stop scheduling client requests
    timer.cancel();
    // Reset the thread synchronization variables
    trafficGenerationCompleted = false;
    loggingCompleted = false;
  }


  /**
   * Returns a <code>TimerTask</code> that should be scheduled only once,
   * as the run method of the timer task will perform repeated invocations
   * and measure the duration, and count the number of invocations to
   * determine the throughput measure (Inovcations / Second).
   */
  public static TimerTask getThroughputTimerTask()
  {
    return new TimerTask() {
      public void run() {
        long invocationCounter = 0;
        long start = System.currentTimeMillis();
        try {
          if (invocationType.equalsIgnoreCase("anycast")) {
            while (!trafficGenerationCompleted) {
              test.test();
              invocationCounter++;
            }
          } else if (invocationType.equalsIgnoreCase("anycast-buffer")) {
            while (!trafficGenerationCompleted) {
              test.test(buffer);
              invocationCounter++;
            }
          } else if (invocationType.equalsIgnoreCase("multicast")) {
            while (!trafficGenerationCompleted) {
              test.mtest();
              invocationCounter++;
            }
          } else if (invocationType.equalsIgnoreCase("multicast-buffer")) {
            while (!trafficGenerationCompleted) {
              test.mtest(buffer);
              invocationCounter++;
            }
          } else {
            throw new IllegalStateException("Unknown invocation type: " + invocationType);
          }
        } catch (RemoteException e) {
          trafficGenerationCompleted = true;
          e.printStackTrace();
          Eventlogger.logEvent(e.getMessage());
        } catch (GroupUnreachableException e) {
          /*
           * The group has become unreachable;
           * stop traffic generation and log the event.
           */
          trafficGenerationCompleted = true;
          Eventlogger.logEvent("GroupUnreachable");
        }
        double invocDuration = (System.currentTimeMillis() - start) / TIME_UNIT;
        double throughput = invocationCounter / invocDuration;
        // log the throughput value
        if (logDelays) {
          Eventlogger.logEventNoSysTime(Double.toString(throughput));
        }
        loggingCompleted = true;
      }
    };
  }


  /**
   * Returns a <code>TimerTask</code> that can be scheduled at 
   * <i>fixed-rate</i> intervals, as required by the performance
   * measurements.
   */
  public static TimerTask getDelayTimerTask()
  {
    return new TimerTask() {
      public void run() {
        try {
          long invocationDelay = 0;
          if (invocationType.equalsIgnoreCase("anycast")) {
            long start = System.nanoTime();
            test.test();
            invocationDelay = System.nanoTime() - start;
          } else if (invocationType.equalsIgnoreCase("anycast-buffer")) {
            long start = System.nanoTime();
            test.test(buffer);
            invocationDelay = System.nanoTime() - start;
          } else if (invocationType.equalsIgnoreCase("multicast")) {
            long start = System.nanoTime();
            test.mtest();
            invocationDelay = System.nanoTime() - start;
          } else if (invocationType.equalsIgnoreCase("multicast-buffer")) {
            long start = System.nanoTime();
            test.mtest(buffer);
            invocationDelay = System.nanoTime() - start;
          } else {
            throw new IllegalStateException("Unknown invocation type: " + invocationType);
          }
          // log the invocation delay
          if (logDelays) {
            Eventlogger.logEvent(Long.toString(invocationDelay));
          }
        } catch (RemoteException e) {
          trafficGenerationCompleted = true;
          e.printStackTrace();
          Eventlogger.logEvent(e.getMessage());
        } catch (GroupUnreachableException e) {
          /*
           * The group has become unreachable;
           * stop traffic generation and log the event.
           */
          trafficGenerationCompleted = true;
          Eventlogger.logEvent("GroupUnreachable");
        } finally {
          loggingCompleted = true;
        }
      }
    };
  }

} // END PerformanceClient
