/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util.log;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.logging.LogManager;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * JDK LogManager with support for using system
 * property references.
 * <p>
 * E.g for setting the system property log.dir, 
 * you can use ${log.dir} in the java logging.properties
 * file.
 * 
 * @author Rohnny Moland
 */
public class JdkLogManager
  extends LogManager
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Buffer holding the property value */
  private StringBuffer buf;
  
  /** Default log dir */
  private static final String DEFAULT_LOG_DIR = File.separator + "tmp" + File.separator + "jgroup-logs";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public JdkLogManager()
  {
    buf = new StringBuffer();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Find all ${..} and replace with the system
   * property if exist.
   */
  @Override
  public String getProperty(String name)
  {
    String property = super.getProperty(name);
    if (property == null)
      return property;
    buf.setLength(0);
    Pattern p = Pattern.compile("\\$\\{[a-zA-Z0-9.]+\\}");
    Matcher m = p.matcher(property);
    while (m.find()) {
      String token = m.group();
      String sysprop = System.getProperty((token.subSequence(2,token.length()-1)).toString());
      if (sysprop != null)
        m.appendReplacement(buf, sysprop);
    }
    m.appendTail(buf);
    return buf.toString();
  }
  
  /**
   * Create the log directory if it does not
   * exist.
   */
  @Override
  public void readConfiguration(InputStream in)
    throws IOException, UnknownHostException
  {
    super.readConfiguration(in);
    System.setProperty("jgroup.log.machine", getMachineName());
    System.setProperty("jgroup.log.time", getTimeString());
    String logDir = System.getProperty("jgroup.log.dir", DEFAULT_LOG_DIR);
    File dir = new File(logDir);
    if (!dir.exists()) {
      /* Directory does not exist; create it. */
      if (!dir.mkdir())
        throw new IOException("Failed to create log directory");
    }
  }

  /**
   * @see jgroup.util.Network#getMachineName(String)
   * (Added here to avoid classloader problems)
   */
  private static String getMachineName()
    throws UnknownHostException
  {
    String hostName = InetAddress.getLocalHost().getHostName();
    int machEndIndex = hostName.indexOf('.');
    if (machEndIndex == -1)
      return hostName;
    else
      return hostName.substring(0, machEndIndex);
  }
  
  /**
   * @see jgroup.util.Util#getTimeString()
   * (Added here to avoid classloader problems)
   */
  private static String getTimeString()
  {
    Calendar gc = Calendar.getInstance();
    StringBuilder buffer = new StringBuilder();
    int hour = gc.get(Calendar.HOUR_OF_DAY);
    if (hour < 10)
      buffer.append("0");
    buffer.append(hour);
    int minute = gc.get(Calendar.MINUTE);
    if (minute < 10)
      buffer.append("0");
    buffer.append(minute);
    int second = gc.get(Calendar.SECOND);
    if (second < 10)
      buffer.append("0");
    buffer.append(second);
    return buffer.toString();
  }
  
} // END JdkLogManager
