/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package com.sun.jini.mahalo;

import java.lang.reflect.Field;
import java.rmi.RemoteException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jgroup.core.ExternalGMIListener;
import jgroup.core.GroupManager;
import jgroup.core.MembershipListener;
import jgroup.core.View;
import jgroup.core.protocols.Anycast;
import jgroup.core.protocols.Atomic;
import jgroup.jini.txn.InternalGroupTransactionManager;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.transaction.CannotAbortException;
import net.jini.core.transaction.CannotCommitException;
import net.jini.core.transaction.CannotJoinException;
import net.jini.core.transaction.TimeoutExpiredException;
import net.jini.core.transaction.UnknownTransactionException;
import net.jini.core.transaction.server.CrashCountException;
import net.jini.core.transaction.server.TransactionManager;
import net.jini.core.transaction.server.TransactionParticipant;

import org.apache.log4j.Logger;

import com.sun.jini.mahalo.log.LogManager;
import com.sun.jini.start.LifeCycle;
import com.sun.jini.thread.TaskManager;
import com.sun.jini.thread.WakeupManager;

/**
 * Transient, non-persistent Transaction Manager with replication support.
 * Uses GMI Leadercast semantics to create transaction and transfer state 
 * to the replicas. GMI with total ordering is used in the transaction 
 * join and voting process.<p>
 * 
 * To start the Transaction Manager, it is preferred to use the  
 * {@link com.sun.jini.start.ServiceStarter} framework, which is included 
 * in the Jini distribution.
 * 
 * @author Rohnny Moland
 */
public class GroupTransientMahalo
  implements TransactionManager, MembershipListener, ExternalGMIListener, InternalGroupTransactionManager
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  private final static Logger log = Logger.getLogger(GroupTransientMahalo.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -1345915566687227879L;

  /** The mahalo object */
  private TxnManagerImpl                  tm;
  private InternalGroupTransactionManager igtm;
  
  /** Reflected fields from mahalo */
  private Map                             txns;
  private LogManager                      logmgr;
  private TaskManager                     taskpool;
  private WakeupManager                   taskWakeupMgr;

  /** Map holding new transactions which should be merged into txns when a state change occur */
  private Map<Long,TxnManagerTransaction> tmpTxns;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Construct a new instance of <code>TxnManagerImpl</code> and 
   * initialize Jgroup services and fields. 
   *
   * @param configArgs <code>String</code> array whose elements are
   *                   the arguments to use when creating the server.
   * @param lifeCycle  instance of <code>LifeCycle</code> that, if
   *                   non-<code>null</code>, will cause this object's
   *                   <code>unregister</code> method to be invoked during
   *                   shutdown to notify the service starter framework that
   *                   the reference to this service's implementation can be
   *                   'released' for garbage collection. A value of
   *                   <code>null</code> for this argument is allowed.
   *
   * @throws Exception If there was a problem initializing the service.
   */
  public GroupTransientMahalo(String[] configArgs, LifeCycle lifeCycle) 
    throws Exception
  {
    /** Obtains the group manager */
    GroupManager gm = GroupManager.getGroupManager(this);
    
    igtm = (InternalGroupTransactionManager) 
      gm.getService(InternalGroupTransactionManager.class);

    /** Create an instance of the mahalo transaction manager */
    tm = new TxnManagerImpl(configArgs, lifeCycle, false);

    /** Get the necessary fields from mahalo using reflection */
    this.txns = (Map) getField(tm, "txns");
    this.logmgr = (LogManager) getField(tm, "logmgr");
    this.taskpool = (TaskManager) getField(tm, "taskpool");
    this.taskWakeupMgr = (WakeupManager) getField(tm, "taskWakeupMgr");
    
    tmpTxns = Collections.synchronizedMap(new HashMap<Long, TxnManagerTransaction>());
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#viewChange(jgroup.core.View)
   */
  public void viewChange(View view)
  {
    if (log.isDebugEnabled())
      log.debug("view: " + view);
  }

  
  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange()
  {
  }


  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft()
  {
  }
 
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from TransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#create(long)
   */
//  @Leadercast
//  public TransactionManager.Created create(long lease)
//    throws LeaseDeniedException, RemoteException
//  {
//    TransactionManager.Created created = tm.create(lease);
//    if (log.isDebugEnabled()) 
//      log.debug("Created transaction with id: " + created.id);
//    
//    TxnManagerTransaction tmt = (TxnManagerTransaction) txns.get(new Long(created.id));
//    tmpTxns.put(new Long(created.id), tmt);
//    return created;
//  }

  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#create(long)
   */
  @Anycast
  public TransactionManager.Created create(long lease) 
    throws LeaseDeniedException, RemoteException
  {
    TransactionManager.Created created = tm.create(lease);
    if (log.isDebugEnabled())
      log.debug("Created transaction with id: " + created.id);

    TxnManagerTransaction tmt = (TxnManagerTransaction) txns.get(new Long(created.id));    
    try {
      igtm.internalCreate(tmt, created.id);
    } catch (RemoteException e) {
      log.debug(e.getMessage());
    }
    return created;
  }
  
  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#join(long, 
   * net.jini.core.transaction.server.TransactionParticipant, long)
   */
  @Atomic 
  public void join(long id, TransactionParticipant part, long crashCount) 
    throws UnknownTransactionException, CannotJoinException, CrashCountException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("Joining TM participant: " + part + " into transaction with ID=" + id);

    tm.join(id, part, crashCount);
  }


  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#getState(long)
   */
  @Anycast 
  public int getState(long id) throws UnknownTransactionException
  {
    int state = tm.getState(id);
    if (log.isDebugEnabled())
      log.debug("GroupTransientMahalo.getState() for txn id: " + id + " is " + state);
    return state;
  }

  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#commit(long)
   */
  @Atomic
  public void commit(long id) 
    throws UnknownTransactionException, CannotCommitException, RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("Started voting process for transaction with ID=" + id);
    tm.commit(id);
    
    if (log.isDebugEnabled())
      log.debug("GroupTransientMahalo.commit() finished with txn id: " + id);
  }

  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#commit(long, long)
   */
  @Atomic 
  public void commit(long id, long waitFor) 
    throws UnknownTransactionException, CannotCommitException, TimeoutExpiredException,
      RemoteException
  {
    tm.commit(id, waitFor);    
  }

  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#abort(long)
   */
  @Atomic 
  public void abort(long id) 
    throws UnknownTransactionException, CannotAbortException
  {
    if (log.isDebugEnabled())
      log.debug("Transaction " + id + " aborted.");
    tm.abort(id);
  }
  
  
  /* (non-Javadoc)
   * @see net.jini.core.transaction.server.TransactionManager#abort(long, long)
   */
  @Atomic 
  public void abort(long id, long waitFor) 
    throws UnknownTransactionException, CannotAbortException, TimeoutExpiredException
  {
    if (log.isDebugEnabled())
      log.debug("Transaction " + id + " aborted.");
    tm.abort(id, waitFor);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from StateListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.StateListener#getState()
   */
  public Object getState()
  {
    Object txnsState = tmpTxns;
//    tmpTxns.clear();
    /**
     * Need a way to clear transactions which
     * have already been transferred from leader
     * to the followers.
     */
    return txnsState;
  }

  
  /* (non-Javadoc)
   * @see jgroup.core.StateListener#putState(java.lang.Object)
   */
  @SuppressWarnings("unchecked")
  public void putState(Object state)
  {
    setField(tm, "txns", this.txns);    
    synchronized (txns) {
      Map<Long, TxnManagerTransaction> transactions = (Map<Long, TxnManagerTransaction>) state;
      for (Iterator<Long> iter = transactions.keySet().iterator(); iter.hasNext();) {
        Long tmpTxnsID = iter.next();        
        TxnManagerTransaction txnMgr = transactions.remove(tmpTxnsID);
        
        if (log.isDebugEnabled())
          log.debug("TM NEWSTATE: Merged in transaction with ID=" + tmpTxnsID);
        
        setField(txnMgr, "logmgr", this.logmgr);
        setField(txnMgr, "threadpool", this.taskpool);
        setField(txnMgr, "wm", this.taskWakeupMgr);
        setField(txnMgr, "settler", tm);
        setField(txnMgr, "leaseLock", new Object());
        setField(txnMgr, "jobLock", new Object());
        setField(txnMgr, "stateLock", new Object());
        txns.put(tmpTxnsID, txnMgr);
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InternalGroupTransactionManager
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.jini.txn.InternalGroupTransactionManager#internalCreate(java.lang.Object, long)
   */
  @SuppressWarnings("unchecked")
  public void internalCreate(Object txnMgr, long id) throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("internalCreate() - txn id: " + id);

    setField(tm, "txns", this.txns);
    TxnManagerTransaction mgrTransaction = (TxnManagerTransaction) txnMgr;

    setField(mgrTransaction, "logmgr", this.logmgr);
    setField(mgrTransaction, "threadpool", this.taskpool);
    setField(mgrTransaction, "wm", this.taskWakeupMgr);
    setField(mgrTransaction, "settler", tm);
    setField(mgrTransaction, "leaseLock", new Object());
    setField(mgrTransaction, "jobLock", new Object());
    setField(mgrTransaction, "stateLock", new Object());
   
    txns.put(new Long(id), mgrTransaction);    
  }


  /* (non-Javadoc)
   * @see jgroup.jini.txn.InternalGroupTransactionManager#internalCommit(long)
   */
  public void internalCommit(long id) 
    throws UnknownTransactionException, CannotCommitException, RemoteException
  {
    // Well. This is not implemented at the moment.
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Instance methods
  ////////////////////////////////////////////////////////////////////////////////////////////
    
  /**
   * Returns value of a field using reflection.
   * 
   * @param obj Object that contains the field to be reflected
   * @param name Name of field to be reflected
   * @return Value of the reflected field  
   */
  private Object getField(Object obj, String name) 
  { 
    Class c = obj.getClass();    
    final Field field;
    Object tmp = null;
       
    try {
      field = c.getDeclaredField(name);      
      AccessController.doPrivileged(new PrivilegedAction<Object>()
      {
        public Object run() {
          field.setAccessible(true);
          return null;
        }
      });
      
      tmp = field.get(obj);
    } catch (SecurityException e) {
      log.error(e.getMessage());
    } catch (NoSuchFieldException e) {
      log.error(e.getMessage());
    } catch (IllegalArgumentException e) {
      log.error(e.getMessage());
    } catch (IllegalAccessException e) {
      log.error(e.getMessage());
    }
    return tmp;
  }
  
  
  /**
   * Set new value of a field using reflection.
   * 
   * @param obj Object that contains the field to be reflected
   * @param name Name of field to be reflected
   * @param value New value of reflected field
   */
  private void setField(Object obj, String name, Object value)
  {
    Class c = obj.getClass();   
    final Field field;
       
    try {
      field = c.getDeclaredField(name);
      AccessController.doPrivileged(new PrivilegedAction<Object>()
      {
        public Object run() {
          field.setAccessible(true);
          return null;
        }
      });
      
      field.set(obj, value);
    } catch (SecurityException e) {
      log.error(e.getMessage());
    } catch (NoSuchFieldException e) {
      log.error(e.getMessage());
    } catch (IllegalArgumentException e) {
      log.error(e.getMessage());
    } catch (IllegalAccessException e) {
      log.error(e.getMessage());
    } 
  }


  public String toString()
  {
    return "GroupTransientMahalo";
  }


  public int hashCode() 
  {
    // FIXME..
    return tmpTxns.hashCode();
  }

} // END GroupTransientMahalo
