/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.arm;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jgroup.arm.GroupData.GroupMergeRecord;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.arm.GroupExistsException;
import jgroup.core.arm.UnknownGroupException;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.types.ViewImpl;

import org.apache.log4j.Logger;


/**
 *  The <code>GroupTable</code> class keeps track of object groups
 *  installed by the replication manager, and the application
 *  information associated with these groups.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class GroupTable
  implements Iterable<GroupData>
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupTable.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields (not shared with other RM replicas)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Table of all applications in the system (group id -> GroupData) */
  private final Map<Integer,GroupData> apptable =
    Collections.synchronizedMap(new HashMap<Integer,GroupData>());

  /** Field indicating if this RM replica is the leader or not */
  private boolean iamLeader;

  /** The period to wait for a view change */
  private int awaitViewPeriod;

  /** The most recently installed view of the RM group */
  private View view;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * The <code>GroupTable</code> constructor will initialize the
   * database maintained by the replication manager over application
   * groups installed/deployed through the replication manager.
   * 
   * @param awaitViewPeriod
   *   The delay to wait for a view before triggering the view
   *   monitor timer.
   * @param gid
   *   The group identifier for the application controling and
   *   interacting with this group table, typically this will be
   *   the identifier of the replication manager group.
   */
  public GroupTable(int awaitViewPeriod, int gid)
  {
    this.awaitViewPeriod = awaitViewPeriod;
    /* Initialize the view and leader fields */
    viewChange(new ViewImpl(gid), true);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // GroupTable methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add the given application group to the table, and set the
   *  assigned hosts set.
   */
  public void add(AppConfig app, HostSet hosts)
    throws GroupExistsException
  {
    if (view == null)
      throw new IllegalStateException("The view has not been initialized");
    int groupId = app.getGroupId();
    if (apptable.containsKey(groupId))
      throw new GroupExistsException("Cannot add application", groupId);
    // This will create a new GroupData object since it didn't exist
    GroupData gd = get(groupId);
    gd.addHosts(hosts);
  }

  /**
   *  Add the given host to an existing application group in the table.
   */
  public void addHost(AppConfig app, Host newHost)
    throws UnknownGroupException
  {
    GroupData gd = getGroupData(app);
    gd.addHost(newHost);
  }

  /**
   *  Remove the given host from an existing application group in the table.
   */
  public void removeHost(AppConfig app, Host oldHost)
    throws UnknownGroupException
  {
    GroupData gd = getGroupData(app);
    gd.removeHost(oldHost);
  }

  /**
   *  Remove the given application group from the system.
   */
  public void remove(AppConfig app)
  {
    GroupData gd = apptable.remove(app.getGroupId());
    if (gd != null)
      gd.cancelViewMonitor();
  }


  /**
   *  Returns true if the given group is contained in this
   *  grouptable object; otherwise false is returned.
   */
  public boolean contains(int gid)
  {
    return apptable.containsKey(gid);
  }


  /**
   *  Returns true if the given application is contained in this
   *  grouptable object; otherwise false is returned.
   */
  public boolean contains(AppConfig app)
  {
    return contains(app.getGroupId());
  }


  /**
   *  Returns the set of assigned hosts for the given application.
   */
  public HostSet getAssignedHosts(AppConfig app)
    throws UnknownGroupException
  {
    GroupData gd = apptable.get(app.getGroupId());
    if (gd == null)
      throw new UnknownGroupException(app.getGroupId());
    return gd.getAssignedHosts();
  }


  /**
   *  Returns an iterator over the application <code>GroupData</code>
   *  objects stored in the table.
   */
  public Iterator<GroupData> iterator()
  {
    return apptable.values().iterator();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for handling view changes and state merging
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Invoked by the replication manager upon a local view change
   * notification of the RM group.
   *
   * @param view
   *   The new view of the RM group
   * @param iamLeader
   *   True if the this member is the leader of the RM group.
   */
  public void viewChange(View view, boolean iamLeader)
  {
    this.view = view;
    this.iamLeader = iamLeader;
    for (GroupData gd : apptable.values()) {
      gd.viewChange(view, iamLeader);
    }
  }

  /**
   *  Invoked by the replication manager upon a view change notification
   *  to any of the groups it manages; i.e., kept in this grouptable.
   *  All replication manager replicas will invoke this method, that is
   *  correlation will be performed locally on all RM replicas, but
   *  recovery will only be performed by the leader RM replica. <p>
   *
   *  This method should never be invoked with a view of a group which
   *  is not already registered in this group table using the
   *  <code>add()</code> method.
   *
   *  @param view
   *    The view of a group for which the group table should be updated.
   *  @exception IllegalStateException
   *    Raised if the specified view's group does not correspond to any
   *    registered groups in the system.
   */
  public void viewChangeEvent(View view)
  {
    /* Obtain the group data object for the associated view */
    GroupData gd = get(view.getGid());
    gd.viewChangeEvent(view);
  }

  /**
   *  Returns a <CODE>MergeRecord</CODE> object containing the 
   *  information needed by the replication manager replicas contained in
   *  <CODE>dests</CODE> set to merge their databases.
   *  
   *  @param dests the set of replicas to which the merging information
   *    is addressed.
   *  @param me identifier of the local replica, needed to identify
   *    the origin of the data.
   */
  MergeRecord getState(MemberId[] dests, MemberId me)
  {
    MergeRecord merging = new MergeRecord(apptable.size());

    int i = 0;
    for (Iterator<GroupData> iter = iterator(); iter.hasNext(); i++) {
      GroupData gd = iter.next();
      merging.groupMerge[i] = gd.getMergeRecord(dests, me);
    }
    return merging;
  }

  /**
   *  Takes a <code>MergeRecord</code> object and inserts the data
   *  contained in it.
   *
   *  @param merging
   *    The merging data sent by another replica
   */
  void putState(MergeRecord merging)
  {
    for (int i = 0; i < merging.groupMerge.length; i++) {
      int gid = merging.groupMerge[i].getGroupId();
      GroupData gd = get(gid);
      gd.putMergeRecord(merging.groupMerge[i]);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the <code>GroupData</code> object associated with the 
   *  given <code>groupId</code>, or creates a new one if not present
   *  in this <code>GroupTable</code> instance.
   */
  private GroupData get(int groupId)
  {
    GroupData gd = apptable.get(groupId);
    if (gd == null) {
      if (log.isDebugEnabled()) {
        log.debug("Group: " + groupId + " not in apptable");
      }
      AppConfig app = AppConfig.getApplication(groupId);
      gd = new GroupData(app, awaitViewPeriod);
      gd.viewChange(view, iamLeader);
      apptable.put(groupId, gd);
    }
    return gd;
  }

  /**
   * Returns the <code>GroupData</code> object associated with the given
   * <code>app</code> object; if no such app has been registered an
   * <code>UnknownGroupException</code> is thrown.
   * 
   * @param app
   * @return
   * @throws UnknownGroupException
   */
  private GroupData getGroupData(AppConfig app)
    throws UnknownGroupException
  {
    int gid = app.getGroupId();
    GroupData gd = apptable.get(gid);
    if (gd == null)
      throw new UnknownGroupException(gid);
    return gd;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[GroupTable: size=");
    buf.append(apptable.size());
    buf.append(":\n");
    for (GroupData gd : apptable.values()) {
      buf.append(gd);
      buf.append("\n");
    }
    buf.append("]");
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class for state merge
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static class MergeRecord
    implements Serializable
  {

    private static final long serialVersionUID = -441736911665721212L;

    GroupMergeRecord[] groupMerge;

    MergeRecord(int size)
    {
      groupMerge = new GroupMergeRecord[size];
    }

  } // END MergeRecord

} // END GroupTable
