/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.arm.recovery;

import jgroup.core.arm.DistributionScheme;
import jgroup.core.arm.RecoveryStrategy;
import jgroup.relacs.config.AppConfig;

import org.apache.log4j.Logger;


/**
 *  This implements only the core initialization of a recovery strategy,
 *  including also handling for the RM recoveree replica.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public abstract class AbstractRecoveryStrategy
  implements RecoveryStrategy
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(AbstractRecoveryStrategy.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The distribution scheme interface */
  protected DistributionScheme distScheme;

  /** The application associated with this recovery strategy instance */
  protected AppConfig app;

  /** The current delta redundancy (the number of missing replicas) */
  protected int missing;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  All recovery strategy implementations should implement a
   *  no-argument constructor.  This pattern cannot be enforced by the
   *  RecoveryStratgy interface, but can be check while parsing.
   */
  public AbstractRecoveryStrategy() {}


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Helper methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the number of replicas that needs to be recovered to
   *  maintain the minimal redundancy level specified.  If no recovery
   *  is needed, zero is returned.
   */
  protected int deltaRedundancy(AppConfig app)
  {
    int delta = app.getMinimalRedundancy() - app.getCurrentGroupSize();
    if (log.isDebugEnabled())
      log.debug("delta=" + delta);
    return (delta > 0) ? delta : 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RecoveryStrategy interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /* (non-Javadoc)
   * @see jgroup.core.arm.RecoveryStrategy#initialize(jgroup.core.arm.DistributionScheme, jgroup.relacs.config.AppConfig)
   */
  public void initialize(DistributionScheme distScheme, AppConfig app)
  {
    this.distScheme = distScheme;
    this.app = app;
  }

  /* (non-Javadoc)
   * @see jgroup.core.arm.RecoveryStrategy#needsRecovery()
   */
  public boolean needsRecovery()
  {
    return (deltaRedundancy(app) > 0);
  }

  /* (non-Javadoc)
   * @see jgroup.core.arm.RecoveryStrategy#prepareRecovery()
   */
  public boolean prepareRecovery()
  {
    missing = deltaRedundancy(app);
    if (log.isDebugEnabled())
      log.debug("Missing " + missing + " replicas for: " + app);
    return (app.getCurrentGroupSize() == 0);
  }

} // END AbstractRecoveryStrategy
