/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This class is an helper utility that provides application developers
 * with additional information about members of their current view. It
 * can be used as a hash table, to associate application-dependent
 * information to member identifier. Furthermore, it can be used to
 * obtain information about the state of a member identifier with
 * respect to both the current view and previous installed views (for
 * example, method <code>getStatus</code> may be used to discover which
 * members survived from the previous view). The table also maintains
 * information about the relative position of a member identifier in the
 * member identifier array associated to the last installed view. <p>
 *
 * The table maintains information about a member until it crashes.  A
 * member is declared crashed when a member identifier from the same
 * endpoint, but with an higher incarnation identifier is inserted in
 * the table.  This means that the JVM hosting the member has crashed,
 * and a new one has started.  The state of crashed members is set to
 * <code>CRASHED</code>.  The information associated to the member is
 * maintained until the next view change, after which the member
 * identifier and all its associated information is removed from the
 * table. <p>
 *
 * This class implements the <code>MembershipListener</code> interface,
 * so it may be passed to the <code>MembershipService</code> to be
 * notified whenever a view change occurs. Otherwise, the application
 * developer may take care of updating the table by explicitly calling
 * method <code>viewChange</code>.
 *
 * @author Alberto Montresor
 * @author Hein Meling
 * @since Jgroup 0.9
 */
public class MemberTable
  implements MembershipListener, Externalizable
{

  private static final long serialVersionUID = -5036038300592577228L;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  State assigned to member identifiers not included in the table
   */
  public final static int NOTMEMBER   = 0;

  /**
   *  State assigned to partitioned members for which no information is
   *  available.
   */
  public final static int PARTITIONED = 1;

  /**
   *  State assigned to members that have been declared crashed because
   *  a member identifier from the same endpoint but with a new
   *  incarnation number has been inserted in the table.
   */
  public final static int CRASHED     = 2;

  /**
   *  State assigned to member identifiers whose incarnation number is
   *  greater than the incarnation number of other members previously
   *  contained in the table.
   */
  public final static int RECOVERING  = 3;

  /**
   *  State assigned to members that survived from the previous view
   */
  public final static int SURVIVED    = 4;

  /**
   *  State assigned to members that merged the view after a previous
   *  partitioning
   */
  public final static int MERGING     = 5;

  /**
   *  State assigned to new members
   */
  public final static int NEWMEMBER   = 6;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Mapping from MemberId to its corresponding status record */
  private Map<MemberId, Record> table = new HashMap<MemberId, Record>(33);

  /** Mapping from IP address to list of members */
  private Map<EndPoint, List<Record>> iptable = new HashMap<EndPoint, List<Record>>(33);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Build a new empty <code>MemberTable</code>.  Also used as the
   *  default constructor for externalization.
   */
  public MemberTable()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Returns the status associated with this member identifier. The
   *  returned value is one of the integer constants defined in this
   *  class.
   */
  synchronized public int getStatus(MemberId id)
  {
    Record r = table.get(id);
    if (r != null)
      return r.status;
    else
      return NOTMEMBER;
  }


  /**
   *  Returns the position of the specified member identifier in the
   *  last view for which method <code>viewChange</code> has been
   *  called; return -1 if the member is not member of that view.
   */
  synchronized public int getIndex(MemberId id)
  {
    Record r = table.get(id);
    if (r != null)
      return r.pos;
    else
      return -1;
  }


  /**
   *  Associates object <code>value</code> to the specified member
   *  identifier.  This association is maintained in the table until it
   *  is explicitly removed from the table, or the member is declared
   *  crashed.  Values may be retrieved using method <code>get</code>.
   *
   *  @param id
   *    The member identifer to which the <code>value</code> should be
   *    associated.
   *  @param value
   *    The value to associate with the given member.
   *  @throws IllegalArgumentException
   *    If the specified member identifier does not exist within this
   *    member table.
   */
  synchronized public void put(MemberId id, Object value)
  {
    Record r = table.get(id);
    if (r != null)
      r.value = value;
    else 
      throw new IllegalArgumentException("The specified member identifer does not exist in this member table.");
  }

  /**
   *  Returns the value associated by this <code>MemberTable</code> to
   *  the specified member id.  Returns null if the map contains no
   *  mapping for this member id.  A return value of null does not
   *  necessarily indicate that the map contains no mapping for the key;
   *  it is possible that the map explicitly maps the key to null.
   *
   *  @param id 
   *    The member identifier used as key.
   *  @return 
   *    The associated value, or null if no such member exists within
   *    this <code>MemberTable</code>.
   */
  synchronized public Object get(MemberId id)
  {
    Record r = table.get(id);
    if (r != null)
      return r.value;
    else
      return null;
  }


  /**
   *
   */
  synchronized public void remove(MemberId id)
  {
    Record r = table.get(id);
    if (r != null)
      table.remove(id);
    List<Record> recordList = iptable.get(id.getEndPoint());
    if (recordList != null) {
      for (Iterator<Record> iter = recordList.iterator(); iter.hasNext();) {
        Record record = iter.next();
        if (record.id.equals(id)) {
          iter.remove();
          break;
        }
      }
    }
  }


  /**
   *  Returns an array containing the values associated to all members
   *  contained in this <code>MemberTable</code>, including members
   *  crashed or partitioned.  The array is sorted according to the member
   *  index in the most recent view; the crashed and partitioned members
   *  are placed at the end of the array with no particular order.
   */
  synchronized public Object[] elements()
  {
    Object[] elements = new Object[table.size()];
    int max = 0;
    List<Object> failed = new ArrayList<Object>();
    for (Record r : table.values()) {
      if (r.pos != -1) {
        elements[r.pos] = r.value;
        if (r.pos > max)
          max = r.pos;
      } else {
        failed.add(r.value);
      }
    }
    for (Iterator iter = failed.iterator(); iter.hasNext();) {
      elements[++max] = iter.next();
    }
    return elements;
  }


  /**
   *  Returns an array containing the identifiers of all members
   *  contained in this <code>MemberTable</code>, including members
   *  crashed or partitioned.
   */
  synchronized public MemberId[] members()
  {
    MemberId[] elements = new MemberId[table.size()];
    int i = 0;
    for (Record record : table.values()) {
      elements[i++] = record.id;
    }
    return elements;
  }


  /**
   * Returns true if the specified member identifier is
   * member of the last view for which method <code>viewChange</code>
   * has been called; return false otherwise.
   *
   * @param id the member whose membership is to be tested.
   */
  synchronized public boolean isMember(MemberId id)
  {
    Record r = table.get(id);
    if (r != null)
      return (r.status != CRASHED && r.status != PARTITIONED);
    else
      return false;
  }


  public String toString(MemberId id)
  {
    StringBuilder buf = new StringBuilder();
    switch (getStatus(id)) {
    case NOTMEMBER:
      /* This member is a new member of this member table */
      buf.append("NOTMEMBER: ");
      break;

    case NEWMEMBER:
      /* This is a new member */
      buf.append("NEWMEMBER: ");
      break;

    case RECOVERING:
      /* This member has crashed and then recovered (new incarnation)*/
      buf.append("RECOVERING: ");
      break;

    case SURVIVED:
      /* This member has survived from the previous view */
      buf.append("SURVIVING: ");
      break;

    case PARTITIONED:
      /* This member has partitioned since the previous view */
      buf.append("PARTITIONED: ");
      break;

    case MERGING:
      /* This member has merged since the previous view */
      buf.append("MERGING: ");
      break;

    case CRASHED:
      /* This member has merged since the previous view */
      buf.append("CRASHED: ");
      break;

    default:
      buf.append("UNDEFINED STATUS CODE: ");
      break;
    }
    buf.append(id);
    return buf.toString();
  }


  synchronized public String toString()
  {
    StringBuilder buf = new StringBuilder();
    for (Record rec : table.values()) {
      buf.append(toString(rec.id));
      buf.append("\n");
    }
    return buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Update the information in the table with the information
   *  maintained in the new view.
   */
  synchronized public void viewChange(View view)
  {
    Record    r;
    List<Record> recordList;

    // Remove crashed elements from table
    for (Iterator<Record> iter = table.values().iterator(); iter.hasNext();) {
      r = iter.next();
      r.pos = -1;
      r.prev = r.status;
      r.status = PARTITIONED;
      if (r.prev == CRASHED)
        iter.remove();
    }

    // Update the status of each member
    MemberId[] members = view.getMembers();
    for (int i=0; i < members.length; i++) {
      r = table.get(members[i]);
      if (r != null) {

        /*
         * The member id is already present in the table; if
         * the member was partitioned, now its status is MERGING.
         * In all other cases (NEWMEMBER, RECOVERING, SURVIVED, MERGING),
         * the new status is SURVIVED
         */
        //
        if (r.prev == PARTITIONED)
          r.status = MERGING;
        else
          r.status = SURVIVED;

      } else {

        /*
         * The member id is not present in the table; may be a new
         * member or a recovered member.  For each endpoint identifier,
         * we maintain a list of member identifiers associated with that
         * endpoint.  This is useful to discover if this is just a new
         * member, or it is a recovering member from a previous JVM
         * (endpoint on the same port).
         */
        recordList = iptable.get(members[i].getEndPoint());
        if (recordList == null) {

          /*
           * First member from that endpoint; new member for sure
           */
          recordList = new ArrayList<Record>();
          iptable.put(members[i].getEndPoint(), recordList);
          r = new Record(NEWMEMBER, members[i]);
          recordList.add(r);
          table.put(members[i], r);

        } else {
          r = recordList.get(0);

          if (members[i].isNewer(r.id)) {
            /*
             * New incarnation! Move old member ids to crashed status;
             * Create a new empty record and put it in the tables.
             */
            for (int k=0; k < recordList.size(); k++) {
              r = recordList.get(k);
              r.status = CRASHED;
              r.pos = -1;
            }
            recordList.clear();
            r = new Record(RECOVERING, members[i]);
            recordList.add(r);
            table.put(members[i], r);

          } else {
            /*
             *  This member id has the same incarnation number of those
             *  contained in the arraylist, but its member identifier
             *  is not present in the table. If members from the same
             *  host are recovering, so is this. Otherwise, it is a
             *  new member.
             */
            if (r.status == RECOVERING)
              r = new Record(RECOVERING, members[i]);
            else
              r = new Record(NEWMEMBER, members[i]);
            recordList.add(r);
            table.put(members[i], r);
          }
        }
      }
      r.pos = i;
    }
  }


  /**
   *  No effect.
   */
  public void hasLeft()
  {
  }


  /**
   *  No effect.
   */
  public void prepareChange()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Internal data structure
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Data structure used to maintain an element in the hashtable and its
   *  status.
   */
  private final class Record
    implements Serializable
  {

    private static final long serialVersionUID = -710112890009914664L;

    MemberId  id;         // MemberId to which this value has been
    int       pos;        // Position of the member in the view
    int       status;     // State (RECOVERED, SURVIVED, MERGED, NEWMEMBER, CRASHED)
    int       prev;       // Previous status
    Object    value;      // The object stored by users of this data structure

    Record(int status, MemberId id) {
      this.id     = id;
      this.status = status;
      this.pos    = -1;
      this.prev   = status;
      this.value  = null;
    }

  } // END Record


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the Externalizable interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * Note that the methods below are synchronized to avoid concurrent
   * access to the internal tables of this membertable object.
   */

  /* (non-Javadoc)
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  synchronized public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeObject(table);
    out.writeObject(iptable);
  }

  /* (non-Javadoc)
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  @SuppressWarnings("unchecked")
  synchronized public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    table = (Map) in.readObject();
    iptable = (Map) in.readObject();
  }

} // END MemberTable
