/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.arm;

import jgroup.core.MergingListener;
import jgroup.core.View;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;


/**
 *  The <code>DistributionScheme</code> interface provides a standard
 *  for implementing various distribution schemes that simplify
 *  interaction with the <code>ReplicationManager</code>.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public interface DistributionScheme
  extends MergingListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods related to replica distribution
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Assign the host location for the specified application.  The
   *  <code>AppConfig</code> object also contains the required redundancy.
   *
   *  @param app
   *    An <code>AppConfig</code> object specifying the class parameters
   *    and redundancy requirements, and group identifier.
   *  @return 
   *    A <code>HostSet</code> specifying the locations where the replicas
   *    should be created.
   *  @exception RedundancyException 
   *    Is raised if the specified redundancy could not be satisfied.
   *  @exception GroupExistsException 
   *    Raised if the specified application group already exists (has been
   *    assigned to some hostset).
   */
  public HostSet assignReplicas(AppConfig app)
    throws RedundancyException, GroupExistsException;


  /**
   *  Remove replica assignments for the specified application.
   *
   *  @param app
   *    An <code>AppConfig</code> object specifying the class parameters
   *    and redundancy requirements, and group identifier.
   *  @return 
   *    A <code>HostSet</code> specifying the locations where the replicas
   *    was located prior to the remove.
   */
  public HostSet removeReplicas(AppConfig app)
    throws UnknownGroupException;


  public HostSet collocateReplicas(AppConfig app, AppConfig collocateWithApp)
    throws GroupExistsException, RedundancyException, UnknownGroupException;


  /**
   *  Reassign the application replica running on the specified host.
   *
   *  @param app
   *    The application group to which the replica belongs.
   *  @param host
   *    The host from which to remove the replica.
   *  @return 
   *    The host on which a new replica should be created.
   *  @exception UnknownGroupException
   *    The given application group is unknown.
   *  @exception RedundancyException
   *    Raised if the system could not reassign the replica to any host
   *    in the system.
   */
  public Host reassignReplica(AppConfig app, Host host)
    throws UnknownGroupException, RedundancyException;


  /**
   *  Update the distribution scheme according to the specified view
   *  for the given application.
   *  <P>
   *  Note that the method behavior should be idempotent.
   */
  public void viewChangeEvent(View view);

} // END DistributionScheme
