/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.multicast;

import java.io.IOException;
import java.io.OutputStream;

import jgroup.core.JgroupException;
import jgroup.core.Layer;


/**
 *  The facilities provided by the reliable multicast service included
 *  in Jgroup may be accessed using the <code>MulticastService</code>
 *  interface.  Methods in this interfaces enable members of a group to
 *  communicate among themselves by multicasting messages or complex
 *  objects. <p>
 *
 *  Instances of <code>MulticastService</code> may be obtained through
 *  the <code>GroupManager</code> class.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.8
 */
public interface MulticastService
  extends Layer
{

  /**
   *  Multicast a message enclosed in the stream <code>out</code> to the
   *  members of the current view of the group. <p>
   *
   *  The invoker may specify an <code>AckListener</code> object, which
   *  will be notified when objects returned by members delivering this
   *  message, are received.
   *
   *  @param out
   *    The stream message to be multicast.
   *  @param ackl
   *    Acknowledgement listener to be notified; if null, no listener
   *    will be notified.
   *  @param chainId
   *    Identifier of a previous message (delivered through a 
   *    {@link jgroup.core.multicast.MulticastListener#deliverStream(java.io.InputStream, jgroup.core.MemberId, int) deliverStream}).
   *    This is only required for multicast messages
   *    that relate to some previous multicast message.  It will only
   *    be used in extremely rare cases, to avoid that the membership service
   *    delays the message while running the agreement protocol, which would
   *    otherwise prevent the agreement protocol from concluding.
   *
   *  @exception UnsupportedOperationException
   *    Raised if the member associated with this group manager is not a
   *    <code>MulticastListener</code>, and thus is unable to receive messages.
   *  @exception JgroupException
   *    Rasied if the group manager receives a request to multicast a
   *    message after having joined a group, but before having received
   *    the first view for it; or after having requested to leave the group.
   */
  public void mcast(OutputStream out, AckListener ackl, ChainIdentifier chainId)
    throws JgroupException;

  /**
   *  Multicast a message enclosed in the stream <code>out</code> to the
   *  members of the current view of the group. <p>
   *
   *  The invoker may specify an <code>AckListener</code> object, which
   *  will be notified when objects returned by members delivering this
   *  message, are received.
   *
   *  @param out
   *    The stream message to be multicast.
   *  @param ackl
   *    Acknowledgement listener to be notified; if null, no listener
   *    will be notified.
   *
   *  @exception UnsupportedOperationException
   *    Raised if the member associated with this group manager is not a
   *    <code>MulticastListener</code>, and thus is unable to receive messages.
   *  @exception JgroupException
   *    Rasied if the group manager receives a request to multicast a
   *    message after having joined a group, but before having received
   *    the first view for it; or after having requested to leave the group.
   */
  public void mcast(OutputStream out, AckListener ackl)
    throws JgroupException;


  /**
   *  Multicast a serializable object to the members of the current view
   *  of a group. <p>
   *
   *  The invoker may specify an <code>AckListener</code> object, which
   *  will be notified when objects returned by members delivering this
   *  message are received.
   *
   *  @param protocol
   *    The protocol for the message object to be multicast.
   *  @param obj
   *    The message object to be multicast.
   *  @param ackl
   *    Acknowledgement listener to be notified; if null, no listener
   *    will be notified.
   *
   *  @exception UnsupportedOperationException
   *    Raised if the member associated with this group manager is not a
   *    <code>MulticastListener</code>, and thus is unable to receive messages.
   *  @exception JgroupException
   *    Rasied if the group manager receives a request to multicast a
   *    message after having joined a group, but before having received
   *    the first view for it; or after having requested to leave the group.
   *  @exception IOException
   *    Raised if the object to be multicast could not be serialized.
   */
  public void mcast(String protocol, Object obj, AckListener ackl)
    throws JgroupException, IOException;


  /**
   *  Returns an output stream to be used to encode a message.  Method
   *  <code>mcast(OutputStream, AckListener)</code> accepts only
   *  multicast messages obtained through this method.  The provided output
   *  stream is a memory data structure optimized to avoid useless data
   *  copies.
   *
   *  @param protocol
   *     The protocol name used for the <code>OutputStream</code> message
   *     object to be multicast.
   */
  public OutputStream getMessage(String protocol)
    throws IOException;

} // END MulticastService
