/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.core.registry;

import java.rmi.RemoteException;
import java.rmi.server.*;

import jgroup.core.JgroupException;


/**
 *  This interface is the service provider interface for the registry
 *  location service.  It is used by Jgroup to identify classes that are
 *  able to retrieve proxies for registry service implementations and
 *  create registry replicas.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.8
 */
public interface RegistryLocator
{

  /**
   *  Returns a proxy for the dependable registry service.
   *
   *  @return 
   *    A proxy for the dependable registry service.
   *
   *  @exception RemoteException 
   *    Raised if the proxy could not be obtained.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry getRegistry()
    throws JgroupException, RemoteException;


  /**
   *  Returns a proxy for the dependable registry service.
   *
   *  Communication with this remote registry will use the supplied
   *  <code>RMIClientSocketFactory</code> to create socket connections
   *  to the dependable registry on the distributed system and port.
   *
   *  @param csf
   *    Client-side socket factory used to make connections to the
   *    dependable registry. If <code>csf</code> is null, then the
   *    default client-side socket factory will be used.
   *
   *  @return 
   *    A proxy for the dependable registry service.
   *
   *  @exception RemoteException 
   *    Raised if the proxy could not be obtained.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry getRegistry(RMIClientSocketFactory csf)
    throws JgroupException, RemoteException;


  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry createRegistry()
    throws JgroupException, RemoteException;


  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  The new instance of the dependable registry listens to incoming
   *  requests using a <code>ServerSocket</code> created from the
   *  supplied <code>RMIServerSocketFactory</code>.  A client that
   *  receives a reference to this registry will then use a
   *  <code>Socket</code> created from the supplied
   *  <code>RMIClientSocketFactory</code>.
   *
   *  @param csf
   *    Client-side <code>Socket</code> factory used to make connections
   *    to the registry.
   *  @param ssf
   *    Server-side <code>ServerSocket</code> factory used to accept
   *    connections to the registry.
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry createRegistry(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws JgroupException, RemoteException;

  
  /**
   * Returns a LookupRegistry interface, which can be used
   * by clients to lookup a remote reference from registry.
   * 
   * @return 
   *   LookupRegistry with methods to lookup remote ref from registry.
   */
  public LookupRegistry getLookupRegistry()
    throws JgroupException, RemoteException;

} // END RegistryLocator
