/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import jgroup.core.ConfigurationException;
import jgroup.util.Abort;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 *  Parser code to read an XML configuration file, using the supplied URL.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class ConfigParser
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(ConfigParser.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants - XML tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String CONFIG_ROOT_TAG_NAME       	= "Configuration";
  private static final String SERVICES_ROOT_TAG_NAME     	= "Services";
  private static final String APPLICATIONS_ROOT_TAG_NAME 	= "Applications";
  private static final String CLIENT_CONFIG_ROOT_TAG_NAME	= "ClientConfiguration";
  private static final String EXPERIMENT_ROOT_TAG_NAME 		= "ExperimentConfiguration";
  private static final String VERSION_ATTRIB                = "version";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Map containing information about configuration classes.  It is a
   *  mapping from configuration class literal to an Object implementing
   *  that class, and containing configuration information.
   */
  private static Map<Class,ConfigurationObject> configMap =
    new HashMap<Class,ConfigurationObject>();

  /** Used to avoid duplicate parsing. */
  private static Set<String> parsedFiles = new HashSet<String>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * The same as <code>parse(urlConfig, false)</code>.
   */
  public static void parse(String urlConfig)
    throws ConfigurationException
  {
    parse(urlConfig, false);
  }


  /**
   *  Parses the file obtained from the specified URL and inserts in an
   *  internal map, associations from class literals of configuration
   *  objects to configuration object instances.
   *    
   *  @param urlConfig  the URL to the XML configuration file containing
   *    the description of the distributed system.
   *  @param parseAgain  If false, all config-files will be parsed only once.
   */
  public static void parse(String urlConfig, boolean parseAgain)
    throws ConfigurationException
  {
    /* To avoid duplicate parsing, but only if parseAgain is false */
    if (parsedFiles.contains(urlConfig) && !parseAgain)
      return;
    parsedFiles.add(urlConfig);

    Document doc = null;
    try {
      DocumentBuilderFactory docBF = DocumentBuilderFactory.newInstance();
      boolean validate = false;
      String validateStr = System.getProperty("jgroup.config.validate");
      if (validateStr != null && validateStr.equals("true")) {
        validate = true;
      }
      docBF.setValidating(validate);
      DocumentBuilder db = docBF.newDocumentBuilder();
      db.setErrorHandler(new ParsingErrorHandler());
      doc = db.parse(urlConfig);
    } catch (Exception exception) {
      log.warn("Parser exception while parsing configuration file", exception);
      throw new ConfigurationException("Unable to parse configuration file: " + urlConfig);
    }

    /* Get and check the root element */
    Element elm = doc.getDocumentElement();
    if (!checkVersion(elm, CONFIG_ROOT_TAG_NAME, "1.1") &&
        !checkVersion(elm, CLIENT_CONFIG_ROOT_TAG_NAME, "1.0") &&
        !checkVersion(elm, SERVICES_ROOT_TAG_NAME, "1.0") &&
        !checkVersion(elm, APPLICATIONS_ROOT_TAG_NAME, "1.1") &&
        !checkVersion(elm, EXPERIMENT_ROOT_TAG_NAME, "0.1" ))
      throw new ConfigurationException("Unexpected root tag in " + urlConfig);

    /* Builds a map with the configuration objects */
    NodeList nl = elm.getChildNodes();
    for (int i = 1; i <= nl.getLength(); i++) {
      Node node = nl.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {

        /* Obtain a configuration object for the node */
        ConfigurationObject obj = parseTag((Element) node);

        /* Insert the configuration object in the map */
        configMap.put(obj.getClass(), obj);
      }
    }
  }


  /**
   *  Retrieve an instance of the given configuration class, that is
   *  containing the configuration information.  For example, the
   *  configuration data object associated with the XML tag
   *  <code>Transport</code>, can be retreived using the class literal
   *  <code>TransportConfig.class</code>.
   *
   *  @param cl
   *    The class literal of the configuration object to retreive.
   *  @return 
   *    An instance of the configuration object retreived.
   *
   *  @exception NullPointerException
   *    If there is no such configuration object in the internal
   *    configuration map.  This may occur for two reasons; (i) there
   *    is no such configuration tag specified in the configuration file,
   *    or (ii) the specified class literal is a non-existant class.
   *  @exception IllegalStateException
   *    If the parser was not invoked prior to invoking this method.
   */
  public static Object getConfig(Class cl)
  {
    if (configMap.isEmpty())
      throw new IllegalStateException("getConfig() invoked before parsing the config files");
    Object obj = configMap.get(cl);
    if (obj == null)
      throw new NullPointerException("No configuration data available for " + cl);
    return obj;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods for parsing the various tags in the XML tree
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses a child of the root element and returns a configuration
   *  object for it.  Given the tag name, a class in the
   *  <code>jgroup.relacs.config</code> is used to parse the element and
   *  to store its configuration data.  The name of this class is
   *  obtained by adding the suffix Config to the class name.  All these
   *  classes must extend the ConfigurationObject interface.
   *  
   *  @param elm 
   *    the element to parse
   *  @exception ConfigurationException 
   *    if the tag given by the <code>elm</code> object could not be parsed.
   */
  private static ConfigurationObject parseTag(Element elm)
    throws ConfigurationException
  {
    try {
      Class cl = Class.forName("jgroup.relacs.config." + elm.getTagName() + "Config");
      ConfigurationObject config = (ConfigurationObject) cl.newInstance();
      config.parse(elm);
      return config;
    } catch (ClassCastException e) {
      e.printStackTrace();
      throw new ConfigurationException("Unexpected tag " + elm.getTagName());
    } catch (ConfigurationException e) {
      throw e;
    } catch (Exception e) {
      e.printStackTrace();
      throw new ConfigurationException("Error parsing configuration file: " + e.getMessage());
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods for parsing the XML tree
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified <code>Element</code> for sub-elements.
   *
   *  @param elm
   *    The <code>Element</code> object to parse for sub-elements.
   *  @return 
   *    An <code>Element[]</code> containing the sub-elements.
   */
  static Element[] parseList(Element elm)
  {
    NodeList tagList = elm.getChildNodes();
    ArrayList<Node> elmList = new ArrayList<Node>(tagList.getLength());

    /* Parse the list of tags */
    for (int i = 1; i <= tagList.getLength(); i++) {
      Node node = tagList.item(i);
      if (node != null && !node.getNodeName().startsWith("#"))
        elmList.add(node);
    }
    elmList.trimToSize();
    return (Element[]) elmList.toArray(new Element[elmList.size()]);
  }


  /**
   *  Check that the configuration file has the expected
   *  version number and root tag.
   *
   *  @param elm
   *    The root element to check.
   *  @param rootTagName
   *    The expected root tag name to compare the XML root element with.
   *  @return 
   *    True if the root tag and version number was correct, and we
   *    can continue parsing the rest of the XML document.  False is
   *    returned if the root tag is incorrect.
   *  @throws ConfigurationException
   *    Raised if the version number is incorrect.
   */
  static boolean checkVersion(Element elm, String rootTagName, String versionValue)
    throws ConfigurationException
  {
    String cfgRoot = elm.getTagName();
    String cfgVersion = elm.getAttribute(VERSION_ATTRIB);
    if (cfgRoot.equalsIgnoreCase(rootTagName))
      if (cfgVersion.equals(versionValue))
        return true;
      else
        throw new ConfigurationException("Wrong version (" + cfgVersion + ") for "
          + rootTagName + " configuration file. Expected version: " + versionValue);
    else
      return false;
  }


  /**
   *  Returns true if the given element in the XML DOM tree corresponds
   *  to the given tag string.
   *
   *  @param elm  An element in the DOM tree.
   *  @param tag  A tag string to match against the DOM tree element.
   *  @return true if the given element in the XML DOM tree corresponds
   *  to the given tag string.
   */
  static boolean checkTag(Element elm, String tag)
  {
    return elm.getTagName().equalsIgnoreCase(tag);
  }


  /**
   *  Get the content of a single attribute within a tag.
   *
   *  @param elm  an element in the DOM tree.
   *  @param attrib  attribute within the given tag.
   *  @param required  true if attribute is required; false otherwise.
   *  @return  value of the attribute found.
   *  @exception ConfigurationException if the given attribute is required and its
   *    undefined.
   */
  static String getAttrib(Element elm, String attrib, boolean required)
    throws ConfigurationException
  {
    /* Get the specified attribute's value; note that we are case-sensitive. */
    String value = elm.getAttribute(attrib);

    /*
     * REQUIRED vs IMPLIED check; if attribute is required and the
     * 'value' is zero length we throw an ConfigurationException.
     */
    if (required && value.length() == 0)
      throw new ConfigurationException("Expected required attribute: " + attrib);

    return value;
  }


  /**
   *  Returns the value of a single attribute within a tag as an integer.
   *  Note that by definition, we also allow the string "ALL" to be represented
   *  in an integer type attribute.  The value "ALL" will return a -1 from this
   *  method.
   *
   *  @param elm  an element in the DOM tree.
   *  @param attrib  attribute within the given tag.
   *  @param required  true if attribute is required; false otherwise.
   *  @return  integer value of the attribute found.
   *  @exception ConfigurationException if the given attribute is required and its
   *    undefined.
   */
  static int getIntAttrib(Element elm, String attrib, boolean required)
    throws ConfigurationException
  {
    String value = getAttrib(elm, attrib, required);
    if ("".equals(value)) {
      throw new ConfigurationException("Attribute " + attrib + " is not defined");
    } else if ("ALL".equalsIgnoreCase(value)) {
      return -1;
    }
    try {
      return Integer.parseInt(value);
    } catch(NumberFormatException e) {
      throw new ConfigurationException("Illegal value for attribute " + 
        attrib+ ": "  + elm.getAttribute(attrib));
    }
  }


  public static boolean getBooleanAttrib(Element elm, String attrib, boolean required)
    throws ConfigurationException
  {
    String strval = getAttrib(elm, attrib, required);
    if (strval == null || strval.length() == 0)
      return false;
    return Boolean.valueOf(strval).booleanValue();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class for handling parser errors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Class <code>ParsingErrorHandler</code> provides methods that will
   *  be called upon parsing errors. This implementation just prints the
   *  content of the error messages, and exits for serious errors.
   *
   *  @author Hein Meling
   */
  static class ParsingErrorHandler
    implements ErrorHandler
  {

    public void warning(SAXParseException ex)
    {
      System.err.println("Warning: " + ex.getMessage());
    }

    public void error(SAXParseException ex)
    {
      dumpContent();
      Abort.exit("Error: " + ex.getMessage(), 1);
    }

    public void fatalError(SAXParseException ex)
      throws SAXException
    {
      dumpContent();
      Abort.exit("Fatal error: " + ex.getMessage(), 2);
    }

    private void dumpContent()
    {
      System.err.println("Parsed files:");
      for (Iterator iter = parsedFiles.iterator(); iter.hasNext(); ) 
        System.err.println(" - " + iter.next());
      System.err.println("Content of configuration map:");
      for (Iterator iter = configMap.values().iterator(); iter.hasNext(); ) 
        System.err.println(" - " + iter.next());
    }

  } // END ParsingErrorHandler

} // END ConfigParser
