/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.net.InetAddress;
import java.net.UnknownHostException;

import jgroup.core.ConfigurationException;
import jgroup.core.MemberId;
import jgroup.util.Network;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *  The <code>DistributedSystemConfig</code> object contains a
 *  description of the distributed system on which distributed
 *  computations can be executed.  Currently, a distributed system is
 *  described as a collection of domains and hosts, and is read into
 *  this object from an XML configuration file.  See the Jgroup
 *  programmers manual for details about the format of this
 *  configuration file.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class DistributedSystemConfig
  implements ConfigurationObject
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String DOMAIN_TAG_NAME            = "Domain";
  private static final String HOST_TAG_NAME              = "Host";
  private static final String NAME_ATTRIB                = "name";
  private static final String ADDRESS_ATTRIB             = "address";
  private static final String PORT_ATTRIB                = "port";
  private static final String JDAEMONS_ATTRIB            = "jdaemons";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Set of all hosts in this distributed system; independent of domains. */
  private static final HostSet allHosts = new HostSet();

  /** The endpoint considered to be the local host. */
  private static Host localHost = null;

  /** The endpoint considered to be the local domain. */
  private static Domain localDomain = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Set of domains in this distributed system */
  private DomainSet domains = new DomainSet();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty distributed system.
   */
  public DistributedSystemConfig()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(Element elm)
    throws ConfigurationException
  {
    /* Nodes within DS should be Domain tags. */
    NodeList domainList = elm.getChildNodes();
    
    /* Parse the list of domains */
    for (int i=1; i <= domainList.getLength(); i++) {
      Node node = domainList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element) node;
        if (ConfigParser.checkTag(elm, DOMAIN_TAG_NAME)) 
          parseDomain(elm);
      }
    }
    /*
     * If this exception is thrown, it is likely that the localhost
     * is not configured correctly (check your local /etc/hosts file)
     */
    if (localHost != null && localHost.getAddress().isLoopbackAddress()) {
      throw new ConfigurationException("Cannot use loopback as a configured server address");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods for parsing
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void parseDomain(Element elm)
    throws ConfigurationException
  {
    /* 
     *  Read attributes; "-" in the domain name corresponds to empty 
     */
    String domainname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
    if ("-".equals(domainname))
      domainname = "";
    String address = ConfigParser.getAttrib(elm, ADDRESS_ATTRIB, true);
    int port = ConfigParser.getIntAttrib(elm, PORT_ATTRIB, false);
    int jdaemons = ConfigParser.getIntAttrib(elm, JDAEMONS_ATTRIB, false);

    /* Create a new domain object with the parsed information */
    Domain domain = null;
    try {
      domain = new Domain(domainname, address, port, jdaemons);
    } catch (UnknownHostException e) {
      throw new ConfigurationException("Unknown multicast address: " + address);
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(e.getMessage());
    }

    /* Add the domain to the system */
    if (!domains.addDomain(domain))
      throw new ConfigurationException("Multiple definitions of domain: " + domain);

    /* Nodes within the domain should be host tags. */
    NodeList hostList = elm.getChildNodes();
    /* Parse the list of hosts */
    for (int i=1; i <= hostList.getLength(); i++) {
      Node node = hostList.item(i);
      if (node != null && !node.getNodeName().startsWith("#")) {
        elm = (Element)node;
        if (ConfigParser.checkTag(elm, HOST_TAG_NAME)) 
          parseHost(domain, elm);
      }
    }
  }


  private void parseHost(Domain domain, Element elm)
    throws ConfigurationException
  {
    String hostname = ConfigParser.getAttrib(elm, NAME_ATTRIB, true);
    int port = ConfigParser.getIntAttrib(elm, PORT_ATTRIB, true);

    /* Create a new host object with the parsed information */
    Host host = null;
    try {
      host = new Host(hostname, domain, port);
    } catch(UnknownHostException e) {
      throw new ConfigurationException("Unknown host: " + hostname);
    }

    /* Add the host to this domain */
    if (!domain.addHost(host))
      throw new ConfigurationException("Multple definitions of host: " + host);

    /* Add the host to the all hosts set */
    allHosts.addHost(host);

    /* Check if this host is a local host */
    if (host.isLocal()) {
      localHost = host;
      localDomain = domain;
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the parsed set of domains.
   */
  public DomainSet getDomainSet()
  {
    return domains;
  }


  /**
   *  Returns a set of all hosts parsed; independent of domain.
   */
  public HostSet getAllHosts()
  {
    return allHosts;
  }


  /**
   *  Returns the size of all hosts parsed; independent of domain.
   */
  public int size()
  {
    return allHosts.size();
  }


  /**
   *  Returns the number of domains parsed.
   */
  public int numOfDomains()
  {
    return domains.size();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods - only for server side
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the host object associated with the given member.
   */
  public static Host getHost(MemberId member)
  {
    return allHosts.getHost(member);
  }


  /**
   *  Returns the host object associated with the given internet address.
   */
  public static Host getHost(InetAddress inetAdr)
  {
    return allHosts.getHost(inetAdr);
  }


  /**
   *  Returns the endpoint that is considered local to this host.
   */
  public static Host getLocalHost()
  {
    return localHost;
  }


  /**
   *  Returns the domain endpoint that is considered local to this
   *  domain.
   */
  public static Domain getLocalDomain()
  {
    return localDomain;
  }


  /**
   *  Returns true if the local host is specified in the distributed
   *  system configuration; false is returned otherwise.
   */
  public boolean containsLocalHost()
  {
    return localHost != null;
  }


  /**
   *  Set the local host and add it to the set of available hosts.
   *  This should be used with care, and only for bootstrap purposes
   *  since other servers will not find the local host unless it is
   *  specified in the <code>config.xml</code> file.
   *
   *  Local host is not set if already defined in the configuration.
   *
   *  @throws UnknownHostException
   */
  public static void setLocalHost()
    throws UnknownHostException
  {
    if (localHost == null) {
      localDomain = new Domain(Network.getLocalDomainName());
      localHost = new Host(Network.getLocalMachineName(), localDomain, 1991);
      allHosts.addHost(localHost);
      System.out.println("hosts: " + allHosts);
    }
  }


  /**
   *  Ping all hosts in the current configuration.  Returns true if
   *  at least one host does not respond; i.e., the <code>ExecService</code>
   *  is not running on one or more hosts in the system.  False indicates
   *  that all is well.
   */
  public static boolean updateAll()
  {
    HostSet availHosts = allHosts.getAvailHosts();
    return !availHosts.containsAll(allHosts);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int hashCode()
  {
    return domains.hashCode() ^ allHosts.hashCode();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    return toString(false);
  }

  /**
   *  Returns a string representation of this object;
   *  
   *  @param full If true, prints also the complete host set for each domain.
   */
  public String toString(boolean full)
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[DistributedSystemConfig:\n");
    buf.append(domains.toString(full));
    buf.append("]");
    return buf.toString();
  }

} // END DistributedSystemConfig
