/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.io.Serializable;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import jgroup.core.MemberId;

/**
 *  Collection class for storing information about a set of domains.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class DomainSet
  implements Iterable<Domain>, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static comparator for domain sets
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Comparator to be provided to domain sets in order to order domains correctly,
   * ignoring the number of replicas in each domain.  That is override the default
   * ordering for domain sets which is based on the number of replicas in each domain.
   */
  public static final Comparator<Domain> domainComparator = new Comparator<Domain>() {
    public int compare(Domain o1, Domain o2) {
      int o1Adr = o1.getIntAddress();
      int o2Adr = o2.getIntAddress();
      if (o1Adr < o2Adr)
        return -1;
      else if (o1Adr == o2Adr)
        return 0;
      else
        return 1;
    }
  };


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * List of <code>Domain</code> objects; the list will be sorted before each access
   * to its iterator.  This sorting is required since the domain objects may change
   * their individual position in the list with time (as more replicas and/or hosts
   * are added to them).  Hence, we cannot use a sorted set, because that will only
   * give correct order at insertion time.
   */
  private final List<Domain> domains;

  /** Dynamic array of domain listeners */
  private final List<DomainListener> domainListeners = new ArrayList<DomainListener>();

  /** Comparator for the domain set iteration ordering (may be undefined) */
  private Comparator<Domain> comparator;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty <code>DomainSet</code>.
   */
  public DomainSet()
  {
    this(null, null);
  }

  /**
   * Constructs an empty <code>DomainSet</code> whose iteration order is
   * defined by the given <code>comparator</code>.
   * 
   * @param comparator
   */
  public DomainSet(Comparator<Domain> comparator)
  {
    this(null, comparator);
  }

  /**
   * Constructs a new <code>DomainSet</code> containg the given domain set, 
   * whose iteration order is defined by the given <code>comparator</code>.
   * 
   * @param domains
   * @param comparator
   */
  public DomainSet(DomainSet domains, Comparator<Domain> comparator)
  {
    if (domains == null) {
      this.domains = new ArrayList<Domain>();
    } else {
      this.domains = new ArrayList<Domain>(domains.domains);
    }
    this.comparator = comparator;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Listener interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void addListener(DomainListener listener)
  {
    domainListeners.add(listener);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // DomainSet interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a new domain to this <code>DomainSet</code>.
   *
   *  @param domainName
   *    The domain name.
   *  @param mcastAdr
   *    The multicast address for this domain.
   *  @param port
   *    The multicast port for this domain.
   *  @param jdaemons
   *    The number of Jgroup daemons that should exist in the new domain.
   *  @return 
   *    True if the <code>DomainSet</code> did not already contain this domain.
   *  @throws UnknownHostException
   *    Is raised if the multicast address is not recognized.
   */
  public boolean addDomain(String domainName, String mcastAdr, int port, int jdaemons)
    throws UnknownHostException
  {
    return addDomain(new Domain(domainName, mcastAdr, port, jdaemons));
  }


  /**
   *  Add a new domain to this <code>DomainSet</code>.
   *
   *  @param domain
   *    A domain object.
   *  @return 
   *    True if the <code>DomainSet</code> did not already contain this domain.
   */
  public boolean addDomain(Domain domain)
  {
    boolean added = !domains.contains(domain);
    if (added) {
      domains.add(domain);
      for (int i = 0, size = domainListeners.size(); i < size; i++) {
        ((DomainListener) domainListeners.get(i)).addDomain(domain);
      }
    }
    return added;
  }


  /**
   *  Remove the specified domain from this <code>DomainSet</code>.
   *
   *  @param domain
   *    A domain object.
   *  @return 
   *    True if the specified domain was removed from this
   *    <code>DomainSet</code>; false otherwise.
   */
  public boolean removeDomain(Domain domain)
  {
    boolean removed = domains.remove(domain);
    if (removed) {
      for (int i = 0, size = domainListeners.size(); i < size; i++) {
        ((DomainListener) domainListeners.get(i)).removeDomain(domain);
      }
    }
    return removed;
  }


  /**
   *  Check if the specified domain is in this <code>DomainSet</code>.
   *
   *  @param domain
   *    A domain object.
   *  @return 
   *    True if the specified domain is in this <code>DomainSet</code>;
   *    false otherwise.
   */
  public boolean containsDomain(Domain domain)
  {
    return domains.contains(domain);
  }


  /**
   * Returns true if this domain set contains the given <code>InetAddress</code>.
   * 
   * @param inetAdr
   * @return
   */
  public boolean contains(InetAddress inetAdr)
  {
    for (Domain domain : domains) {
      for (Host host : domain.getHostSet()) {
        if (host.getAddress().equals(inetAdr))
          return true;
      }
    }
    return false;
  }


  /**
   * Returns the <code>Domain</code> object stored in this domain set
   * corresponding to the provided domain.  If no domain matching the provided
   * domain is available in this domain set, <code>null</code> is returned.
   */
  public Domain getDomain(Domain theDomain)
  {
    for (Domain domain : domains) {
      if (domain.equals(theDomain)) {
        return domain;
      }
    }
    return null;
  }


  /**
   * Returns the <code>Domain</code> object stored in this domain set
   * corresponding to the provided multicast <code>InetAddress</code>.
   * If no domain matching the provided domain is available in this
   * domain set, <code>null</code> is returned.
   * 
   * @throws IllegalArgumentException if the provided <code>InetAddress</code>
   *   is not a multicast address.
   */
  public Domain getDomain(InetAddress mcastAdr)
  {
    if (!mcastAdr.isMulticastAddress())
      throw new IllegalArgumentException(mcastAdr + " is not a multicast address");
    for (Domain domain : domains) {
      if (domain.getAddress().equals(mcastAdr)) {
        return domain;
      }
    }
    return null;
  }

  /**
   * Returns the <code>Domain</code> object stored in this domain set
   * corresponding to the provided host <code>InetAddress</code>.
   * If no domain contains the provided host address in this domain set,
   * <code>null</code> is returned.
   * 
   * @param hostAdr the host address whose domain object to return.
   * @return the domain in which the given host resides.
   * @throws IllegalArgumentException if the provided <code>InetAddress</code>
   *   is a multicast address instead of a regular host address.
   */
  public Domain getHostDomain(InetAddress hostAdr)
  {
    if (hostAdr.isMulticastAddress())
      throw new IllegalArgumentException(hostAdr + " is a multicast address");
    for (Domain domain : domains) {
      if (domain.getHostSet().containsHost(hostAdr)) {
        return domain;
      }
    }
    return null;
  }

  /**
   * @param comparator The comparator to set.
   */
  public void setComparator(Comparator<Domain> comparator)
  {
    this.comparator = comparator;
  }

  /**
   *  Returns the number of domains stored in this <code>DomainSet</code>.
   */
  public int size()
  {
    return domains.size();
  }


  /**
   *  Returns true if this <code>DomainSet</code> is empty.
   */
  public boolean isEmpty()
  {
    return domains.isEmpty();
  }


  /**
   *  Returns an iterator over the values of the <code>DomainSet</code>.
   *  The values returned by the <code>next</code> method call of the
   *  iterator should be cast to <code>Domain</code>.
   *  
   *  This iterator returns the sorted order of the domains in this set.
   *
   *  @return 
   *    An <code>Iterator</code> for the <code>DomainSet</code>.
   */
  @SuppressWarnings("unchecked")
  public Iterator<Domain> iterator()
  {
    /*
     * Note that we need to sort here because the domain objects in this set
     * may change their individual positions with time, depending on the number
     * of replicas and number of hosts associated with the domain.  Hence, we
     * cannot use a sorted set, because that will only give correct order at
     * insertion time.
     */
    Collections.sort(domains, comparator);
    return domains.iterator();
  }


  /**
   * Returns an array of domain objects, in sorted order.
   * 
   * @return
   */
  @SuppressWarnings("unchecked")
  public Domain[] toArray()
  {
    Collections.sort(domains, comparator);
    return (Domain[]) domains.toArray(new Domain[domains.size()]);
  }


  /**
   *  Returns the first <code>Domain</code> in the domain set; the domain
   *  returned is the one with the lowest replica count.
   *
   *  @return 
   *    The <code>Domain</code> object found.
   *  @throws NoSuchElementException
   *    Raised if there are no domains in this <code>DomainSet</code>.
   */
  @SuppressWarnings("unchecked")
  public Domain first()
  {
    Collections.sort(domains, comparator);
    return domains.get(0);
  }


  /**
   *  Returns the domain that contains the largest number of hosts.
   */
  public Domain getLargestDomain()
  {
    if (isEmpty()) 
      throw new NoSuchElementException("No domains have been added to the domain set.");
    Domain largestDomain = null;
    for (Domain domain : domains) {
      if (largestDomain == null || largestDomain.size() < domain.size()) {
        largestDomain = domain; 
      }
    }
    return largestDomain;
  }


  /**
   *  Given a set of members, return the <code>Domain</code>
   *  that contains the most members.  It is not specified which
   *  domain is returned when more than one domain contains an
   *  equal number of members.
   * 
   *  @param members The members from which to find the domain.
   *  @return The domain with the largest number of members amongst
   *     the provided set.
   */
  public static Domain getLargestReplicaDomain(MemberId[] members)
  {
    DomainSet domains = new DomainSet();
    for (int i = 0; i < members.length; i++) {
      Host host = DistributedSystemConfig.getHost(members[i]);
      /*
       * Obtain the domain of the host object, and check if this
       * domain already exists within the domain set.
       */
      Domain hostDomain = host.getDomain();
      Domain domain = domains.getDomain(hostDomain);
      if (domain == null) {
        /*
         * Construct a new Domain without any hosts associated with it,
         * and add it to the domain set, since it was not previously
         * stored in the domain set.
         */
        domain = new Domain(hostDomain, false);
        domains.addDomain(domain);
      }
      domain.addHost(host);
    }
    return domains.getLargestDomain();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int hashCode()
  {
    return domains.hashCode();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    return toString(false);
  }

  /**
   *  Returns a string representation of this object
   */
  public String toString(boolean full)
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[DomainSet: ");
    for (Domain domain : domains) {
      /* One domain per line */
//      buf.append("\n ");
      buf.append(domain.toString(full));
    }
    buf.append("]");
    return buf.toString();
  }

} // END DomainSet
