/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.config;

import java.util.HashMap;
import java.util.Map;

import jgroup.core.ConfigurationException;


/**
 *  Object for holding service information.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class ServiceConfig
  implements ConfigurationObject
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Tags
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Attributes for the Service tag
  private static final String NAME_ATTRIB      = "name";
  private static final String INTERFACE_ATTRIB = "interface";
  private static final String LISTENER_ATTRIB  = "listener";
  private static final String EQUALS_ATTRIB    = "equals";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static collections
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Map of services (association from service name to Service object).
   * Note that this map must be static to allow cross access between
   * instances of the ServiceConfig class.  This is because the parser
   * creates a new ServiceConfig object for each Service entry in the
   * XML file.
   */
  private static final Map<String, ServiceConfig> services =
    new HashMap<String, ServiceConfig>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The service interface */
  private String iface;
  private Class serviceIface;

  /** The listener interface */
  private String listener;
  private Class listenerIface;

  /** True if the listener interface is an empty marker interface */
  private boolean listenerIsMarker;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs an empty service object.
   */
  public ServiceConfig() { }


  /**
   *  Constructs a service config object without listener interface.
   */
  public ServiceConfig(String iface)
    throws ConfigurationException
  {
    this.iface = iface;
    this.listener = null;
    /*
     * Set marker to false; that is we should never call addListener()
     * for services that does not provide a marker interface.
     */
    listenerIsMarker = false;
  }


  /**
   *  Constructs a service info object.
   */
  public ServiceConfig(String iface, String listener)
    throws ConfigurationException
  {
    this.iface = iface;
    this.listener = listener;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Class checkInterface(String clname)
    throws ConfigurationException
  {
    Class cl;
    try {
      cl = Class.forName(clname);
    } catch (ClassNotFoundException e) {
      throw new ConfigurationException("Service interface not found: " + clname);
    }
    if (!cl.isInterface())
      throw new ConfigurationException("Service " + clname + " is not an interface");
    return cl;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ConfigurationObject
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Parses the specified element.
   */
  public void parse(org.w3c.dom.Element elm)
    throws ConfigurationException
  {
    String servicename  = ConfigParser.getAttrib(elm, NAME_ATTRIB,      true);
    String iface        = ConfigParser.getAttrib(elm, INTERFACE_ATTRIB, false);
    String listener     = ConfigParser.getAttrib(elm, LISTENER_ATTRIB, false);
    String equalService = ConfigParser.getAttrib(elm, EQUALS_ATTRIB, false);

    ServiceConfig serviceConfig = null;
    if (iface != null && iface.length() > 0) {
      if (listener != null && listener.length() > 0) {
        /* Create a new service object with the parsed information */
        serviceConfig = new ServiceConfig(iface, listener);
      } else {
        /* Create a new service object with the parsed information */
        serviceConfig = new ServiceConfig(iface);
      }
    } else if (equalService != null && equalService.length() > 0) {
      serviceConfig = ServiceConfig.getService(equalService);
    } else {
      throw new ConfigurationException("Service interfaces not defined: " + servicename);
    }
    services.put(servicename, serviceConfig);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static ServiceConfig getService(String serviceName)
    throws ConfigurationException
  {
    ServiceConfig serviceInfo = (ServiceConfig) services.get(serviceName);
    if (serviceInfo == null)
      throw new ConfigurationException("The service " + serviceName + " is not specified");
    return serviceInfo;
  }
  

  public static Class getServiceIface(String serviceName)
    throws ConfigurationException
  {
    ServiceConfig serviceInfo = (ServiceConfig) services.get(serviceName);
    if (serviceInfo == null)
      throw new ConfigurationException("The service " + serviceName + " is not specified");
    return serviceInfo.getInterface();
  }
  

  public static Class getListenerIface(String serviceName)
    throws ConfigurationException
  {
    ServiceConfig serviceInfo = (ServiceConfig) services.get(serviceName);
    if (serviceInfo == null)
      throw new ConfigurationException("The service " + serviceName + " is not specified");
    return serviceInfo.getListener();
  }


  public static boolean contains(String serviceName)
  {
    return services.containsKey(serviceName);
  }


  public Class getInterface()
    throws ConfigurationException
  {
    serviceIface = checkInterface(iface);
    return serviceIface;
  }


  public Class getListener()
    throws ConfigurationException
  {
    listenerIface = checkInterface(listener);
    return listenerIface;
  }


  /**
   *  Returns true if the listener interface is a marker (empty) interface;
   *  otherwise false is returned.
   *
   *  @throws ConfigurationException
   *    Thrown if the expected listener interface does not exist.
   */
  public boolean isMarker()
    throws ConfigurationException
  {
    if (listenerIface == null) {
      listenerIface = checkInterface(listener);
    }
    java.lang.reflect.Method[] methods = listenerIface.getDeclaredMethods();
    listenerIsMarker = (methods.length == 0) ? true : false;
    return listenerIsMarker;
  }


  public String toString()
  {
    return "[service=" + iface + ", listener=" + listener + "]";
  }

} // END ServiceConfig
