/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jgroup.core.EndPoint;

import org.apache.log4j.Logger;


/**
 * The <code>GroupTable</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 1.2
 */
final class GroupTable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupTable.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Hash table containing the groups in the system */
  private final Map<Integer,Group> totgroups = new HashMap<Integer,Group>();

  /** Hash map: host to a list of Group objects on that host */
  private final Map<EndPoint,List<Group>> grouplists = new HashMap<EndPoint,List<Group>>();

  /** Hash table that maps host to symmetry sets */ 
  private final Map<EndPoint,EndPoint[]> symsets = new HashMap<EndPoint,EndPoint[]>();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  GroupTable() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  @Override
  public String toString()
  {
    StringBuilder buf = new StringBuilder("GroupTable: groups=");
    for (Group group : totgroups.values()) {
      buf.append(group.getGid());
      buf.append(" ");
    }
    for (Map.Entry<EndPoint,List<Group>> groupEntry : grouplists.entrySet()) {
      buf.append("endpoint=");
      buf.append(groupEntry.getKey());
      buf.append(": ");
      for (Group group : groupEntry.getValue()) {
        buf.append(group.getGid());
      }
      buf.append("; ");
    }
    return buf.toString();
  }

  /**
   *  Add a new host/endpoint to the given group.
   */
  void addEndpointToGroup(EndPoint endpoint, Group group)
  {
    if (log.isDebugEnabled()) 
      log.debug("GroupTable: addGroupEndPoint: " + endpoint);
    if (!containsGroup(group.getGid()))
      throw new IllegalStateException("Cannot add group endpoint to a non-local group.");
    List<Group> groups = grouplists.get(endpoint);
    if (groups == null) {
      groups = new ArrayList<Group>(10);
      grouplists.put(endpoint, groups);
    }
    if (!groups.contains(group))
      groups.add(group);

    if (log.isDebugEnabled())
      log.debug(toString());
  }

  /**
   *  Insert a new group in the group table.
   */
  void insertGroup(Group group)
  {
    totgroups.put(group.getGid(), group);
  }

  /**
   *  Remove the given group from the group table.
   */
  void removeGroup(int gid)
  {
    if (log.isDebugEnabled())
      log.debug("Removing group: " + gid);

    Group group = getGroup(gid);
    for (Iterator iter = grouplists.entrySet().iterator(); iter.hasNext();) {
      Map.Entry groupEntry = (Map.Entry) iter.next();
      List groups = (List) groupEntry.getValue();
      groups.remove(group);
      if (log.isDebugEnabled())
        log.debug("Removed group " + group.getGid() + " from host " + groupEntry.getKey());
      if (groups.isEmpty())
        iter.remove();
    }
    totgroups.remove(gid);

    if (log.isDebugEnabled())
      log.debug(toString());
  }

  /**
   *  Return the group object for the given group identifier.
   */
  Group getGroup(int gid)
  {
    return totgroups.get(gid);
  }

  /**
   *  Returns true if the given group is contained in this group table.
   */
  boolean containsGroup(int gid)
  {
    return totgroups.containsKey(gid);
  }

  /**
   *  Returns true if the group table contains no local group members.
   */
  boolean isEmpty()
  {
    return totgroups.isEmpty();
  }
    
  /**
   *  Returns an iterator containing all the the groups contained in
   *  this group table.
   */
  Iterator iterator()
  {
    return totgroups.values().iterator();
  }
  
  /**
   *  Returns an iterator containing the list of groups
   *  joined by the specified host.
   */
  Iterator getGroups(EndPoint endpoint) 
  {
    List<Group> groups = grouplists.get(endpoint);
    /* 
     * Ensures that we return an empty iterator if there are no groups
     * for the specified host.
     */
    if (groups == null)
      groups = new ArrayList<Group>(0);
    return groups.iterator();
  }

  EndPoint[] getSymset(EndPoint endpoint)
  {
    return symsets.get(endpoint);
  }
  
  void setSymset(EndPoint host, EndPoint[] symset)
  {
    symsets.put(host, symset);
  }

} // END GroupTable
