/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssHost;

/**
 * The <code> HostData</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class HostData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** MssHost to which this information is related */
  private final MssHost host;

  /** Coordinator table */ 
  private MsgProp ctbl;

  /** If the host is leaving the group */ 
  private boolean leaving;

  /** Index inside a view */ 
  private int viewIndex;
  
  /** Version number */ 
  private int version;

  /** Agreed version number */ 
  private int agreed;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * Creates a new additional info object for host <code>host</code>.
   */
  HostData(MssHost host)
  {
    this.host   = host;
    /*
     * FIXED Hein: viewIndex was 0, but this may cause the agreement
     * protocol to think that this host has a valid view index before
     * a view has been installed which includes this host.
     */
    resetViewIndex();
    version     = 0;
    agreed      = 0;
    leaving     = false;
    ctbl        = null;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the endpoint associated with this host. 
   */
  EndPoint getEndPoint()
  {
    return host.getEndPoint();
  }
  
  /**
   *  Returns the <code>MssHost</code> associated with this object. 
   */
  MssHost getHost()
  {
    return host;
  }

  /**
   *  Returns the last proposal received from this host
   */
  MsgProp getProposal()
  {
    return ctbl;
  }

  /**
   *  Store the last proposal received from this host. 
   */
  void setProposal(MsgProp ctbl) 
  {
    this.ctbl = ctbl;
  }

  /**
   *  Returns true if this host is leaving.
   */
  boolean isLeaving()
  {
    return leaving;
  }

  /**
   *  Set this host as leaving.
   */
  void setLeaving(boolean leaving) 
  {
    this.leaving = leaving;
  }

  /**
   *  Returns the version number.
   */
  int getVersion()
  {
    return version;
  }

  /**
   *  Set the version number. 
   */
  void setVersion(int version) 
  {
    this.version = version;
  }

  /**
   *  Returns the agreed version number.
   */
  int getAgreed()
  {
    return agreed;
  }

  /**
   *  Set the agreed version number. 
   */
  void setAgreed(int agreed) 
  {
    Group.log.assertLog(this.agreed <= agreed, getEndPoint() +
        ": new agreed is smaller ("+agreed+") than the previous ("+this.agreed+")");
    if (this.agreed < agreed)
      this.agreed = agreed;
  }

  /**
   *  Returns the index of this host in the current view.
   */
  int getViewIndex()
  {
    return viewIndex;
  }

  /**
   *  Set the index of this host in the current view.
   */
  void setViewIndex(int viewIndex) 
  {
    this.viewIndex = viewIndex;
  }

  /**
   *  Reset the view index of this host.  This indicates that this host
   *  is not part of the current view.
   */
  void resetViewIndex()
  {
    viewIndex = -1;
  }

  /**
   *  Returns true if this host's view index is valid; false otherwise.
   */
  boolean hasValidViewIndex() 
  {
    return (viewIndex >= 0);
  }

  /**
   *  Resets the data associated to this host because a new incarnation
   *  of it is present.
   */
  void newIncarnationId()
  {
    version = 0;
    agreed  = 0;
    resetViewIndex();
    /*
     * FIXED HEIN: A new incarnation may no longer be member of the same group.
     * 
     * FIXME this causes problems when a new member suddenly see an old member
     * from a different partition.
     *  
     * It should be handled elsewhere; not related to a new incarnation, but that
     * a member/host is no longer interested in this group.  I cannot recall when
     * this caused problems, so we need to fix this after we can identify the
     * problem fully.
     */
//    leaving = true;
  }

  /**
   *  Creates a new version number.
   */
  void newVersion()
  {
    version++;
    agreed++;
  }
  
  public String toString()
  {
    StringBuilder buf = new StringBuilder("[HostData: ");
    buf.append(host.getEndPoint().toString());
    if (leaving)
      buf.append(", leaving");
    buf.append(", version=");
    buf.append(version);
    buf.append(", agreed=");
    buf.append(agreed);
    buf.append(", viewIndex=");
    if (hasValidViewIndex())
      buf.append(viewIndex);
    else
      buf.append("NO");
    buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the list of <code>HostData</code> objects associated with
   *  the hosts contained in the given <code>list</code>.  The
   *  <code>HostData</code> objects are contained in <code>map</code>.
   */
  static List<HostData> extract(Map<EndPoint,HostData> map, EndPoint[] list)
  {
    List<HostData> ret = new ArrayList<HostData>();
    for (int i=0; i < list.length; i++) {
      HostData hostData = map.get(list[i]);
      if (hostData != null)
        ret.add(hostData);
    }
    return ret;
  }

} // END HostData
