/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.VersionId;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

/**
 * The <code> MsgSynch</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgSynch
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START  = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message
   * instance is sufficient. In this way, we save the costs of
   * allocating and garbage collecting messages.
   */
  private static MsgSynch msg = new MsgSynch();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  int                   vsend;          // Sender version number
  int                   vdest;          // Destination version number
  EndPoint[]            rset;           // Reachable set (32 bit adressess)


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Decodes a <code>MsgSynch</code> message starting from a m-received
   *  message input stream.
   */
  static MsgSynch unmarshal(ObjectInput stream)
    throws IOException, ClassNotFoundException
  {
    // Read data
    msg.vsend = VersionId.unmarshal(stream);
    msg.vdest = VersionId.unmarshal(stream);
    int len = GroupIndex.unmarshal(stream);     // Length of the rset field
    msg.rset = new EndPoint[len];
    for (int i=0; i < len; i++) {
      msg.rset[i] = new EndPointImpl();
      msg.rset[i].readExternal(stream);
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Computes the encoding of a <code> MsgSynch</code> message.
   *
   * @param     gid     group identifier.
   * @param     vsend   version number of the sender host
   * @param     vdest   version number of the destination host
   * @param     rset    reachable set
   */
  static MsgSynch marshal(int gid, int vsend, int vdest, EndPoint[] rset)
    throws IOException
  {
    /* Allocates data buffer */
    int size = START + GroupId.SIZE + 2*VersionId.SIZE + GroupIndex.SIZE
      + rset.length * EndPointImpl.SIZE;
    outmsg = MsgFactory.get(size, 1);
    outmsg.seek(START);

    /* Write group identifier and sender, destination version numbers */
    GroupId.marshal(outmsg, gid);
    VersionId.marshal(outmsg, vsend);
    VersionId.marshal(outmsg, vdest);

    /*
     *  Writes the identifier of the endpoints contained in rset
     */
    GroupIndex.marshal(outmsg, rset.length);
    for (int i=0; i < rset.length; i++)
      rset[i].writeExternal(outmsg);
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    /* Note; it is ok to use the outmsg size since MsgSynch will never be
     * forwarded or resent (it is not queued). */
    return outmsg.size();
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgSynch: vsend=");
    buffer.append(vsend);
    buffer.append(", vdest=");
    buffer.append(vdest);
    buffer.append(", rset={");
    for (int i=0; i < rset.length; i++) {
      buffer.append(rset[i]);
      if (i < rset.length-1)
        buffer.append(", ");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END MsgSynch
