/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.daemon;


import java.io.IOException;
import java.io.ObjectInput;

import jgroup.core.EndPoint;
import jgroup.relacs.mss.MssConstants;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.GroupId;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.IncarnationId;
import jgroup.relacs.types.LocalId;
import jgroup.relacs.types.MessageLen;
import jgroup.relacs.types.VersionId;
import jgroup.relacs.types.ViewId;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;


/**
 * The <code> MsgView</code> class
 *
 * @author  Alberto Montresor
 * @since  Jgroup 0.1
 */
final class MsgView
  implements MssConstants, DaemonMsg
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int START  = MSS_HEADER_SIZE;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message
   * instance is sufficient.  In this way, we save the costs of
   * allocating and garbage collecting messages.
   */
  private static MsgView msg = new MsgView();

  /** The marshalled message data (also static for same reason as above) */
  private static OutMessage outmsg;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Original stream received from a coordinator, used to resend messages. */
  InMessage stream;
  
  /** New complete view identifier */
  long cvid;
  
  /** New compete view composition */
  EndPoint[] cvcomp;
  
  /** */                                  
  int[] incarns;
  
  /** Agreed version numbers */          
  int[] agreed;
  
  /** Partial view identifiers */         
  int[] pvids;
  
  /** Members tables */
  LocalId[][] members;

  /** MsgView message size (excluding the mss header) */
  private int size;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a <code>MsgView</code> object that contains the decoding
   *  of the m-received input stream.
   *
   *  @param stream    the message input stream to decode
   */
  static MsgView unmarshal(ObjectInput stream)
    throws IOException, ClassNotFoundException
  {
    msg.stream = (InMessage) stream;
    msg.cvid = ViewId.unmarshal(stream);
    msg.size = MessageLen.unmarshal(stream);
    int len  = GroupIndex.unmarshal(stream);
    msg.cvcomp  = new EndPoint[len];
    msg.incarns = new int[len];
    msg.agreed  = new int[len];
    msg.pvids   = new int[len];
    msg.members = new LocalId[len][];
    for (int i=0; i < len; i++) {
      msg.cvcomp[i] = new EndPointImpl();
      msg.cvcomp[i].readExternal(stream);
    }
    for (int i=0; i < len; i++) 
      msg.incarns[i] = IncarnationId.unmarshal(stream);
    for (int i=0; i < len; i++) 
      msg.agreed[i]  = VersionId.unmarshal(stream);
    for (int i=0; i < len; i++) 
      msg.pvids[i]   = GroupIndex.unmarshal(stream);
    for (int i=0; i < len; i++) {
      msg.members[i] = new LocalId[GroupIndex.unmarshal(stream)];
      for (int j=0; j < msg.members[i].length; j++) {
        msg.members[i][j] = new LocalId();
        msg.members[i][j].readExternal(stream);
      }
    }
    return msg;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message to send
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Encode a message starting from a previously received message.
   */
  MsgView marshal()
  {
    outmsg = new OutMessage(stream);
    return msg;
  }


  /**
   *  Returns a output stream containing the encoding of a
   *  <code>MsgView</code> message.
   *
   *  @param     gid     group identifier
   *  @param     cvid    identifier of the current complete view
   *  @param     prop    proposal on which the agreement has been reached
   *  @param     pvids   identifiers of previous partial views (needed to compute new partial
   *                     views
   *  @param     props   proposal array (containing members lists)
   */
  static MsgView marshal(int gid, long cvid, MsgProp prop, int[] pvids,
                         MsgProp[] props, int nreceivers)
    throws IOException
  {
    // Allocates data buffer
    msg.size = GroupId.SIZE + ViewId.SIZE + MessageLen.SIZE + GroupIndex.SIZE
      + prop.host_plen + pvids.length * GroupIndex.SIZE;
    for (int i=0; i < props.length; i++)
      msg.size += props[i].memb_plen;
    outmsg = MsgFactory.get(START + msg.size, nreceivers);
    outmsg.seek(START);

    // Writes single data
    GroupId.marshal(outmsg, gid);
    /*
     * There should've been a GroupIndex.marshal() here, but this is
     * instead in MsgProp.encodeHostProp().
     */
    ViewId.marshal(outmsg, cvid);
    MessageLen.marshal(outmsg, msg.size);

    // Copies composition and agreed values
    prop.stream.seek(prop.host_start);
    byte[] tempbuffer = new byte[prop.host_plen];
    prop.stream.read(tempbuffer);
    outmsg.write(tempbuffer);

    // Writes coordinator table informations
    for (int i=0; i < pvids.length; i++)
      GroupIndex.marshal(outmsg, pvids[i]);
    for (int i=0; i < props.length; i++) {
      props[i].stream.seek(props[i].memb_start);
      tempbuffer = new byte[props[i].memb_plen];
      props[i].stream.read(tempbuffer);
      outmsg.write(tempbuffer);
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DaemonMsg
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int size()
  {
    return size;
  }

  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[MsgView: cvid=");
    buffer.append(cvid);
    buffer.append(", cvcomp={");
    for (int i=0; i < cvcomp.length; i++) {
      buffer.append(cvcomp[i]);
      if (i < cvcomp.length-1)
        buffer.append(", ");
    }
    buffer.append("}, agreed={");
    for (int i=0; i < agreed.length; i++) {
      buffer.append(agreed[i]);
      if (i < agreed.length-1)
        buffer.append(", ");
    }
    buffer.append("}, pvids={");
    for (int i=0; i < pvids.length; i++) {
      buffer.append(pvids[i]);
      if (i < pvids.length-1)
        buffer.append(", ");
    }
    buffer.append("}, members={");
    for (int i=0; i < members.length; i++) {
      buffer.append("{");
      for (int j=0; j < members[i].length; j++) {
        buffer.append(members[i][j]);
        if (j < members[i].length-1)
          buffer.append(",");
      }
      buffer.append("}");
      if (i < members.length-1)
        buffer.append(",");
    }
    buffer.append("}]");
    return buffer.toString();
  }
  
} // END MsgView

