/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.MemberId;
import jgroup.relacs.types.Flag;
import jgroup.relacs.types.GroupIndex;
import jgroup.relacs.types.MemberIdImpl;
import jgroup.relacs.types.MessageId;
import jgroup.util.InMessage;

/**
 *  The <code>DeliveryEvent</code> class represents a local message 
 *  exchanged between members and the daemon; it is used by a daemon 
 *  to notify members that a message is ready to be delivered.
 *  
 *  @author Alberto Montresor
 *  @since  Jgroup 0.2
 */
public final class DeliveryEvent
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -327290828044626238L;

  /** Identifier of the sender of the message */
  private MemberId sender;

  /** Index of the sender host in the current view */
  private int hpos;

  /** True if the payload contains a single object */
  private boolean isObject;

  /** Message identifier */
  private int mid;

  /** True if an ack for this message is needed */
  private boolean ackRequired;

  /** Message contents */
  private InMessage payload;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public DeliveryEvent() { }


  /**
   *  Builds a new <CODE>DeliveryEvent</CODE> object.
   * 
   *  @param gid
   *    Group identifier
   *  @param isObject
   *    True if the payload contains a single object; false if it
   *    contains a stream
   *  @param mid
   *    Message identifier
   *  @param sender
   *    Sender identifier
   *  @param ackRequired
   *    True if acknowledgement is required; false otherwise
   *  @param payload
   *    A stream or object payload
   */
  public DeliveryEvent(int gid, boolean isObject, int mid, int hpos, 
    MemberId sender, boolean ackRequired, InMessage payload)
  {
    super(DELIVERY_EVENT, gid);
    this.isObject = isObject;
    this.mid = mid;
    this.hpos = hpos;
    this.sender = sender;
    this.ackRequired = ackRequired;
    this.payload = payload;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessor methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the sender of this multicast message.
   */
  public MemberId getSender()
  {
    return sender;
  }
  
  /**
   *  
   */
  public int getHostIndex()
  {
    return hpos;
  }

  /**
   *  Returns the message identifier for this multicast message.
   */
  public int getMessageId()
  {
    return mid;
  }
  
  /**
   *  
   */
  public boolean isObject()
  {
    return isObject;
  }
  
  /**
   *  
   */
  public boolean isAckRequired()
  {
    return ackRequired;
  }

  /**
   *  
   */
  public InMessage getPayload()
  {
    return payload;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[");
    buf.append(super.toString());
    buf.append(", hpos=");
    buf.append(hpos);
    buf.append(", mid=");
    buf.append(mid);
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", isObj=");
    buf.append(isObject);
    buf.append(", ackRequired=");
    buf.append(ackRequired);
    buf.append(", payload=");
    buf.append(payload);
    return buf.toString();    
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    sender = new MemberIdImpl();
    sender.readExternal(in);
    hpos = GroupIndex.unmarshal(in);
    isObject = Flag.unmarshal(in);
    ackRequired = Flag.unmarshal(in);
    mid = MessageId.unmarshal(in);
    payload = new InMessage();
    payload.readExternal(in);
  }


  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    sender.writeExternal(out);
    GroupIndex.marshal(out, hpos);
    Flag.marshal(out, isObject);
    Flag.marshal(out, ackRequired);
    MessageId.marshal(out, mid);
    payload.writeExternal(out);
  }

} // END DeliveryEvent
