/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 *  The <code>Event</code> class is the common superclass for all events
 *  that may be exchanged between the daemon and members.  Different
 *  types of events are distinguished using an integer tag, that may be
 *  obtained using method <code>getTag()</code>.  Each event is
 *  associated to a group identifier, that may be obtained using method
 *  <code>getGid()</code>.  The group identifier is needed because an
 *  application may be member of most than one group at a time.
 *
 *  @author Alberto Montresor
 *  @since  Jgroup 1.2
 */
public class Event
  implements EventTags, Externalizable
{

  private static final long serialVersionUID = 7265112762984225812L;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Tag identifying the event */
  protected int tag;

  /** Group identifier */
  protected int gid;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public Event()
  {
  }

  /**
   *  Constructs a new generic <CODE>Event</CODE> object.
   *
   *  @param tag tag identifying the event
   *  @param gid group identifier
   */
  public Event(int tag, int gid)
  {
    this.tag = tag;
    this.gid = gid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag identifying this event.
   */
  public int getTag()
  {
    return tag;
  }


  /**
   *  Returns the group identifier associated with this event.
   */
  public int getGid()
  {
    return gid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restores the content of this object from the marshalled data
   *  contained in the specified input stream.
   *
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    tag = in.readInt();
    gid = in.readInt();
  }


  /**
   *  Marshals the content of this object to the specified output
   *  stream.
   *
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(tag);
    out.writeInt(gid);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[Event: tag=");
    switch (tag) {
    case DELIVERY_EVENT:
      buf.append("DELIVERY_EVENT");
      break;
    case DELIVERY_ACK:
      buf.append("DELIVERY_ACK");
      break;
    case INSTALL_EVENT:
      buf.append("INSTALL_EVENT");
      break;
    case INSTALL_ACK:
      buf.append("INSTALL_ACK");
      break;
    case PREPARE_EVENT:
      buf.append("PREPARE_EVENT");
      break;
    case PREPARE_ACK:
      buf.append("PREPARE_ACK");
      break;
    case MULTICAST_REQUEST:
      buf.append("MULTICAST_REQUEST");
      break;
    case JOIN_REQUEST:
      buf.append("JOIN_REQUEST");
      break;
    case LEAVE_REQUEST:
      buf.append("LEAVE_REQUEST");
      break;
    case MEMBER_LEFT_EVENT:
      buf.append("MEMBER_LEFT_EVENT");
      break;
    case DAEMON_FAILURE_EVENT:
      buf.append("DAEMON_FAILURE_EVENT");
      break;
    default:
      buf.append("UNKNOWN EVENT TAG: " + tag);
      break;
    }
    buf.append(", gid=");
    buf.append(gid);
    buf.append("]");
    return buf.toString();
  }

} // END Event
