/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.events;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.MemberId;
import jgroup.relacs.types.MemberIdImpl;



/**
 *  The <code>PrepareAck</code> class represents a local message
 *  exchanged between members and the daemon; it is used by local
 *  members to notify the daemon that it has notified its membership
 *  listener(s) that a new view is to be installed.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since  Jgroup 0.1
 */
public final class PrepareAck
  extends Event
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 6671374717319612210L;

  /** Local member identifier */
  private MemberId memberId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public PrepareAck() {}

  /**
   *  Creates a <code>PrepareAck</code> object for the specified group
   *  identifier.
   */
  public PrepareAck(int gid, MemberId member)
  {
    super(PREPARE_ACK, gid);
    this.memberId = member;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the member identifier for the member acknowledging
   *  the prepare event.
   */
  public MemberId getMemberId()
  {
    return memberId;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    super.readExternal(in);
    memberId = new MemberIdImpl();
    memberId.readExternal(in);
  }
  
  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    super.writeExternal(out);
    memberId.writeExternal(out);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append(super.toString());
    buf.append("[PrepareAck: memberId=");
    buf.append(memberId);
    buf.append("]");
    return buf.toString();
  }

} // END PrepareAck
