/*
 * Copyright (c) 2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


package jgroup.relacs.gm;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MemberTable;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.MembershipListener.AllowDuplicateViews;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.daemon.DaemonInteraction;
import jgroup.relacs.events.DaemonFailureEvent;
import jgroup.relacs.events.Event;
import jgroup.relacs.events.EventTags;
import jgroup.relacs.events.InstallAck;
import jgroup.relacs.events.InstallEvent;
import jgroup.relacs.events.JoinRequest;
import jgroup.relacs.events.LeaveRequest;
import jgroup.relacs.events.MemberLeftEvent;
import jgroup.relacs.events.PrepareAck;
import jgroup.relacs.events.PrepareEvent;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

/**
 *  The <code>MembershipLayer</code> class implements the group membership
 *  service interface and provide membership events to listeners.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 2.2
 */
public final class MembershipLayer
  implements MembershipService, DispatcherListener, EventTags
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private final static Logger log = Logger.getLogger(MembershipLayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Used in status to identify that no group is joined at the moment */ 
  private static final int S_UNUSED = 0;
  
  /** Used in status to identify that the member is joining a group */
  private static final int S_JOINING = 1;
  
  /** Used in status to identify that the member is member of a group */
  private static final int S_NORMAL = 2;
  
  /** Used in status to identify that the member is leaving a group */
  private static final int S_LEAVING = 3;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Information about the group and other objects used to provide group membership
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The lock preventing concurrent event handling */
  private final Lock lock = new ReentrantLock();

  /** Condition for awaiting the group to be joined */
  private final Condition joined = lock.newCondition();

  /** Condition for awaiting the handling of some event */
  private final Condition eventHandlingCompleted = lock.newCondition();

  /** Condition variable for enabling/disabling event handling */
  private boolean handlingEvent = false;

  /** Member identifier */
  private final MemberId me;

  /** Dispatcher layer */
  private final DispatcherService dispatcher;

  /** MemberTable provided to upper layers */
  private final MemberTable membertable;
  
  /**
   *  Dynamic list of membership listeners.  Note that the ordering of
   *  these listeners are important.
   */
  private final List<MembershipListener> membershipListeners =
    new ArrayList<MembershipListener>();

  /** Current status of membership */
  private volatile int status;

  /** Group identifier */
  private int gid;

  /**
   *  Position index of this member in the array containing
   *  the members of the current view.
   */
  private int viewIndex;

  /**
   *  Position index of the local host in the array 
   *  containing the hosts of the current view.
   */
  private int hostIndex;
  
  /**
   *  Position index of this member in the array containing 
   *  the local members of the current view.
   */
  private int memberIndex;

  /**
   *  The number of members of this group.
   */
  private int numMembers;

  /**
   * Field holding the previous view installed; it is used to suppress identical views
   * so that the application do not see them.  Duplicate suppression is the default.
   * If the <code>viewChange()</code> method is annotated with the
   * <code>@AllowDuplicateViews</code> marker, then applications or layers will
   * receive duplicate views.  This is sometimes necessary, when the
   * <code>prepareChange()</code> method is used to block some processing; and the
   * <code>viewChange()</code> is used to unblock things again.
   */
  private View previousView = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize a new <code>BaseLayer</code> object using information
   *  contained in the configuration objects.
   *
   *  @param dispatcher
   *    The dispatcher service.
   *  @exception JgroupException
   *    Raised if the distributed system configuration was not available.
   */
  private MembershipLayer(DispatcherService dispatcher)
    throws JgroupException
  {
    /* Stores initialization information */
    this.dispatcher = dispatcher;

    /*
     * Initialize the Jgroup daemon, either locally or remotely
     * depending on the configuration.
     */
    DaemonInteraction.initDaemon();

    /* Obtain an identifier for this member */
    me = DaemonInteraction.getMemberId();

    /* Initialize local data structures */
    membertable = new MemberTable();
    status = S_UNUSED;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static MembershipLayer getLayer(DispatcherService dispatcher)
    throws JgroupException
  {
    return new MembershipLayer(dispatcher);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a listener object for this layer to provide upcalls to,
   *  in response to membership events.
   * 
   *  @see jgroup.core.Layer#addListener(java.lang.Object)
   */
  public void addListener(Object listener)
  {
    if (listener instanceof MembershipListener && !membershipListeners.contains(listener))
      membershipListeners.add((MembershipListener)listener);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // FinalizeLayer interface methods (invoked to complete the group manager construction)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Post initialization of the <code>MembershipLayer</code>; invoked once
   *  all group manager layers have been constructed.  This method is defined
   *  in the <code>FinalizeLayer</code> inherited through the
   *  <code>MembershipSerivce</code>. <p>
   * 
   *  This method is used to determine if the provided server should join the
   *  group automatically, as soon as the group manager layer stack has been
   *  completed.  By default, automatic join is disabled.  To enable auto join
   *  you must configure it in the <code>applications.xml</code> file. 
   */
  public void complete(Object server)
    throws JgroupException
  {
    /* Check if the server application has auto join enabled */
    AppConfig thisApp = AppConfig.getApplication(server);
    if (thisApp != null && thisApp.getBooleanParam("PGMS.auto")) {
      int gid = thisApp.getGroupId();
      if (log.isDebugEnabled())
        log.debug("Automatic join enabled for group: " + gid);
      /* Join the application group */
      join(gid);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#join()
   */
  public void join()
    throws JgroupException
  {
    join1(-1);
  }

  /* (non-Javadoc)
   *  @see jgroup.core.MembershipService#join(int)
   */
  public void join(int gid)
    throws JgroupException
  {
    if (gid < 0)
      throw new IllegalArgumentException("Negative group identifier: " + gid);
    join1(gid);
  }

  /**
   *  The actual implementation of join.
   *
   *  @param gid
   *    Non-negative integer identifying with the group
   *  @exception JgroupException
   *    Raised if the group manager receives a request to join a group
   *    for which it is already a member.
   */
  private void join1(int gid)
    throws JgroupException
  {
    if (log.isDebugEnabled())
      log.debug("Joining: " + gid);

    // Acquire the lock
    lock.lock();
    try {
      if (status != S_UNUSED)
        throw new JgroupException("Already member of group " + gid);
      this.gid = gid;
      status = S_JOINING;

      dispatcher.start(gid);
      JoinRequest joinRequest = new JoinRequest(gid, me, dispatcher.getRemoteDispatcher());
      DaemonInteraction.addEvent(joinRequest);
      if (log.isDebugEnabled()) {
        log.debug("Sent to Daemon: " + joinRequest);
      }

      /*
       * The join method blocks pending a response from the membership service
       * (the Jgroup daemon).  However, if there is no response within ten
       * seconds, we throw an exception.
       */
      while (status != S_NORMAL) {
        try {
          boolean success = joined.await(10, TimeUnit.SECONDS);
          if (!success)
            throw new JgroupException("Join timeout elapsed for group " + gid);
        } catch (InterruptedException e1) { }
      }
      if (log.isDebugEnabled())
        log.debug("Join completed");
    } finally {
      // Release the lock
      lock.unlock();
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#leave()
   */
  public void leave()
    throws JgroupException
  {
    if (log.isDebugEnabled()) {
      log.debug("leave");
    }
    // Acquire the lock
    lock.lock();
    try {
      if (status == S_UNUSED)
        throw new JgroupException("The gm is not member of any group");
      if (status == S_LEAVING)
        throw new JgroupException("The gm has already received a request to leave the group");
      status = S_LEAVING;
    } finally {
      // Release the lock
      lock.unlock();
    }
    DaemonInteraction.addEvent(new LeaveRequest(gid, me));
    if (log.isDebugEnabled()) {
      log.debug("leave completed");
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#isMember()
   */
  public synchronized boolean isMember()
  {
    return (status == S_NORMAL);
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#isJoining()
   */
  public synchronized boolean isJoining()
  {
    return (status == S_JOINING);
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#isMemberOrJoining()
   */
  public boolean isMemberOrJoining()
  {
    lock.lock();
    try {
      return (status == S_NORMAL || status == S_JOINING);
    } finally {
      lock.unlock();
    }
  }

  // Not accessible through the MembershipService interface (maybe added later)
  public synchronized boolean isUnused()
  {
    return (status == S_UNUSED);
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#isLeader()
   */
  public synchronized boolean isLeader()
  {
    return viewIndex == 0;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#getMyIdentifier()
   */
  public MemberId getMyIdentifier()
  {
    return me;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#getGid()
   */
  public int getGid()
  {
    if (isUnused())
      throw new IllegalStateException("The membership service has not yet joined a group");
    return gid;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#getMemberTable()
   */
  public MemberTable getMemberTable()
  {
    return membertable;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#members()
   */
  public synchronized int members()
  {
    return numMembers;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#getViewIndex()
   */
  public synchronized int getViewIndex()
  {
    return viewIndex;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipService#getMemberIndex()
   */
  public synchronized int getMemberIndex()
  {
    return memberIndex;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DispatcherListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherListener#eventTypes()
   */
  public int[] eventTypes()
  {
    return new int[] {
      INSTALL_EVENT,
      PREPARE_EVENT,
      MEMBER_LEFT_EVENT,
      DAEMON_FAILURE_EVENT
    };
  }


  /* (non-Javadoc)
   * @see jgroup.relacs.gm.DispatcherListener#notifyEvent(jgroup.relacs.events.Event)
   */
  public void notifyEvent(Event event)
  {
    /*
     * Block waiting for completion of the current event, if any.  This is
     * so as to serialize the events received from the DispatcherService.
     * These events are all membership events, and should only be processed
     * in sequence, while however, the events processed by the MulticastLayer
     * is not required to be blocked while completing one of these events.
     */
    lock.lock();
    try {
      while (handlingEvent) {
        try {
          eventHandlingCompleted.await();
        } catch (InterruptedException e) { }
      }
      handlingEvent = true;
    } finally {
      lock.unlock();
    }

    switch (event.getTag()) {
      case INSTALL_EVENT:
        handleInstallEvent((InstallEvent) event);
        break;
      case PREPARE_EVENT:
        handlePrepareEvent((PrepareEvent) event);
        break;
      case MEMBER_LEFT_EVENT:
        handleMemberLeftEvent((MemberLeftEvent) event);
        break;
      case DAEMON_FAILURE_EVENT:
        handleDaemonFailureEvent((DaemonFailureEvent) event);
        break;
    }

    /*
     * Signal the completion of the current event.
     */
    lock.lock();
    try {
      handlingEvent = false;
      eventHandlingCompleted.signal();
    } finally {
      lock.unlock();
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  This method is invoked after a view install event is sent has been
   *  notified by the daemon.  This method is only partially synchronized
   *  for updating the state of this object; it makes calls to application
   *  developed code, and we cannot keep the lock for these calls, as we
   *  have no control over what the application might do.
   */
  private void handleInstallEvent(final InstallEvent event)
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + event.getGid() + "]");
      log.debug("handleInstallEvent gid: " + event.getGid());
    }

    final View view = event.getView();

    if (isMemberOrJoining()) {
      lock.lock();
      try {
        // Update view information while holding the lock
        numMembers = view.size();
        viewIndex = view.getMemberPosition(me);
        // FIXME hostIndex is never used.  Do we need it?
        hostIndex = event.getHostIndex();
        memberIndex = event.getMemberIndex();
      } finally {
        lock.unlock();
      }

      // Update member table
      membertable.viewChange(view);

      notifyViewListeners(view, previousView, membershipListeners);

      // Sends ack to daemon
      InstallAck installAck = new InstallAck(gid, memberIndex);
      DaemonInteraction.addEvent(installAck);

      // A few more things needs to be done while we hold the lock
      lock.lock();
      try {
        previousView = view;
        if (isJoining()) {
          // Notify the member waiting for join to complete
          joined.signal();
        }
        // Now we can enter the normal membership status
        status = S_NORMAL;
      } finally {
        lock.unlock();
      }
    }
  }


  /**
   * Here we check if the new view has the same set of members as the previous view.
   * If the two views have the same set of members, then the default is to suppress
   * calls to the <code>listener.viewChange()</code> methods.  If a particular listener
   * marks the <code>viewChange()</code> method with the <code>@AllowDuplicateViews</code>
   * marker, then that method will still be called even if the new view is a duplicate
   * of the pervious.
   * 
   * Note that, the Daemon layer could perhaps be improved to generate less such
   * identical views, however we cannot exclude them entirely since the agreement
   * protocol will be reexecuted if a member did not yet discover the failure of
   * an excluded member; that is before the conclusion of the agreement protocol.
   * 
   * @param view the new view to be installed
   * @param prevView the previously installed view to compare against to identify duplicate views
   * @param listeners the set of listeners
   */
  static void notifyViewListeners(final View view, final View prevView,
      final List<MembershipListener> listeners)
  {
    if (view.hasSameMembers(prevView)) {
      if (log.isDebugEnabled())
        log.debug("GOT IDENTICAL VIEW");
      for (MembershipListener listener : listeners) {
        if (allowDuplicateViewsFor(listener)) {
          if (log.isDebugEnabled())
            log.debug("DELIVERING IDENTICAL VIEW TO " + listener.getClass().getSimpleName());
          notifyListener(view, listener);
        } else {
          if (log.isDebugEnabled())
            log.debug("DISCARDING IDENTICAL VIEW TO " + listener.getClass().getSimpleName());
        }
      }
    } else {
      for (MembershipListener listener : listeners) {
        notifyListener(view, listener);
      }
    }
  }

  /**
   * Returns true if the given listener should have its <code>viewChange()</code>
   * method called to deliver duplicate views.  If duplicates should be suppressed
   * for this listener, then false is returned.  Suppression is the default.
   */
  private static boolean allowDuplicateViewsFor(MembershipListener listener)
  {
    try {
      Method m = listener.getClass().getMethod("viewChange", new Class[] { View.class } );
      return m.isAnnotationPresent(AllowDuplicateViews.class);
    } catch (NoSuchMethodException e) {
      e.printStackTrace();
      log.error("Cannot happen unless MembershipListener.viewChange() has changed its signature", e);
      return false;
    }
  }

  /**
   * Notify the give membership listener of the given view.
   * 
   * Log any exception trace that may occur when executing code in a
   * <code>viewChange</code> handler method.  Continue to invoke the
   * remaining listeners, if an exception occurred.
   */
  private static void notifyListener(View view, MembershipListener listener)
  {
    try {
      if (log.isDebugEnabled())
        log.debug("notifying: " + listener.getClass().getSimpleName());
      listener.viewChange(view);
    } catch (Exception e) {
      e.printStackTrace();
      log.warn("A viewChange() implementation has thrown an exception", e);
    }
  }

  /**
   *  This is method is invoked after a daemon has sent a notification
   *  for a PREPARE event.  The membership listeners are notified about
   *  this event, and an acknowledgement for it is sent to the daemon.
   *  
   *  This method must not be synchronized since it calls external code
   *  which we do not have control over.  The isMember() and isJoining()
   *  methods however does perform synchronized access to state variables
   *  of this class; only those need to be synchronized.
   */
  private void handlePrepareEvent(PrepareEvent event)
  {
    if (log.isDebugEnabled())
      log.debug("handlePrepareEvent");

    if (isMemberOrJoining()) {
      /*
       * We should not synchronize on calls to external code such as prepareChange()
       * since we have no control over what such code might dream up to do.
       */
      for (MembershipListener listener : membershipListeners) {
        try {
          listener.prepareChange();
        } catch (Exception e) {
          e.printStackTrace();
          log.warn("A prepareChange() implementation has thrown an exception", e);
        }
      }
      // No need to synchronize here either since gid and me never change after join
      DaemonInteraction.addEvent(new PrepareAck(gid, me));
    }
  }


  /**
   *  This method is invoked after a daemon has sent a notification that
   *  the current member has left.  The membership listeners are
   *  notified and the dispatcher service is halted.
   */
  private void handleMemberLeftEvent(MemberLeftEvent event)
  {
    if (log.isDebugEnabled())
      log.debug("handleMemberLeftEvent");
    if (!isUnused()) {
      /*
       * We should not synchronize on calls to external code such as hasLeft()
       * since we have no control over what such code might dream up to do.
       */
      for (MembershipListener listener : membershipListeners) {
        try {
          listener.hasLeft();
        } catch (Exception e) {
          e.printStackTrace();
          log.warn("A hasLeft() implementation has thrown an exception", e);
        }
      }
      lock.lock();
      try {
        // Mark member as unused
        status = S_UNUSED;
      } finally {
        lock.unlock();
      }
    }
    dispatcher.halt();
  }


  /**
   *  This is invoked when this member detects that the Jgroup daemon has failed.
   *  We try to rejoin the group on a restarted daemon.
   */
  private void handleDaemonFailureEvent(DaemonFailureEvent event)
  {
    if (log.isDebugEnabled())
      log.warn("handleDaemonFailureEvent");
    lock.lock();
    try {
      // Mark member as unused
      status = S_UNUSED;
    } finally {
      lock.unlock();
    }
    dispatcher.halt();
    try {
      join(gid);
    } catch (JgroupException e) {
      log.warn("Cannot recover from daemon failure");
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object; currently only the
   *  status and group identifier.
   */
  public String toString() 
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[MembershipLayer: ");
    buf.append("groupId=");
    buf.append(gid);
    buf.append(", status=");
    switch (status) {
      case S_UNUSED:
        buf.append("UNUSED");
        break;
      case S_JOINING:
        buf.append("JOINING");
        break;
      case S_NORMAL:
        buf.append("MEMBER");
        break;
      case S_LEAVING:
        buf.append("LEAVING");
        break;
      default:
        buf.append("INCORRECT STATUS");
        break;
    }
    buf.append("]");
    return buf.toString();
  }

} // END MembershipLayer
