/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import static jgroup.util.log.ViewEvent.Type.Client;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.lang.reflect.Proxy;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import jgroup.core.GroupUnreachableException;
import jgroup.core.MemberId;
import jgroup.core.View;
import jgroup.core.registry.LookupRegistry;
import jgroup.core.registry.RegistryFactory;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.simulator.SocketStatusImpl;
import jgroup.util.log.Eventlogger;
import jgroup.util.log.ViewEvent;
import net.jini.core.constraint.InvocationConstraints;
import net.jini.jeri.ObjectEndpoint;
import net.jini.jeri.OutboundRequest;
import net.jini.jeri.OutboundRequestIterator;

import org.apache.log4j.Logger;

/**
 * Handles identity of groups.
 * Knows endpoints of the current view and the group leader.
 * Responsible for updating itself when needed.
 * 
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public class GroupEndPoint
  implements ObjectEndpoint, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupEndPoint.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 4971189422261856392L;

  /**
   * An object used to obtain information about link connectivity when
   * the partition simulator has been activated.
   */
  private static SocketStatusImpl status = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Contains all members */
  private List<MemberId> members = new ArrayList<MemberId>(10);

  /** The group identifier */
  private int gid;

  /** The service lookup name associated with this group endpoint */
  private String serviceName;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructor for externaliztion.
   */
  public GroupEndPoint()
  {
    /* Check if the partition simulator should be activated in the client JVM. */
    if (Boolean.getBoolean("jgroup.simulator")) {
      status = SocketStatusImpl.instance();
    }
  } 

  /**
   * Creates an empty group endpoint for the given serviceName.
   */
  GroupEndPoint(String serviceName)
  {
    this.serviceName = serviceName;
    this.gid = -1;
  }

  /**
   * Creates a group endpoint with given <code>MemberId</code>.  This 
   * constructor uses the group identifier to obtain the registry service 
   * name.
   */
  GroupEndPoint(int gid, MemberId me)
  {
    this.gid = gid;
    this.serviceName = AppConfig.getApplication(gid).getRegistryName();
    addMember(me);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the group identifier associated with this group endpoint proxy.
   */
  public int getGroupId()
  {
    return gid;
  }

  /**
   * Add <code>MemberId</code>s to the given set of members.
   */
  public void addMembers(Iterator<MemberId> iter)
  {
    while (iter.hasNext())
      addMember(iter.next());
  }

  /**
   * Add the given <code>MemberId</code> to the this 
   * group endpoint.
   */
  public void addMember(MemberId member)
  {
    if (!members.contains(member))
      members.add(member);
  }

  /**
   * Returns an iterator over the set of <code>MemberId</code>s 
   * in this group endpoint.
   */
  public Iterator<MemberId> getMembers()
  {
    return members.iterator();
  }

  /**
   *  Update our list of group member endpoints with the
   *  most recent entries known to the naming service used,
   *  for example the dependable registry or the group enabled
   *  Reggie (greg).
   */
  private void updateGroupEndpoints()
    throws Exception
  {
    Proxy newProxy = null;
    if (gid == 1) {
      /*
       * If the proxy itself is the dependable registry, then we
       * we search the bootstrap registries for new dependable
       * registry instances.
       */
      newProxy = (Proxy) RegistryFactory.getRegistry();
    } else {
      LookupRegistry reg = RegistryFactory.getLookupRegistry();
      newProxy = (Proxy) reg.lookup(serviceName);
    }
    if (log.isDebugEnabled())
      log.debug("Updated proxy for: " + serviceName);
    GroupInvocationHandler newHandler =
      (GroupInvocationHandler) Proxy.getInvocationHandler(newProxy);
    members = newHandler.getGroupEndPoint().members;
    /* Check if the reference list is still empty */
    if (members.isEmpty())
      throw new RemoteException("No members in group associated with: " + serviceName);
    Collections.sort(members);
    if (log.isDebugEnabled())
      log.debug("New member list: " + members);
    if (Eventlogger.ENABLED)
      Eventlogger.logEvent("UpdatedGroupEndpoints: members.size=" + members.size() + ": " + members);
  }

//  private long clientViewId = -1;
//  private boolean first = true;

  /**
   * Set the most recent view identifier for this group.  This method is
   * invoked on the client-side to update the client-side view identifier.
   * This is used to detect changes in the membership, and perform client-side
   * proxy updates.
   * 
   * This method is used only to test the periodic refresh technique, and
   * should not be used to update the group endpoints in a real system.
   */
//  public void notifyView(long serverViewId)
//    throws Exception
//  {
//    if (serverViewId != clientViewId) {
//      if (clientViewId != -1) {
//        if (Debug.MEASURE)
//          Debug.logMem("NewViewDetected view.vid=" + serverViewId);
//        if (log.isDebugEnabled())
//          log.debug("New server-side view detected");
//        if (!first) {
//          new Thread("UpdateGroupEndpoints") {
//            public void run() {
//              try {
//                Thread.sleep(rnd.nextInt(15000));
//                updateGroupEndpoints();
//              } catch (Exception e) {
//              }
//            }
//          }.start();
//        }
//        first = false;
//      }
//      clientViewId = serverViewId;
//    }
//  }

  /**
   * Update the client-side side member list according to the given view.
   * <p>
   * This updating is only done on the client-side in response to receiving
   * a new view from the server-side.
   *
   * @param view the new server-side view; <code>null</code> views will
   *   be ignored.
   */
  public void updateClientView(View view)
  {
    if (view == null)
      return;
    if (Eventlogger.ENABLED)
      Eventlogger.logEvent(new ViewEvent(Client, view));
    members.clear();
    MemberId[] members = view.getMembers();
    for (int i = 0; i < members.length; i++) {
      addMember(members[i]);
    }
    if (log.isDebugEnabled())
      log.debug("ClientViewUpdate: " + members);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ObjectEndpoint
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see net.jini.jeri.ObjectEndpoint#newCall(net.jini.core.constraint.InvocationConstraints)
   */
  public OutboundRequestIterator newCall(final InvocationConstraints constraints)
  {
    throw new UnsupportedOperationException("Invocation constraints not supported");
  }
  
  
  /** 
   * Returns an <code>OutboundRequestIterator</code> to use to send remote call to the 
   * referenced remote object. 
   */
  public OutboundRequestIterator newCall(MethodSemantics semantics) 
  {
    // Get the member containing the endpoint to use
    final MemberId member = selectMember(semantics);
    // Get the request iterator to wrap
    final OutboundRequestIterator iter = member.getTcpEndpoint().newRequest(InvocationConstraints.EMPTY);

    return new OutboundRequestIterator() {

      public boolean hasNext()
      {
        /*
         * Assume that Endpoint returns false if unable to communicate
         * with the server endpoint; i.e. there is no point in retrying.
         */
        boolean hasMore = iter.hasNext();
        if (!hasMore) {
          // Remove failed members from client-side proxy
          members.remove(member);
          if (log.isDebugEnabled())
            log.debug("Removed failed member from client-side proxy: " + member);
        }
        return hasMore;
      }

      public OutboundRequest next()
        throws IOException
      {
        if (log.isDebugEnabled())
          log.debug(GroupEndPoint.this);
        return iter.next();
      }
      
    };
  } 

  /* (non-Javadoc)
   * @see net.jini.jeri.ObjectEndpoint#executeCall(net.jini.jeri.OutboundRequest)
   */
  public RemoteException executeCall(OutboundRequest call)
    throws IOException
  {
    int response = call.getResponseInputStream().read();
    switch (response) {
      case InvocationResult.INVOCATION_COMPLETED:
        return null;

      case InvocationResult.INVOCATION_BLOCKED:
        try {
          Thread.sleep(100);
        } catch (InterruptedException e) { }
        if (log.isDebugEnabled())
          log.debug("Invocation was blocked; retrying on another node.");
        throw new IOException("Invocation was blocked; retrying on another node.");

      case InvocationResult.INVOCATION_FAILED:
        // There was an error at the server-side request handler level
        return new RemoteException("Invocation failed; server-side was unable to handle request");

      case InvocationResult.UNKNOWN_INVOCATION_SEMANTICS:
        return new RemoteException("Unknown invocation semantics");

      default :
        return new RemoteException("Unknown response code: " + response);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for selecting the communication endpoint (the group member)
  ////////////////////////////////////////////////////////////////////////////////////////////

  private MemberId selectMember(MethodSemantics semantics)
  {
    List<MemberId> eps = members;
    Collections.sort(members);
    if (status != null)
      eps = accessibleMembers();
    MemberId member = semantics.selectMember(eps);
    if (member == null) {
      /* Try to refresh the proxy endpoint list */
      try {
        updateGroupEndpoints();
      } catch (Exception e) {
        throw new GroupUnreachableException(
            "unable to update the client-proxy from the naming service", e);
      }
      if (status != null) {
        eps = accessibleMembers();
        member = semantics.selectMember(eps);
      } else
        member = semantics.selectMember(members);
    }
    if (member == null)
      throw new GroupUnreachableException("no remaining members");
    if (log.isDebugEnabled())
      log.debug("Selected member: " + member);
    return member;
  }


  /**
   * Returns a list of <code>MemberId</code>s that is accessible
   * according to the partition simulator.
   */
  private List<MemberId> accessibleMembers()
  {
    List<MemberId> list = new ArrayList<MemberId>(members.size());
    for (Iterator<MemberId> iter = members.iterator(); iter.hasNext();) {
      MemberId member = iter.next();
      String hostname = member.getCanonicalHostName();
      try {
        if (status.isReliable(hostname)) {
          list.add(member);
        } else {
          if (log.isDebugEnabled())
            log.debug("Partition simulator: Broken link to: " + hostname);
        }
      } catch (UnknownHostException e) {
        e.printStackTrace();
      }
    }
    return list;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public int hashCode()
  {
    return members.hashCode();
  }

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[GroupEndPoint: serviceName=");
    buf.append(serviceName);
    buf.append(", members=");
    buf.append(members);
    buf.append("]");
    return buf.toString();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  @SuppressWarnings("unchecked")
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    serviceName = in.readUTF();
    gid = in.readInt();
    members = (List) in.readObject();
  }

  /* (non-Javadoc)
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeUTF(serviceName);
    out.writeInt(gid);
    out.writeObject(members);
  }

} // END GroupEndPoint
