/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.Remote;
import java.util.Collection;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import jgroup.core.ExternalGMIListener;
import jgroup.core.GroupManager;
import jgroup.core.View;
import net.jini.jeri.InboundRequest;
import net.jini.jeri.InvocationDispatcher;

import org.apache.log4j.Logger;

/**
 * Dispatches external invocations to the local group member by unmarshalling
 * the invocation method and parameters.
 * 
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public class GroupInvocationDispatcher
  implements InvocationDispatcher
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(GroupInvocationDispatcher.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Method table for converting hash to method */
  private final MethodTable methodTable;

  /** The group manager */
  private final GroupManager groupManager;

  /** The server-side view */
  private View serverView;

  /**
   * True if invocation requests should be blocked awaiting a new view.
   * Initially, we do not accept any invocations until we have installed a view.
   */
  private volatile boolean blockInvocation = true;

  /** Lock and condition used to enable/disable method dispatching */
  private final Lock lock = new ReentrantLock();
  private final Condition notDispatching = lock.newCondition();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a new invocation dispatcher, using the given table of
   * available server side EGMI methods.
   */
  public GroupInvocationDispatcher(GroupManager gm, MethodTable methodTable)
  {
    this.groupManager = gm;
    this.methodTable = methodTable;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Invokes serialized request on object.
   * 
   * @param stream InputStream to read invocation from
   * @return an <code>InvocationResult</code> instance containing the
   *  invocation result, or null if the execution failed in some way.
   */
  public InvocationResult dispatch(ObjectInputStream in)
  {
    // Unmarshall the method hash
    long hash;
    long clientView;
    try {
      clientView = in.readLong();
      hash = in.readLong();
    } catch (IOException ioex) {
      /* Something went wrong when reading from stream */
      log.warn("Could not read method hash from stream, aborting request!", ioex);
      return null;
    }

    // Lookup the method for the unmarshalled hash
    MethodDetails m = methodTable.get(hash);
    Method method = m.getMethod();
    if (m == null) {
      log.warn("The provided hash does not correspond to a known server-side method: " + hash);
      return null;
    }

    // Set up the params array to hold the unmarshalled method parameters
    Class[] paramTypes = method.getParameterTypes();
    Object[] params = new Object[paramTypes.length];
    try {
      // Unmarshall the method parameters
      for (int i = 0; i < params.length; ++i) {
        params[i] = in.readObject();
        if (params[i] instanceof ExternalGMIListener) {
          /*
           * If the parameter is a egmi remote reference ensure
           * to set the group manager reference on the handler.
           */
          GroupInvocationHandler handler =
            (GroupInvocationHandler) Proxy.getInvocationHandler(params[i]);
          handler.setGroupManager(groupManager);
        }
      }
    } catch (ClassNotFoundException cnfex) {
      log.warn("ClassNotFoundException caught when unmarshalling arguments", cnfex);
      return null;
    } catch (IOException ioex) {
      // Problems reading arguments from stream; could be wrong number of arguments
      log.warn("Could not read arguments from stream, aborting request!", ioex);
      return null;
    }

    // Try to invoke the server whose method is associated with the given hash
    Object server = m.getServer();

    // Do block invocation if we are pending a new view installtion.
//    doBlock(m);

    if (log.isDebugEnabled())
      log.debug("Invoking " + m + " on server " + server.getClass().getSimpleName());
    try {
      return new InvocationResult(method, params, server, serverView, clientView);
    } catch (IllegalAccessException e) {
      log.warn("Tried to invoke the non-public method: " + m, e);
      return null;
    } catch (IllegalArgumentException e) {
      log.warn("An IllegalArgumentException occured while invoking method: " + m, e);
      return null;
    }
  }

  /**
   * Do block invocation if we are pending a new view.
   * 
   * @param m the method which may get blocked
   */
  private void doBlock(MethodDetails m)
  {
    lock.lock();
    try {
      while (blockInvocation) {
        if (log.isDebugEnabled())
          log.debug("Blocking invocation " + m);
        // wait for signal to start dispatching again
        notDispatching.awaitUninterruptibly();
        if (!blockInvocation && log.isDebugEnabled())
          log.debug("Unblocking invocation " + m);
      }
    } finally {
      lock.unlock();
    }
  }

  /**
   * Update local view, and unblock invocation dispatching.
   */
  void enableDispatching()
  {
    if (log.isDebugEnabled())
      log.debug("Enable invocation dispatching");
    lock.lock();
    try {
      // Unblock all pending invocations
      blockInvocation = false;
      notDispatching.signalAll();
    } finally {
      lock.unlock();
    }
  }

  /**
   * To block invocation dispatching during view change periods.
   */
  void disableDispatching()
  {
    if (log.isDebugEnabled())
      log.debug("Disable invocation dispatching");
    lock.lock();
    try {
      // Block new invocations from now on and until enableDispatching is called
      blockInvocation = true;
    } finally {
      lock.unlock();
    }
  }

  /**
   * Set the server-side view for this invocation.
   */
  void setView(View serverView)
  {
    this.serverView = serverView;
  }


  /* (non-Javadoc)
   * @see net.jini.jeri.InvocationDispatcher#dispatch(java.rmi.Remote, net.jini.jeri.InboundRequest, java.util.Collection)
   */
  public void dispatch(Remote impl, InboundRequest request, Collection context)
  {
    throw new UnsupportedOperationException("Should not be called.");
  }

} // END GroupInvocationDispatcher
