/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
package jgroup.relacs.gmi;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException;

import org.apache.log4j.Logger;

import jgroup.core.View;

/**
 * Encapsulates invocation results and the server-side view identifier
 * at the time of result generation.
 * 
 * @author Tor Arve Stangeland
 * @author Hein Meling
 */
public final class InvocationResult
  implements Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(InvocationResult.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -416389148892987660L;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RequestHandler return codes
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Indicates that the invocation completed, and a result will follow. */
  public static final int INVOCATION_COMPLETED         = 0x00;

  /** Indicates that the invocation completed, and a result will follow. */
  public static final int INVOCATION_BLOCKED           = 0x01;

  /** Indicates that the invocation failed with the attached exception. */
  public static final int INVOCATION_FAILED            = 0x02;

  /**
   * Indicates that the invocation could not be performed since it
   * had an unknown invocation semantics for the given request.
   */
  public static final int UNKNOWN_INVOCATION_SEMANTICS = 0x03;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The result of the invocation; it may be an exception. */
  private Object result;

  /** Indicates if the results contains an exception. */
  private boolean isException = false;

  /**
   * The server-side view identifier of the group being invoked
   * (always transmitted)
   */
  private long serverViewId;

  /**
   * The server-side view of the group being invoked
   * (only transmitted when a change has been detected).
   */
  private View serverView = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Empty constructor for externalizable
   */
  public InvocationResult() {} 

  /**
   * Invokes method <code>m</code> on object <code>o</code> with the
   * provided arguments.  The method may return a result which is stored
   * in the <code>result</code> field of this object, or it may return an
   * exception.  The latter is also stored in the <code>result</code>
   * field, but in addition the boolean field <code>isException</code> is
   * set to true.
   * 
   * @param m Method to invoke
   * @param args Arguments to method <code>m</code>
   * @param o Object to invoke method on
   * @param serverView The server-side view in which the method was invoked
   * @param clientViewId The client-side view in which this method was invoked
   */
  public InvocationResult(Method m, Object[] args, Object o,
                          View serverView, long clientViewId)
    throws IllegalArgumentException, IllegalAccessException
  {
    this.serverViewId = serverView.getVid();
    this.serverView = (clientViewId != serverViewId) ? serverView : null;
    try {
      result = m.invoke(o, args);
    } catch (InvocationTargetException itex) {
      // The method threw an exception
      isException = true;
      result = itex.getCause();
      log.error("Invocation threw an exception", itex.getCause());
    }
  }

  public InvocationResult(RemoteException e)
  {
    isException = true;
    result = e;
    log.error("Invocation returned with a remote exception", e);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Client-side access methods (GroupInvocationHandler)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the result of the invocation.
   * If the invoked object threw an exception this exception is thrown.
   */
  Object getResult()
    throws Exception
  {
    if (isException)
      throw (Exception) result;
    return result;
  }

  /**
   * Returns the server-side view identifier attached with this
   * <code>InvocationResult</code> message.
   */
  long getServerViewId()
  {
    return serverViewId;
  }

  /**
   * Returns the server-side view attached with this
   * <code>InvocationResult</code> message.
   */
  View getServerView()
  {
    return serverView;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-javadoc)
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    result = in.readObject();
    isException = in.readBoolean();
    serverViewId = in.readLong();
    boolean containsView = in.readBoolean();
    if (containsView)
      serverView = (View) in.readObject();
  }

  /* (non-javadoc)
   * @see java.io.Externalizable#writeExternal(ObjectOutput)
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeObject(result);
    out.writeBoolean(isException);
    out.writeLong(serverViewId);
    out.writeBoolean(serverView != null);
    if (serverView != null)
      out.writeObject(serverView);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("InvocationResult[");
    if (isException)
      buf.append("Result is an exception, ");
    buf.append("serverViewId=");
    buf.append(serverViewId);
    buf.append(", result=");
    buf.append(result == null ? "null" : result.toString());
    buf.append("]");
    return buf.toString();
  }

} // END InvocationResult
