/*
 * Copyright (c) 1998-2003 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */ 

package jgroup.relacs.gmi;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.rmi.Remote;
import java.rmi.server.ExportException;
import java.util.Collection;

import jgroup.core.ExternalGMIService;
import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.MalformedLayerException;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.View;
import jgroup.core.Layer.FinalizeLayer;
import jgroup.core.multicast.MulticastService;
import jgroup.core.registry.DependableRegistry.RegistryEntry;
import jgroup.relacs.gmi.protocols.BasicDispatcher;
import jgroup.relacs.gmi.protocols.ProtocolDispatcher;
import net.jini.jeri.AbstractILFactory;
import net.jini.jeri.InvocationDispatcher;
import net.jini.jeri.ObjectEndpoint;
import net.jini.jeri.ServerCapabilities;

import org.apache.log4j.Logger;


/**
 *  This layer implements the external group method invocation service
 *  based on JERI. <p>
 * 
 *  It enables a server to listen for group method invocations from
 *  clients external to the object group that the server is a member of
 *  and dispatch these invocations to one or more servers within the group.
 *
 *  @author Tor Arve Stangeland
 *  @author Hein Meling
 *  @since Jgroup 2.1
 */
public class JeriEGMILayer
  extends AbstractILFactory
  implements ExternalGMIService, MembershipListener, FinalizeLayer
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(JeriEGMILayer.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The local group manager */
  private final GroupManager gm;

  /** The group identifier for this group manager stack */
  private final int gid;

  /** Method table for converting hash to method */
  private final MethodTable methodTable = new MethodTable();

  /** Request handler */
  private final GroupRequestHandler requestHandler;

  /** Invocation dispatcher */
  private final GroupInvocationDispatcher dispatcher;
  
  private final MemberId me;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private JeriEGMILayer(GroupManager gm, MulticastService mcast, MembershipService pgms)
  {
    this.gm = gm;
    gid = gm.getGroupId();
    me = pgms.getMyIdentifier();
    dispatcher = new GroupInvocationDispatcher(gm, methodTable);
    requestHandler = new GroupRequestHandler(gid, me);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static JeriEGMILayer getLayer(
      GroupManager gm, MulticastService mcast, MembershipService pgms)
  {
    return new JeriEGMILayer(gm, mcast, pgms);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Layer interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Add a server/layer that is listening for external group method invocations.
   */
  public void addListener(Object listener)
  {
    if (listener instanceof Remote) {
      methodTable.addMethods(listener);
      if (log.isDebugEnabled()) {
        log.debug(listener.getClass().getName() + " implements a Remote interface; "
            + "added remote methods to the method table.");
      }
    } else {
      if (log.isDebugEnabled()) {
        log.debug(listener.getClass().getName() + " does not implement a Remote interface");
      }
    }
  }

  
  /**
   * This method is called by the <code>GroupManager</code> after the construction
   * of all the layers, i.e., all <code>addListener</code> calls have been completed,
   * and the <code>MethodTable</code> is fully populated.
   * 
   * The purpose of this method is to determine which protocols are actually needed
   * by the server (and its required layers) and from that construct the protocol
   * instances and add them to the <code>GroupRequestHandler</code> object.  This
   * method will also pass the server reference to the protocol instances that need
   * it.
   * 
   * @throws JgroupException if a protocol instance could not be constructed,
   * 
   * @see jgroup.core.Layer.FinalizeLayer#complete(java.lang.Object)
   */
  public void complete(Object server)
    throws JgroupException
  {
    for (MethodDetails method : methodTable) {
      MethodSemantics semantics = method.getInvocationSemantics();
      if (!requestHandler.hasProtocol(semantics)) {
        // Get protocol if the request handler does not have it yet
        ProtocolDispatcher protocol = getProtocol(semantics);
        // The protocol implementation may need access to the server object
        protocol.setServer(server);
        requestHandler.addProtocol(semantics, protocol);
      }
    }
  }


  /**
   * Construct a <code>ProtocolDispatcher</code> instance for the given
   * method invocation semantics.
   * 
   * The protocol instances must follow these rules of design:
   * <ol>
   * <li> The protocol implementation must implement the <code>ProtocolDispatcher</code>
   *      interface, or do so implicitly through inheritence.
   * <li> The constructor must provide only a single constructor.
   * <li> The constructor cannot throw checked exceptions.
   * <li> The parameter types of the constructor must correspond to a *Service.class
   *      available from the group manager; that is it must be a layer.  In addition,
   *      one of the parameters may be of type <code>GroupInvocationDispatcher</code>.
   * </ol>
   * 
   * @param semantics the method invocation semantics whose protocol to obtain
   * @return the protocol instance for the provided method invocation semantics.
   * @throws JgroupException if the protocol could not be constructed, one of the above
   *   rules of design was broken.
   */
  public ProtocolDispatcher getProtocol(MethodSemantics semantics)
    throws JgroupException
  {
    Class pClass = semantics.getProtocol();
    if (!ProtocolDispatcher.class.isAssignableFrom(pClass)) {
      throw new MalformedLayerException(
          "Protocol class must implement ProtocolDispatcher: " + pClass);
    }
    Constructor[] constructors = pClass.getDeclaredConstructors();
    if (constructors.length > 1) {
      throw new MalformedLayerException(
          "Protocol cannot provide multiple constructors: " + pClass);
    }
    Constructor constr = constructors[0];
    if (constr.getExceptionTypes().length > 0) {
      throw new MalformedLayerException(
          "Protocol constructor is not allowed to throw checked exceptions: " + pClass);
    }
    Class[] parameters = constr.getParameterTypes();
    Object[] args = new Object[parameters.length];
    if (parameters.length > 0) {
      for (int i = 0; i < parameters.length; i++) {
        if (GroupInvocationDispatcher.class.isAssignableFrom(parameters[i])) {
          args[i] = dispatcher;
        } else {
          args[i] = gm.getService(parameters[i]);
        }
      }
    }

    try {
      return (ProtocolDispatcher) constr.newInstance(args);
    } catch (Exception e) {
      throw new JgroupException(
          "Failed to construct protocol for: " + pClass, e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AbstractILFactory
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /* (non-Javadoc)
   * @see net.jini.jeri.AbstractILFactory#createInvocationHandler(Class[], Remote, ObjectEndpoint)
   */
  public InvocationHandler createInvocationHandler(Class[] interfaces, Remote impl, ObjectEndpoint oe)
    throws ExportException
  {
    return new GroupInvocationHandler(oe, methodTable);
  }
  
  /* (non-Javadoc)
   * @see net.jini.jeri.AbstractILFactory#createInvocationDispatcher(Collection, Remote, ServerCapabilities)
   */
  public InvocationDispatcher createInvocationDispatcher(Collection methods, Remote impl, ServerCapabilities caps)
    throws ExportException
  {
    return dispatcher;
  }

  /**
   * By overriding this method we may avoid patching the Mahalo
   * implementation to support our GroupInvocationHandler implementation.
   */
  protected Class[] getExtraProxyInterfaces(Remote impl)
    throws ExportException
  {
    if (impl == null) {
      throw new NullPointerException("impl is null");
    }
    return new Class[0];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Notify the invocation dispatcher and the request handler
   * of the most recent server-side view identifier.
   * 
   * @see jgroup.core.MembershipListener#viewChange(jgroup.core.View)
   */
  @AllowDuplicateViews public void viewChange(View view)
  {
//    requestHandler.enableDispatching();
//    dispatcher.enableDispatching();
    dispatcher.setView(view);
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#prepareChange()
   */
  public void prepareChange()
  {
//    requestHandler.disableDispatching();
//    dispatcher.disableDispatching();
  }

  /* (non-Javadoc)
   * @see jgroup.core.MembershipListener#hasLeft()
   */
  public void hasLeft() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ExternalGMIService interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public IID getIdentifier()
  {
    return BasicDispatcher.getIdentifier();
  }


  public RegistryEntry getRegistryEntry(String serviceName)
    throws JgroupException
  {
    try {
      requestHandler.initServerEndpoint();
      return new JeriRegistryEntry(serviceName, me, methodTable);
    } catch (IOException ioex) {
      throw new JgroupException("Could not get Endpoint", ioex);
    }
  }


  public BootstrapInvocationHandler getBootstrapHandler()
    throws JgroupException
  {
    try {
      requestHandler.initServerEndpoint();
      GroupEndPoint gep = new GroupEndPoint(gid, me);
      if (log.isDebugEnabled())
        log.debug(gep);
      return new GroupInvocationHandler(gep, methodTable);
    } catch (IOException ioex) {
      throw new JgroupException("Could not create bootstrap invocation handler", ioex);
    }
  }

  /* (non-Javadoc)
   * @see jgroup.core.ExternalGMIService#getProxy(java.rmi.Remote)
   */
  public Remote getProxy(Remote impl)
    throws ExportException
  {
    try {
      requestHandler.initServerEndpoint();
      GroupEndPoint gep = new GroupEndPoint(gid, me);
      if (log.isDebugEnabled())
        log.debug(gep);
      Instances inst = createInstances(impl, gep, requestHandler.getServerEndpoint());
      return inst.getProxy(); 
    } catch (IOException e) {
      throw new ExportException("Could not export remote object", e);
    }
  }

} // END JeriEGMILayer
