/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi;

import static jgroup.relacs.gmi.MethodSemantics.ANYCAST;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.security.AccessController;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivilegedAction;

/**
 * @author Hein Meling
 * @since Jgroup 3.0
 */
public class MethodDetails
  implements Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6799041492486905534L;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Hash identifier of this method */
  private final long hash;

  /** The server object associated with this method (only on server side) */
  private final transient Object server;

  /** The actual method object associated with this method (only on server side) */
  private final transient Method method;

  /** The method semantics for this method */
  private MethodSemantics semantics = null;

  /** Prepared string for toString() to return */
  private final String methString;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Determine the method hash and invocation semantics associated
   * with the given method and store these values.  Also store the
   * associated server object.
   * 
   * Note that the provided method object should be obtained from an
   * interface unless annotations are only declared in the server class.
   * This is since annotations are not inherited from interfaces
   * implemented by the server.
   * 
   * Annotation markers declared in the server class takes precedence over
   * those that <i>may</i> be declared in the interface.
   */
  MethodDetails(final Method method, final Object server)
  {
    /* Store the server object associated with the given method. */
    this.server = server;
    /* Store the method object (used only on the server-side) */
    this.method = method;

    /*
     * Compute the "method hash" of a remote method.  The method hash
     * is a long containing the first 64 bits of the SHA digest from
     * the UTF encoded string of the method name and descriptor.
     */
    hash = computeHash(method);

    /*
     * Note that the following lines may seem unecessary, but the Method
     * object provided to this constructor 'may' be from an interface.
     * We want to obtain the Method object from the server class since we
     * then can support annotation markers directly in the server implementation,
     * and they do not have to appear in the interface.
     */
    Method srvMethod;
    try {
      Class[] params = method.getParameterTypes();
      srvMethod = server.getClass().getMethod(method.getName(), params);
    } catch (NoSuchMethodException e) {
      // Ignored; cannot happen unless things where not compiled correctly.
      throw new Error("Should never happen", e);
    }

    /*
     * Determine the method invocation semantics based on the defined annotations.
     */
    Annotation[] ann = srvMethod.getAnnotations();
    if (ann.length == 0) {
      /*
       * Fallback to the interface method declaration and check if there
       * are annotations declared in the interface instead.
       */
      ann = method.getAnnotations();
    }
    for (Annotation a : ann) {
      MethodSemantics s = MethodSemantics.getInvocationSemantics(a.annotationType());
      if (s != null) {
        if (semantics == null)
          semantics = s;
        else
          throw new Error("Cannot declare multiple semantics for a single method: "
              + method.getName() + " in server: " + server.getClass().getSimpleName());
      }
    }
    // If no annotation marker is given, fallback to default
    if (semantics == null)
      semantics = ANYCAST;

    /* Buffer used to prepare string for this method's details */
    StringBuilder buf = new StringBuilder();
    buf.append(semantics);
    buf.append(" ");
    buf.append(method.getReturnType().getSimpleName());
    buf.append(" ");
    buf.append(method.getName());
    buf.append("(), hash=");
    buf.append(hash);
    methString = buf.toString();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return the hash value associated to this method.
   */
  public long getHash()
  {
    return hash;
  }

  /**
   * Return the server object associated to this method.  This method
   * should only be used on the server-side.
   */
  public Object getServer()
  {
    return server;
  }

  /**
   * Returns the actual method object associated with this method details
   * object.  This method should only be used on the server-side.
   */
  public Method getMethod()
  {
    return method;
  }

  /**
   * Return the invocation semantics associated with this method.
   */
  public MethodSemantics getInvocationSemantics()
  {
    return semantics;
  }

  /**
   * Returns a string representation of this method object. 
   */
  public String toString()
  {
    return methString;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Compute the "method hash" of a remote method.  The method hash
   * is a long containing the first 64 bits of the SHA digest from
   * the UTF encoded string of the method name and descriptor.
   */
  @SuppressWarnings("unchecked")
  static long computeHash(final Method m)
  {
    /*
     * Set this Method object to override language access checks so that
     * the dispatcher can invoke methods from non-public remote interfaces.
     */
    AccessController.doPrivileged(new PrivilegedAction() {
      public Object run() {
        m.setAccessible(true);
        return null;
      }
    });

    long hash = 0;
    ByteArrayOutputStream sink = new ByteArrayOutputStream(127);
    try {
      MessageDigest md = MessageDigest.getInstance("SHA");
      DataOutputStream out = new DataOutputStream(
        new DigestOutputStream(sink, md));

      String s = getMethodNameAndDescriptor(m);

      out.writeUTF(s);

      // use only the first 64 bits of the digest for the hash
      out.flush();
      byte hasharray[] = md.digest();
      for (int i = 0; i < Math.min(8, hasharray.length); i++) {
        hash += ((long) (hasharray[i] & 0xFF)) << (i * 8);
      }
    } catch (IOException ignore) {
      /* can't happen, but be deterministic anyway. */
      hash = -1;
    } catch (NoSuchAlgorithmException complain) {
      throw new SecurityException(complain.getMessage());
    }
    return hash;
  }


  /**
   * Return a string consisting of the given method's name followed by
   * its "method descriptor", as appropriate for use in the computation
   * of the "method hash".
   *
   * See section 4.3.3 of The Java Virtual Machine Specification for
   * the definition of a "method descriptor".
   */
  private static String getMethodNameAndDescriptor(Method m)
  {
    StringBuilder desc = new StringBuilder(m.getName());
    desc.append("(");
    Class[] paramTypes = m.getParameterTypes();
    for (int i = 0; i < paramTypes.length; i++) {
      desc.append(getTypeDescriptor(paramTypes[i]));
    }
    desc.append(")");
    Class returnType = m.getReturnType();
    if (returnType == void.class) { // optimization: handle void here
      desc.append("V");
    } else {
      desc.append(getTypeDescriptor(returnType));
    }
    return desc.toString();
  }

  /**
   * Get the descriptor of a particular type, as appropriate for either
   * a parameter or return type in a method descriptor.
   */
  private static String getTypeDescriptor(Class type)
  {
    if (type.isPrimitive()) {
      if (type == int.class) {
        return "I";
      } else if (type == boolean.class) {
        return "Z";
      } else if (type == byte.class) {
        return "B";
      } else if (type == char.class) {
        return "C";
      } else if (type == short.class) {
        return "S";
      } else if (type == long.class) {
        return "J";
      } else if (type == float.class) {
        return "F";
      } else if (type == double.class) {
        return "D";
      } else if (type == void.class) {
        return "V";
      } else {
        throw new Error("unrecognized primitive type: " + type);
      }
    } else if (type.isArray()) {
      /*
       * According to JLS 20.3.2, the getName() method on Class does
       * return the VM type descriptor format for array classes (only);
       * using that should be quicker than the otherwise obvious code:
       *
       *     return "[" + getTypeDescriptor(type.getComponentType());
       */
      return type.getName().replace('.', '/');
    } else {
      return "L" + type.getName().replace('.', '/') + ";";
    }
  }

} // END MethodDetails
