/*
 * Copyright (c) 1998-2004 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi;

import java.io.Serializable;
import java.util.List;
import java.util.Random;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.protocols.Anycast;
import jgroup.core.protocols.Atomic;
import jgroup.core.protocols.Internal;
import jgroup.core.protocols.Leadercast;
import jgroup.core.protocols.Multicast;

/**
 * Enumeration of the supported method invocation semantics.
 *
 * @author Hein Meling
 * @since Jgroup 3.0
 */
public enum MethodSemantics
  implements Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Enumeration constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  ANYCAST(Anycast.class),

  LEADERCAST(Leadercast.class) {
    /*
     * Returns the leader member always.
     */
    MemberId selectMember(List<MemberId> members)
    {
      if (members.isEmpty())
        return null;
      return members.get(0);
    }
  },

  MULTICAST(Multicast.class),

  ATOMIC(Atomic.class),
  
  INTERNAL(Internal.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator for "load-balancing" */
  private static final Random rnd = new Random();

  /** Array representation of the enumerated values */
  private static MethodSemantics[] values;

  /** The annotation type associated with this invocation semantics object */
  private final Class annotationType;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  private MethodSemantics(Class annotationType)
  {
    this.annotationType = annotationType;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods associated with this enumeration type
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the protocol class object associated with this method semantics
   * (or protocol annotation type).
   * 
   * @throws JgroupException Raised if this protocol has no implementation class.
   */
  public Class getProtocol()
    throws JgroupException
  {
    String protocolName = annotationType.getSimpleName();
    try {
      return Class.forName("jgroup.relacs.gmi.protocols." + protocolName);
    } catch (ClassNotFoundException e) {
      throw new JgroupException("Unsupported protocol: " + protocolName, e);
    }
  }

  /**
   * Returns the <code>MemberId</code> containing the endpoint that the client 
   * should use for communication with the object group.  If the provided list 
   * of members is empty, <code>null</code> may be returned. <p>
   * 
   * By default the member is selected at random.
   * 
   * @param endpoints
   *   The list of available endpoints that the client can choose from.
   */
  MemberId selectMember(List<MemberId> members)
  {
    String sequencerType = System.getProperty("jgroup.egmi.endpointSelectionScheme", "random");
    int size = members.size();
    if (size > 0)
      if (sequencerType.equals("leader"))
        return members.get(0);
      else
        return members.get(rnd.nextInt(size));
    else
      return null;
  }

  /**
   * Returns the invocation semantics constant for the given
   * annotation if such exists.  If the given annotation is unrelated
   * to these constants, <code>null</code> is returned.
   *
   * @param annotation
   *   The annotation for which to obtain a constant if such exists.
   * @return
   *   The constant associated with the given annotation, or <code>null</code>
   *   if no associatation exists to the given annotation.
   */
  static MethodSemantics getInvocationSemantics(Class annotation)
  {
    //FIXME can we not find the MethodSemantics object more elegantly.
    for (MethodSemantics semantics : values()) {
      if (semantics.annotationType.equals(annotation)) {
        return semantics;
      }
    }
    return null;
  }

  /**
   * Returns the <code>MethodSemantics</code> object for the given ordinal.
   */
  public static MethodSemantics valueOf(int ordinal)
  {
    if (values == null)
      values = values();
    return values[ordinal];
  }

} // END MethodSemantics
