/*
 * Copyright (c) 1998-2006 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.gmi.protocols;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import jgroup.core.JgroupException;
import jgroup.core.MemberId;
import jgroup.core.MembershipService;
import jgroup.core.multicast.MulticastListener;
import jgroup.core.multicast.MulticastService;
import jgroup.relacs.gmi.GroupAckListener;
import jgroup.relacs.gmi.GroupInvocationDispatcher;
import jgroup.relacs.gmi.InvocationResult;
import jgroup.relacs.gmi.MethodSemantics;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

/**
 * Multicast represents a protocol that dispatch method invocations
 * received through the local endpoint to all the other members of
 * the group, including the local server.  The result from all members
 * is returned to the local endpoint whom first received the method
 * invocation request, and one of them is selected and passed back to
 * the client.
 *
 * @author Hein Meling <hein.meling@uis.no>
 */
public class Multicast
  extends BasicDispatcher
  implements MulticastListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Multicast.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Protocol name for demux in the multicast layer */
  private static final String PROTOCOL_NAME = MethodSemantics.MULTICAST.toString();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The multicast service interface */
  private MulticastService mcast;

  /** The membership service interface */
  private MembershipService pgms;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Multicast(GroupInvocationDispatcher dispatcher,
      MulticastService mcast, MembershipService pgms)
  {
    super(dispatcher);
    this.mcast = mcast;
    this.pgms = pgms;
    /*
     * We need to listen to multicast events, and since this is
     * not a layer it will not be handled automatically by the
     * group manager construction mechanism.
     */
    mcast.addListener(this);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // ProtocolDispatcher interface (overriding BasicDispatcher)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Handle inbound request to the local endpoint with <i>multicast</i>
   * invocation semantics.
   * 
   * The local endpoint will send a multicast message to all the other group
   * members for method invocation dispatching at each members.  They each
   * return a result to the local endpoint, and one of these values are selected
   * and returned to the client (GroupInvocationHandler).
   * 
   * @see jgroup.relacs.gmi.protocols.ProtocolDispatcher#dispatch(java.io.InputStream)
   */
  public InvocationResult dispatch(InputStream in)
    throws IOException
  {
    // Reduce overhead if group has a single member only
    if (pgms.members() == 1)
      return super.dispatch(in);

    //HACK: buffer the inputstream and mark the current position.
    BufferedInputStream bis = new BufferedInputStream(in);
    bis.mark(50000);
    OutputStream mout = mcast.getMessage(PROTOCOL_NAME);
    // Write request into multicast output stream
    byte[] buf = new byte[1500];
    int bytesRead;
    while ((bytesRead = bis.read(buf)) != -1) {
      mout.write(buf, 0, bytesRead);
    }
    // Do multicast
    GroupAckListener ackListener = new GroupAckListener(this);
    try {
      if (log.isDebugEnabled())
        log.debug("Multicast: multicasting invocation to group members");
      mcast.mcast(mout, ackListener.getRemoteAckListener());
    } catch (JgroupException jex) {
      log.warn("Multicast invocation failed; member is not ready", jex);
      return null;
    }
    // Wait for ack listener
    boolean hasResults = ackListener.pendingCompletionOrTimeout(2);
    if (hasResults) {
      return (InvocationResult) ackListener.getResult();
    } else {
      //HACK: if no results received within timeout value; dispatch only locally.
      bis.reset();
      return super.dispatch(bis);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MulticastListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#getProtocolName()
   */
  public String getProtocolName()
  {
    return PROTOCOL_NAME;
  }

  /**
   * Received a multicast message from a mediating endpoint for invocation
   * dispatching at the local server.
   * 
   * @param msg
   *   Multicast message received
   * @return
   *   The result of the multicast invocation, or <code>null</code>
   *   if the invocation failed.
   * 
   * @see jgroup.core.multicast.MulticastListener#deliverStream(java.io.InputStream, jgroup.core.MemberId, int)
   */
  public Object deliverStream(InputStream msg, MemberId sender, int seqNo)
  {
    if (log.isDebugEnabled()) {
      MDC.put("group", "[Group: " + pgms.getGid() + "]");
      log.debug("deliverStream() start");
    }
    try {
      // Perform the actual method invocation on the local server (BasicDispatcher)
      return super.dispatch(msg);
    } catch (IOException e) {
      log.warn("Failed to dispatch multicast invocation", e);
    }
    return null;
  }

  /* (non-Javadoc)
   * @see jgroup.core.multicast.MulticastListener#deliverObject(java.lang.Object, jgroup.core.MemberId, int)
   */
  public Object deliverObject(Object msg, MemberId sender, int seqNo)
  {
    throw new UnsupportedOperationException();
  }

} // END Multicast
