/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import jgroup.core.ConfigurationException;
import jgroup.core.EndPoint;
import jgroup.relacs.config.Domain;
import jgroup.relacs.config.TransportConfig;

import org.apache.log4j.Logger;


/**
 *  The <code> Cluster </code> class
 *
 *  FIXME HEIN: This class should contain a SortedMap of members in the
 *  cluster, sorted according to each members reachability value.  in
 *  addition there should be a sub-view of this map, reflecting those
 *  deemed as reachable.
 *
 *  @author Salvatore Camarrata
 *  @author Hein Meling
 *  @since Jgroup 0.5
 */
final class Cluster
  implements MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(Cluster.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random generator */
  private static final Random rnd = new Random();

  /**
   *  Constant null array to be returned for the case of no reachable
   *  hosts (object reuse).  See Item 27 in Effective Java.
   */
  private final static MssHost[] NULL_MSS_HOST_ARRAY = new MssHost[0];


  ///////////////////////////////////////////////////////////////////////
  // Fields
  ///////////////////////////////////////////////////////////////////////

  /** The distributed system configuration */
  private MssDS mssds;

  /** Configuration parameters */
  private TransportConfig config;

  /** EndPoints */
  private RoutingTableEntry rt;

  /** Data link layer */
  private NI ni;

  /** Hosts that are member of the cluster */
  private final List<MssHost> members;

  /** True if this object is the local cluster */
  private boolean local;

  /** Number of members in the cluster that are reachable */
  private int reachableCounter;

  /** Message Flow Layer */
  private MsgFlowSndrSide msgFlow;

  private MsgCntrl msgCntrl;

  /**
   *  Message receivers associated with this cluster, when the message
   *  is not addressed to all members of the cluster.
   */
  private final List<EndPoint> receivers;


  ///////////////////////////////////////////////////////////////////////
  // Constructors
  ///////////////////////////////////////////////////////////////////////

  /**
   *  Builds a cluster identified by <code>domain</code>.
   */
  Cluster(MssDS mssds, TransportConfig config, Domain domain, NI ni)
  {
    /* Copy initialization parameters */
    this.mssds = mssds;
    this.ni = ni;
    this.config = config;
    local = domain.isLocal();
    reachableCounter = (local ? 1 : 0);
    int cost = (local ? 0 : config.getMaxPathLength());

    /*
     * Note that when marshalling the routing table entry, it will use
     * the marshalling of <code>EndPointImpl</code> since currently the
     * <code>Domain</code> class do not need any marshalling.  Further,
     * note that EndPointImpl do not detect local clusters correctly.
     */
    rt = new RoutingTableEntry(domain.getEndpoint(), domain.getEndpoint(), cost);

    /* Initialize list of members */
    members = new ArrayList<MssHost>(domain.size());
    receivers = new ArrayList<EndPoint>(domain.size());

    /* Flow msg management */
    msgFlow= new MsgFlowSndrSide(this, config);

    if (log.isDebugEnabled())
      log.debug("Cluster:<init>: " + rt.key);
  }


  ///////////////////////////////////////////////////////////////////////
  // Public methods
  ///////////////////////////////////////////////////////////////////////

  /**
   *  Returns the endpoint associated to this cluster.
   */
  EndPoint getEndPoint()
  {
    return rt.key;
  }


  String getName()
  {
    return rt.key.getAddress().getHostName();
  }


  MsgCntrl getControl()
  {
    return msgCntrl;
  }


  void setControl(MsgCntrl msgCntrl)
  {
    this.msgCntrl = msgCntrl;
  }


  /**
   *  
   */
  RoutingTableEntry getRoutingEntry()
  {
    return rt;
  }


  /**
   *  Reset the routing entry for this cluster.  This should only be used
   *  for the local cluster. <p>
   *
   *  It is used by the clustertable update method, and is required to
   *  ensure that remote failures are detected.
   */
  void resetRoute()
  {
    rt.cost = 0;
    rt.TTL = config.getMaxTTL();
    rt.route = rt.key;
  }


  /**
   *  Returns true if this is the local cluster.
   */
  boolean isLocal()
  {
    return local;
  }


  /**
   *  Returns the <CODE>MsgFlowSndrSide</CODE> object associated to
   *  this host.
   */
  MsgFlowSndrSide getMsgFlow()
  {
    return msgFlow;
  }


  /**
   *  Returns the number of hosts in this cluster.
   */
  int size()
  {
    return members.size();
  }


  /**
   *  Returns and iterator over the hosts in this cluster.
   */
  Iterator<MssHost> iterator()
  {
    return members.iterator();
  }


  /**
   *
   */
  void insertMember(MssHost member)
    throws ConfigurationException
  {
    if (members.contains(member))
      throw new ConfigurationException("Trying to insert a previously inserted member in cluster");
    else {
      member.setClusterIndex(members.size());
      members.add(member);
    }
  }


  /**
   *  Returns the endpoint of a host included in this cluster whose
   *  reachability value is greater than the given threshold; returns
   *  the cluster endpoint if no hosts in this cluster has a
   *  reachability value above the given threshold.
   *
   *  The leader is selected randomly among the endpoints in the cluster
   *  whose reachability value is greater than the threshold.
   *
   *  FIXME POSSIBLE OPTIMIZATION: If the hosts were stored in a
   *  SortedMap (on a more permanent basis), we could use the sorted
   *  order to determine the range from which the random number should
   *  be picked.
   */
  EndPoint getLeaderEndPoint(int threshold)
  {
    int size = size();
    List<EndPoint> candidates = new ArrayList<EndPoint>(size);
    for (MssHost candidate : members) {
      if (candidate.isReachable(threshold))
        candidates.add(candidate.getEndPoint());
    }
    return (candidates.isEmpty() ? getEndPoint()
            : (EndPoint) candidates.get(rnd.nextInt(candidates.size())));
  }


  /**
   *  Returns the endpoint of a host included in this cluster whose
   *  reachability value is greater than threshold obtained from the
   *  transport configuration object; returns the cluster endpoint if no
   *  hosts in this cluster has a reachability value above the
   *  associated threshold.
   */
  EndPoint getLeaderEndPoint()
  {
    return getLeaderEndPoint(config.getReachabilityThreshold());
  }


  /**
   *  Set all reachable members of the cluster to the given reachability
   *  warning level, and check if there are unreachable members.
   */
  void setWarning(int TTLwarning)
  {
    for (MssHost host : members) {
      if (host.isReachable(TTLwarning)) {
        host.pingOK(TTLwarning);
      } else if (!host.pingKO() && host.wasReachable()) {
        /* If the host was reachable, but is no longer reachable */
        mssds.setAsUnreachable(host);
      }
    }
  }


  /**
   *  Check for previously reachable members of the cluster, and mark
   *  them as unreachable.
   */
  void updateReachability()
  {
    for (MssHost host : members) {
      if (!host.isLocal() && host.isReachable()) {
        host.printReachParams();
      }
      if (!host.isLocal() && !host.pingKO() && host.wasReachable()) {
        /* If the member was reachable, but is no longer reachable */
        mssds.setAsUnreachable(host);
      }
    }
  }


  /**
   *  Mark all reachable members of the cluster as unreachable, and
   *  check if there are new unreachable members.
   */
  void setClusterAsUnreachable()
  {
    for (MssHost host : members) {
      if (host.wasReachable()) {
        /* If the member was reachable, mark it as unreachable */
        mssds.setAsUnreachable(host);
      }
    }
  }


  MssHost[] getReachable(int threshold)
  {
    int size = members.size();
    List<MssHost> reachable = new ArrayList<MssHost>(size);
    for (MssHost host : members) {
      if (host.isReachable(threshold))
        reachable.add(host);
    }
    return (MssHost[]) reachable.toArray(NULL_MSS_HOST_ARRAY);
  }


  MssHost[] getReachable()
  {
    return getReachable(MINIMUM_ALIVE_VALUE);   
  }


  void incrementReachableCounter()
  {
    reachableCounter++;
  }


  void decrementReachableCounter()
  {
    reachableCounter--;
  }
  

  int getReachableCounter()
  {
    return reachableCounter;
  }
  
  
  boolean isReachable()
  {
    return (reachableCounter != 0);
  }


  /**
   *  Returns true if the local cluster is directly connected to this
   *  cluster.  False is returned if there is no direct connection, and
   *  messages to this cluster (from the local cluster) must be routed
   *  through another cluster.
   */
  boolean directlyConnected()
  {
    return (rt.key.equals(rt.route));
  }


  /**
   *  Returns the cluster that we need to route through to get to this
   *  cluster from the local cluster.
   */
  private Cluster getRouteCluster()
  {
    return mssds.clusterLookup(rt.route);
  }


  /**
   *  Reset the sender side message flow for this cluster.  This method
   *  should be invoked only during initialization.
   */
  void resetMsgFlow()
  {
    msgFlow.reset(rnd.nextInt());
  }


  /**
   *  Reset the sender side message flow for this cluster to the given
   *  value.  This is used to indicate that this cluster is unreachable.
   */
  void resetMsgFlow(int mid)
  {
    msgFlow.reset(mid);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Receiver management
  ////////////////////////////////////////////////////////////////////////////////////////////

  void addReceiver(EndPoint member)
  {
    receivers.add(member);
  }

  void clearReceivers()
  {
    receivers.clear();
  }

  boolean hasReceivers()
  {
    return !(receivers.isEmpty());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Cluster sending methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Forward a raw message fragment to this cluster.
   */
  void forward(byte[] buffer, int bufLen)
  {
    if (local) {

      /* If it is my local cluster (LAN) */

      ni.send(buffer, bufLen);
      if (log.isDebugEnabled())
        log.debug("Forwarded fragment to local cluster");

    } else {

      /*
       * Sending to a remote cluster, preferably through the cluster
       * leader; otherwise, we send directly to each member of the
       * remote cluster.
       */
      EndPoint leader = getLeaderEndPoint();
      if (!leader.isMulticastEndPoint()) {

        /*
         * If there is a reachable leader member for this cluster, then
         * send the message directly to the leader.  The leader will
         * broadcast the message within its local cluster to the other
         * cluster members.
         */
        ni.send(leader, buffer, bufLen);
        if (log.isDebugEnabled())
          log.debug("Forwarded fragment to remote cluster through leader: " + leader);

      } else {

        /*
         * Set the receiver list to all members, unless the message is
         * associated with a set of receivers.
         */
        if (receivers.isEmpty()) {
          /* Sending to all members in cluster */
          for (MssHost host : members) {
            ni.send(host.getEndPoint(), buffer, bufLen);
            if (log.isDebugEnabled())
              log.debug("Forwarded fragment to receiver " + host);
          }
        } else {
          /* Sending to receivers in cluster */
          for (EndPoint recv : receivers) {
            ni.send(recv, buffer, bufLen);
            if (log.isDebugEnabled())
              log.debug("Forwarded fragment to receiver " + recv);
          }
        }
      }
    }
  }


  /**
   *  Method for resending Jgroup related messages.  This involves
   *  setting the cluster, receiver set, and flow control information
   *  for this cluster.  The actual Jgroup level message remains the
   *  same; we only changed the cluster, flow control data and receiver
   *  set, in addition to the fragment identifier in the case the
   *  original message was sent to multiple clusters (each with
   *  different fragId sequence numbers).
   */
  void resend(FragmentIterator fragIter, int fragId)
  {
    MsgJG msgjg = (MsgJG) fragIter.getMsg();
    EndPoint[] rcvrs = msgjg.getReceivers();

    /*
     * Add receivers in the destination cluster to the receiver set of
     * the cluster.
     */
    for (int i = 0; i < rcvrs.length; i++) {
      MssHost host = mssds.hostLookup(rcvrs[i]);
      if (!rcvrs[i].isLocal() && host.isIn(this)) {
        addReceiver(rcvrs[i]);
        if (log.isDebugEnabled()) {
          log.debug("Added host to the destination list: " + host);
        }
      } else if (log.isDebugEnabled()) {
        log.debug("Removed host from the destination list: " + host);
      }
    }

    /*
     * Update the cluster header field and the receiver set of the
     * message.
     */
    msgjg.setCluster(this);
    msgjg.setReceivers(receivers);

    /*
     * Reset the fragment identifier, in case it was changed due to
     * sending the same fragment to different clusters (with different
     * fragment id sequences).
     */
    byte[] msgFragment = fragIter.getFragment(fragId);
    FragmentHeader.marshal(false, fragId, msgFragment);

    /*
     * At this point the message should be updated again, allowing us to
     * send a correct fragment.
     */
    forward(msgFragment, fragIter.fragmentLength());
  }


  /**
   *  Method for sending Jgroup related messages.  This involves setting
   *  the cluster, receiver set, and flow control information for this
   *  cluster.
   */
  void send(MsgJG msgjg)
  {
    /*
     * Update the cluster header field and the receiver set of the
     * message.
     */
    msgjg.setCluster(this);
    msgjg.setReceivers(receivers);

    if (ROUTING_ENABLED) {

      /*
       * Check if the message has to be routed (that is, this cluster is
       * not directly connected to the local cluster)
       */
      if (!directlyConnected()) {
        if (log.isDebugEnabled()) {
          log.debug("Stack trace", new Exception());
          log.debug("localCluster=" + ClusterTable.getLocalCluster() + ", Route=" + rt);
        }

        /*
         * The message must be forwarded; simply forward it to the next
         * cluster in route for the destination (this) cluster.
         */
        getRouteCluster().send((Msg) msgjg);
      }

    } else {

      send((Msg) msgjg);

    }
  }


  /**
   *  Send a generic mss level message to this cluster.
   *
   *  Note that this method used to be synchronzied.  However, I believe
   *  that this is unecessary since the fragment iteration loops will have
   *  all their data in the local stack frame of the current invocation.
   *  Only those methods whom modify the object state needs to be synchronized.
   */
  void send(Msg msg)
  {
    /* Send the msg directly to the cluster */

    if (local) {

      /* If it is my local cluster (LAN) */

      for (FragmentIterator frag = msg.iterator(msgCntrl); frag.hasNext(); ) {
        /* Generate and send a fragment */
        byte[] buffer = frag.next(false);
        int bufLen = frag.fragmentLength();

        if (log.isDebugEnabled())
          log.debug("Sending msg fragment (" + frag.getFid() + ") to cluster " + rt.key);
        ni.send(buffer, bufLen);
      }

    } else {

      /*
       * Sending to a remote cluster, preferably through the cluster
       * leader; otherwise, we send directly to each member of the
       * remote cluster.
       */

      EndPoint leader = getLeaderEndPoint();
      if (!leader.isMulticastEndPoint()) {

        /*
         * If there is a reachable leader member for this cluster, then
         * send the message directly to the leader.  The leader will
         * broadcast the message within its cluster to the other cluster
         * members.
         */

        for (FragmentIterator frag = msg.iterator(msgCntrl); frag.hasNext(); ) {
          /* Generate and send a fragment */
          byte[] buffer = frag.next(true);
          int bufLen = frag.fragmentLength();

          if (log.isDebugEnabled())
            log.debug("Sending msg fragment (" + frag.getFid() + ") to cluster leader " + leader);
          ni.send(leader, buffer, bufLen);
        }

      } else {

        /*
         * Otherwise send the message to all the members of the
         * cluster; alternatively, if only a subset of the cluster
         * members are receivers, only those will receive the message.
         */

        for (FragmentIterator frag = msg.iterator(msgCntrl); frag.hasNext(); ) {
          /* Generate and send a fragment */
          byte[] buffer = frag.next(false);
          int bufLen = frag.fragmentLength();

          /*
           * Set the receiver list to all members, unless the message is
           * associated with a set of receivers.
           */
          if (receivers.isEmpty()) {
            /* Sending to all members in cluster */
            for (MssHost host : members) {
              if (log.isDebugEnabled())
                log.debug("Sending msg fragment (" + frag.getFid() + ") to member " + host);
              ni.send(host.getEndPoint(), buffer, bufLen);
            }
          } else {
            /* Sending to receivers in cluster */
            for (EndPoint recv : receivers) {
              if (log.isDebugEnabled())
                log.debug("Sending msg fragment (" + frag.getFid() + ") to receiver " + recv);
              ni.send(recv, buffer, bufLen);
            }
          }

        } // end for-loop frag iterator
      } // end if-else leader mcast endpoint
    } // end if-else local/remote cluster
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the hash code for this object.
   */
  public int hashCode()
  {
    return rt.key.hashCode();
  }


  /**
   *  Returns a string representation of this object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[Cluster: ");
    buf.append(rt.key);
    buf.append("]");
    return buf.toString();
  }


  /**
   *  Compares two objects for content equality.
   *
   *  @param obj
   *    The object to compare this object with.
   *  @return 
   *    True if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof Cluster))
      return false;
    return this.rt.equals(((Cluster) obj).rt);
  }

} // END Cluster
