/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import jgroup.core.EndPoint;
import jgroup.util.BooleanSet;


/**
 *  The <code>DSView</code> class describe the various member sets.  It
 *  includes the complete reachable set, the new reachable set, the new
 *  unreachable set, and the set of new incarnations associated with
 *  this view.
 *
 *  @author Salvatore Cammarata
 *  @since Jgroup 1.2
 */
final class DSView
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Hosttable */
  private HostTable hosttable;
  
  /** Boolean set of the reachable members */
  private BooleanSet reachableSet;    
  
  /** Boolean set of the new unreachable members */
  private BooleanSet newUnreachableSet;
  
  /** Boolean set of the new reachable members */
  private BooleanSet newReachableSet; 
  
  /** Boolean set of the member with new incarnation */
  private BooleanSet newIncarnationSet; 


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  DSView(HostTable hosttable, MssHost localhost)
  {
    /* Copy references */
    this.hosttable = hosttable;
    
    /* Instantiates bit sets */
    reachableSet      = new BooleanSet(hosttable.size());
    reachableSet.set(localhost.getIndex());
    newUnreachableSet = new BooleanSet(hosttable.size());
    newReachableSet   = new BooleanSet(hosttable.size());   
    newIncarnationSet = new BooleanSet(hosttable.size()); 
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Set host member as unreachable 
   */
  void setAsUnreachable(MssHost member)
  {
    reachableSet.unset(member.getIndex());
    newReachableSet.unset(member.getIndex());
    newUnreachableSet.set(member.getIndex());
    newIncarnationSet.unset(member.getIndex());
  }
  
  /**
   *  Set host member as reachable 
   */
  void setAsReachable(MssHost member)
  {
    reachableSet.set(member.getIndex());
    newReachableSet.set(member.getIndex());
    newUnreachableSet.unset(member.getIndex());
  }

  /**
   *  Set host member as having a new incarnation
   */
  void setNewIncarnation(MssHost member)
  {
    newIncarnationSet.set(member.getIndex());
  }

  /**
   *  Clear the set of new reachable, new unreachable and new
   *  incarnation hosts.
   */
  void clear()
  {
    newReachableSet.clear();
    newUnreachableSet.clear();
    newIncarnationSet.clear();
  }

  /**
   *  Returns true if the current view contains new reachable hosts
   *  or new unreachable hosts.
   */
  boolean hasChanged()
  {
    return (newReachableSet.size() > 0) || (newUnreachableSet.size() > 0);
  }
 
  /**
   *  Return the array of reachable hosts 
   */
  MssHost[] getReachableHosts()
  {
    return bitToHosts(reachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  MssHost[] getNewReachableHosts()
  {
    return bitToHosts(newReachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  MssHost[] getNewUnreachableHosts()
  {
    return bitToHosts(newUnreachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  MssHost[] getNewIncarnationHosts()
  {
    return bitToHosts(newIncarnationSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  EndPoint[] getReachableEndPoints()
  {
    return bitToEndPoints(reachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  EndPoint[] getNewReachableEndPoints()
  {
    return bitToEndPoints(newReachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  EndPoint[] getNewUnreachableEndPoints()
  {
    return bitToEndPoints(newUnreachableSet);
  }
  
  /**
   *  Return the array of reachable hosts 
   */
  EndPoint[] getNewIncarnationEndPoints()
  {
    return bitToEndPoints(newIncarnationSet);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private EndPoint[] bitToEndPoints(BooleanSet bitSet) 
  { 
    EndPoint[] members = new EndPoint[bitSet.size()];
    for (int i=0, j=0; i < bitSet.lastIndex(); i++) {
      if (bitSet.lookup(i)) 
        members[j++] = hosttable.get(i).getEndPoint();
    }
    return members;
  }
  
  private MssHost[] bitToHosts(BooleanSet bitSet) 
  { 
    MssHost[] members = new MssHost[bitSet.size()];
    for (int i=0, j=0; i < bitSet.lastIndex(); i++) {
      if (bitSet.lookup(i)) 
        members[j++] = hosttable.get(i);
    }
    return members;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[ReachableSet: ");
    boolean first = true;
    for (int i = 0; i < reachableSet.lastIndex(); i++) {
      if (reachableSet.lookup(i)) {
        if (first) {
          first = false;
        } else {
          buf.append(", ");
        }
        buf.append(hosttable.get(i));
      }
    }
    buf.append("]");
    return buf.toString();
  }

} // END DSView
