/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.core.ConfigManager;
import jgroup.core.EndPoint;
import jgroup.core.JgroupException;
import jgroup.relacs.types.EndPointImpl;
import jgroup.relacs.types.Flag;
import jgroup.relacs.types.IncarnationId;
import jgroup.relacs.types.MessageId;
import jgroup.relacs.types.MessageTag;
import jgroup.util.InMessage;
import jgroup.util.Util;

/**
 *  This class contains the <i>mss</i> level fragment header data.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class FragmentHeader
  implements MssTag, MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size of this object in bytes */
  private static final int SIZE = HEADER_SIZE;

  /** The position in the msg fragment where the tag byte can be set */
  private static final short TAG_POSITION = 0;

  /** The position in the msg fragment where the broadcast bit can be set */
  private static final short BROADCAST_POSITION = 1;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Header Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The tag associated with this message fragment */
  private byte tag;

  /** True if the message fragment has to be broadcast in the cluster */
  private boolean toBeBroadcast;

  /** The fragment identifier */
  int fragId;

  /** The message sender (only the endpoint is transmitted, thus this field is transient) */
  private transient MssHost sender;

  /** The incarnation identifier */
  private int incarnationId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The message fragment associated with this header */
  private transient byte[] msgFragment;

  /** The length of the fragment associated with this header */
  private transient int fragLen;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshalling / Unmarshalling methods and Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshals the given <code>FragmentHeader</code> values into a byte
   *  array and returns it.  The size of the array is equal to the
   *  fragment header standard size.  This method is useful if we want
   *  to send a single message fragment, in which only the tag reveals
   *  the semantics of the message.  For example to notify others of
   *  some event, that does not need additional data.
   *
   *  At the sender side, we don't keep any of the input parameters, we
   *  just marshal it directly into the created <code>outstream</code>
   *  byte array that is returned from this method.
   */
  public static byte[] marshal(byte tag, boolean broadcast, int fragId)
  {
    byte[] outstream = new byte[SIZE];
    marshal(tag, broadcast, fragId, outstream);
    return outstream;
  }


  /**
   *  Marshals the given <code>FragmentHeader</code> values into the
   *  specified <code>byte[] outstream</code>, without creating an
   *  object.  This marshal method does not change the sender related
   *  fields, such as the inet address, port and incarnation identifier.
   *  It will only set the broadcast and fragment identifier values. <p>
   *
   *  At the sender side, we don't keep any of the input parameters, we
   *  just marshal it directly into the <code>outstream</code> byte
   *  array.
   */
  public static void marshal(boolean broadcast, int fragId, byte[] outstream)
  {
    int offset = BROADCAST_POSITION;
    outstream[offset++] = (byte) (broadcast ? 1 : 0);

    outstream[offset++] = (byte) (fragId >>> 24);
    outstream[offset++] = (byte) (fragId >>> 16);
    outstream[offset++] = (byte) (fragId >>>  8);
    outstream[offset]   = (byte) (fragId >>>  0);
  }


  /**
   *  Marshals the given <code>FragmentHeader</code> values into the
   *  specified <code>byte[] outstream</code>, without creating an
   *  object. <p>
   *
   *  At the sender side, we don't keep any of the input parameters, we
   *  just marshal it directly into the <code>outstream</code> byte
   *  array.
   */
  public static void marshal(byte tag, boolean broadcast, int fragId, byte[] outstream)
  {
    /* At the sender side the local host is equal to the sender */
    MssHost sender    = HostTable.getLocalHost();
    int incId         = sender.getIncarnationId();
    EndPoint endpoint = sender.getEndPoint();
    int intAddress    = endpoint.getIntAddress();
    int port          = endpoint.getPort();

    int offset = TAG_POSITION;
    outstream[offset++] = tag;

    outstream[offset++] = (byte) (broadcast ? 1 : 0);

    outstream[offset++] = (byte) (fragId >>> 24);
    outstream[offset++] = (byte) (fragId >>> 16);
    outstream[offset++] = (byte) (fragId >>>  8);
    outstream[offset++] = (byte) (fragId >>>  0);

    outstream[offset++] = (byte) (intAddress >>> 24);
    outstream[offset++] = (byte) (intAddress >>> 16);
    outstream[offset++] = (byte) (intAddress >>>  8);
    outstream[offset++] = (byte) (intAddress >>>  0);

    outstream[offset++] = (byte) (port >>>  8);
    outstream[offset++] = (byte) (port >>>  0);

    outstream[offset++] = (byte) (incId >>> 24);
    outstream[offset++] = (byte) (incId >>> 16);
    outstream[offset++] = (byte) (incId >>>  8);
    outstream[offset]   = (byte) (incId >>>  0);
  }


  public static FragmentHeader unmarshal(MssDS mssds, byte[] instream, int fragLen)
    throws IOException, ClassNotFoundException, JgroupException
  {
    return new FragmentHeader(mssds, instream, fragLen);
  }


  private FragmentHeader(MssDS mssds, byte[] instream, int fragLen)
    throws IOException, ClassNotFoundException, JgroupException
  {
    /* Keep the message fragment for later reference */
    this.msgFragment = instream;
    this.fragLen = fragLen;
    InMessage inmsg = new InMessage(SIZE, 0, 0, SIZE);
    inmsg.insert(instream, SIZE);
    tag = MessageTag.unmarshal(inmsg);
    toBeBroadcast = Flag.unmarshal(inmsg);
    fragId = MessageId.unmarshal(inmsg);
    EndPoint endpoint = new EndPointImpl();
    endpoint.readExternal(inmsg);
    sender = mssds.hostLookup(endpoint);
    if (sender == null) {
      throw new JgroupException("Received a packet from an unknown host: " + endpoint);
    }
    incarnationId = IncarnationId.unmarshal(inmsg);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Access methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated with this message.
   */
  public byte getTag() 
  {
    return tag;
  }


  /**
   *  Set the tag associated with this message.
   */
  public void setTag(byte tag)
  {
    this.tag = tag;
    msgFragment[TAG_POSITION] = tag;
  }


  /**
   *  Returns the message fragment associated with this header.
   */
  public byte[] getFragment() 
  {
    return msgFragment;
  }


  /**
   *  Returns the length of the fragment associated with this header.
   */
  public int getFragmentLength() 
  {
    return fragLen;
  }


  /**
   *  Returns the sender of this message fragment.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns incarnation identifier for the sender of this message.
   */
  public int getIncarnationId()
  {
    return incarnationId;
  }


  /**
   *  Returns true if the sender of this message is the local host.
   */
  public boolean isLocal()
  {
    return sender.isLocal();
  }


  /**
   *  Returns true if this message is a broadcast message.  False is
   *  returned otherwise.
   */
  public boolean isBroadcast()
  {
    return toBeBroadcast;
  }


  public void broadcastOff()
  {
    toBeBroadcast = false;
    msgFragment[BROADCAST_POSITION] = 0;
  }


  /**
   *  Returns true if this message represents the last fragment of a
   *  message flow.
   */
  boolean isLastFragment()
  {
    return (tag != NOTLASTFRAGMENT);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Print the content of the <code>FragmentHeader</code>.
   */
  public String toString() {
    StringBuilder buf = new StringBuilder("[FragmentHeader: tag=");
    switch (tag) {
    case IAMALIVE:
      buf.append("IAMALIVE");
      break;
    case SENT:
      buf.append("SENT");
      break;
    case ROUTING:
      buf.append("ROUTING");
      break;
    case FWDROUTING:
      buf.append("FWDROUTING");
      break;
    case SYN:
      buf.append("SYN");
      break;
    case NACK:
      buf.append("NACK");
      break;
    case JG:
      buf.append("JG");
      break;
    case NOTLASTFRAGMENT:
      buf.append("NLF");
      break;
    default:
      buf.append("Unknown tag=");
      buf.append(tag);
      break;
    }
    buf.append(", fragId=");
    buf.append(fragId);
    buf.append(", sender=");
    buf.append(sender);
    buf.append(", incId=");
    buf.append(incarnationId);
    buf.append(", fragSize=");
    buf.append(fragLen);
    buf.append(", bcast=");
    buf.append(toBeBroadcast);
    if (ConfigManager.logMsgContent) {
      buf.append(", packet=[");
      for (int i = 0; i < fragLen; i++) {
        buf.append(" ");
        buf.append(Util.byte2str(msgFragment[i]));
      }
    }
    buf.append("]");
    return buf.toString();
  }

} // END FragmentHeader
