/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.util.InMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgCntrl</code> class
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgCntrl
  implements MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MsgCntrl.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Event handler */
  private EventHandler ehandler;
  
  /** Upper layer */
  private MssUser mssuser;

  /** Distributed system */
  private MssDS mssds;
  
  /** Local host */
  private MssHost me;

  /**
   *  Counts the number of messages received before the message flow is
   *  initialized.
   */
  private int flowCnt = 0;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  MsgCntrl(EventHandler ehandler, MssUser mssuser, MssDS mssds)
  {
    this.ehandler = ehandler;
    this.mssuser = mssuser;
    this.mssds = mssds;
    this.me = HostTable.getLocalHost();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message send/receive handling
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Compute a fragment identifier for the given fragment iterator (message),
   *  and store the message in the <i>sent</i> queue.  A message may have many
   *  fragments, and thus multiple fragment identifiers associated with
   *  it.  Each of these fragment identifiers point to the same message
   *  instance.
   *
   *  @param fragIter
   *    The fragment iterator (message) for which to compute a fragment identifier.
   *  @return
   *    A fragment identifier associated with the given message.
   */
  int msgSend(FragmentIterator fragIter)
  {
    Msg msg = fragIter.getMsg();
    MsgFlowSndrSide mf = msg.getMsgFlow();

    if (!msg.hasToBeRouted()) {
      /*
       *  This operation is needed for the correct computation of
       *  lastMsgDelivered, in particular when localhost is the
       *  only member of the cluster. It is needed to avoid that
       *  the sent queue grows too big. The sent queue is flushed
       *  periodically when a IAMALIVE message sent by me is received.
       */
      me.getMsgFlow().clusterWindow.set(me.getClusterIndex(), mf.getLastMsgSent() + 1);
    }
    log.assertLog(mf.getLastMsgAcked() != UNDEF,
      "The sender side message flow has not yet been initialized");

    /*
     *  Try to send a message; wait if the window is closed 
     */
    while (true) {
      if (mf.isWindowOpen()) {
        /* 
         *  The sender is not blocked 
         */
        return mf.insertSentMsgFrag(fragIter);

      } else {
        /* 
         *  The sender is blocked 
         */
        if (log.isDebugEnabled())
          log.warn("Sender is blocked waiting for window to open");
        mf.waitOpen();
      }
    }
  }


  /**
   *  Handles the receiving of a normal message from the net.  There are
   *  three cases: (i) the message is delivered immediately; (ii) the
   *  message is enqueued because is not complete; (iii) the message is
   *  enqueued because it is not in FIFO order.
   *
   *  @return 
   *    False if the message could not be delivered because the message
   *    flow was not yet initialized; otherwise, true is returned.
   */
  boolean msgReceive(FragmentHeader header)
  {
    if (log.isDebugEnabled())
      log.debug("--- Enter: msgReceive --------------------------------------------------");

    MssHost sender = header.getSender();
    MsgFlowRcvrSide mf = sender.getMsgFlow();
    SWindow swindow = mf.swindow;

    /*
     * Check if the receive window is closed.
     */
    if (swindow.getLastMsgDlvr() == UNDEF) {
      if ((++flowCnt % 20) == 0) {
        log.warn("The receiver side message flow has not yet been initialized: " + flowCnt);
      }
      return false;
    }

    log.assertLog(swindow.inWindow(header.fragId),
      "Flow control overflow: Msg.header=" + header + ", MsgFlow=" + mf);
    /*
     * FIXED Hein Nov. 3rd 2005; ignore messages not in the sliding window;
     * we return true to pretend that it has been delivered.
     */
    if (!swindow.inWindow(header.fragId))
      return true;

    /*
     * Check if there is a NACK timeout not yet expired for this
     * message; if so we remove it.
     */
    if (mf.hasNACK(header.fragId)) {
      if (log.isDebugEnabled())
        log.debug("msgReceive : Check nackTimeout event associated with msg.header=" + header);
      ScheduledEvent event = mf.getScheduledEvent(header.fragId);
      if (event != null) {
        MsgNACK nack = (MsgNACK) event.getData();
        if (log.isDebugEnabled())
          log.debug("msgReceive : Remove NACK timeout " + nack);
        ehandler.abortTimeout(event);

        // Remove the nack from the queue
        mf.removeScheduledEvent(header.fragId);

        /* With adaptive timeouts, timeouts are delayed */
        sender.updateTimeout(nack.getRTT());
      }
    }

    if (log.isDebugEnabled())
      log.debug("msgReceive: START CONTROL - mid = " + header.fragId + ", lastMsgDlvr = "
                + swindow.lastMsgDlvr + ", lastMsgRcvd = " + mf.getLastMsgRcvd());

    if (mf.isFIFO(header.fragId)) {

      /*
       * The message can be delivered in FIFO order.
       */
      if (log.isDebugEnabled())
        log.debug("msgReceive: FIFO OK " + header.fragId);

      InMessage msgstream = mf.addFragment(header);
      if (msgstream != null) {
        if (log.isDebugEnabled())
          log.debug("msgReceive: Delivering msg (" + header.fragId + ") to mssuser (Daemon)");
        try {
          MsgJG msgjg = MsgJG.unmarshal(msgstream, header, mssds);
          InMessage inmsg = msgjg.complete();
          deliverMsg(msgjg, inmsg, sender);

        } catch (IOException e) {
          log.warn("Could not unmarshal message stream (" + header.fragId + ")\n" + msgstream, e);
        } catch (ClassNotFoundException e) {
          log.warn("Could not unmarshal message stream (" + header.fragId + ")\n" + msgstream, e);
        }
      }

      /* Check if we can deliver delayed messages from this sender. */
      checkReceived(sender);

    } else if (!mf.causeGap(header.fragId) && mf.isNotFIFO(header.fragId)) {

      /*
       * The message cannot be delivered since it is not in FIFO order,
       * and it may fill a existing hole.  We insert the message in the
       * queue unless it has already been received.
       */
      if (!mf.isDelivered(header.fragId)) {
        if (log.isDebugEnabled())
          log.debug("msgReceive: Queuing received msg (" + header.fragId
                    + "); it does not respect FIFO order");
        mf.insertMsgFrag(header);
      }

    } else if (mf.causeGap(header.fragId)) {

      /*
       * The message cannot be delivered and there is a hole in the
       * sequence not yet notified.
       */
      if (log.isDebugEnabled())
        log.debug("msgReceive: recevied msg causing a gap in the sequence (" + header.fragId + ") from sender " + sender
                  + "; lastMsgRcvd=" + mf.getLastMsgRcvd());
      mf.insertMsgFrag(header);
      int oldLastMsgRcvd = mf.setLastMsgRcvd(header.fragId);

      // FIXME No local recovery (???)  We don't need to NACK request
      // all msgs between, use the window to determine the missing
      // messages.  Maybe we can use: mf.isDelivered(i)
      for (int i = oldLastMsgRcvd + 1; i < header.fragId; i++) {
        /*
         * Sending NACK messages for the missing message(s)
         */
        if (log.isDebugEnabled())
          log.debug("msgReceive: Sending NACK to " + sender + " for msg fragment (" + i + ")");
        sender.sendNACK(i);
      }
    }
    if (log.isDebugEnabled()) {
      log.debug(swindow.toString());
      log.debug("--- Exit: msgReceive --------------------------------------------------");
    }
    return true;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private message receive handling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check if the message is for me, and if so deliver to the upper
   *  layer.  Messages not for me, are either routed to the correct
   *  destination or discarded.
   */
  private void deliverMsg(MsgJG msgjg, InMessage inmsg, MssHost sender)
  {
    if (msgjg.hasToBeRouted()) {

      //FIXME routing is currently not supported (at least it has not
      //been tested, at all)

      /* The message has to be routed to its actual destination */
      if (log.isDebugEnabled())
        log.warn("Routing message to destination: " + msgjg);
      Cluster destCluster = msgjg.getCluster();
//       MssHost[] receivers = msgjg.getReceivers();

//       /*
//        * If the message is unicast (only one receiver) and the cluster
//        * is directly connected, we send it directly to the receiver;
//        * otherwise we use the route of the destination cluster.
//        */
//       if (receivers.length == 1 && destCluster.directlyConnected()) {
//         if (log.isDebugEnabled())
//           log.debug("Routing unicast msg directly to member: " + receivers[0]);
//         receivers[0].unicastRouteSend(msgjg);
//       } else {
        if (log.isDebugEnabled())
          log.debug("Routing multicast msg destined for cluster " + destCluster);
        destCluster.send(msgjg);
//       }

    } else if (!msgjg.isForMe()) {

      /* The message is not for me; discard it */
      if (log.isDebugEnabled())
        log.debug("Discarding message not addressed to me: " + msgjg);

    } else {

      /* The message is for me; deliver it to the upper layer */
      if (log.isDebugEnabled())
        log.debug("Delivering message: " + msgjg);
      mssuser.remoteReceive(msgjg.getJGTag(), inmsg, sender.getEndPoint());

    }
  }


  /**
   *  Check if there are messages in the FIFO-order queue of
   *  <code>sender</code> that can be delivered.
   */
  private void checkReceived(MssHost sender) 
  {
    MsgFlowRcvrSide msgFlow = sender.getMsgFlow();
    for (int i = 0, dlvrCount = msgFlow.getDeliveryCount(); i < dlvrCount; i++) {
      /* Get and remove the first message in the queue */
      FragmentHeader header = msgFlow.removeMsgFrag();
      MsgFlowRcvrSide mFlow = header.getSender().getMsgFlow();
      InMessage msgstream = mFlow.addFragment(header);
      if (msgstream != null) {
        if (log.isDebugEnabled())
          log.debug("checkReceived: Delayed delivery msg (" + header.fragId + ") to mssuser (Daemon)");
        try {
          MsgJG msgjg = MsgJG.unmarshal(msgstream, header, mssds);
          InMessage inmsg = msgjg.complete();
          deliverMsg(msgjg, inmsg, msgjg.getSender());
        } catch (IOException e) {
          log.warn("Could not unmarshal message stream\n" + msgstream, e);
        } catch (ClassNotFoundException e) {
          log.warn("Could not unmarshal message stream\n" + msgstream, e);
        }
      }
    }
  }

} // END MsgCntrl
