/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;
import java.util.NoSuchElementException;

import jgroup.util.OutMessage;

/**
 *  Provides an iterator over the fragments of a message.
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
class MsgFragmentIterator
  implements FragmentIterator, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Constants for the duration of this iterator */
  private byte[][] fragments;
  private int fragmentCount;
  private int defaultFragLen;
  private int lastFragLen;

  /* The actual message associated with this fragment iterator */
  private Msg msg;

  /* The message control object */
  private MsgCntrl msgCntrl;

  /* Indicates if the iterator has been initialized */
  private boolean initialized;

  /* Pointer to the next fragment to be returned by the iterator */
  private int nextFragment;

  /* The actual length of the fragment previously returned */
  private int fragLen;

  /* The fragment identifier of the fragment previously returned */
  private int fragId;

  /* The first fragment identifier represented by this iterator */
  private int firstFragId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Creates a new <code>MsgFragmentIterator</code> instance, used to
   *  access the fragments of the given <code>OutMessage</code>.
   */
  MsgFragmentIterator(Msg msg, MsgCntrl msgCntrl)
  {
    this.msgCntrl = msgCntrl;
    this.msg = msg;
    reset(msg.getOutMessage());
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Reset the iterator to start from the first fragment again.  Note
   *  that this method should be used with care, especially in the case
   *  when a message may potentially be resent, based on the fragId.
   *  This is since, for each reuse, this iterator will change its
   *  fragment identifier.  This requires that the iterator must be
   *  reinitialized with the correct fragment identifier before reuse.
   */
  void reset(OutMessage outmsg)
  {
    initialized    = false;
    fragments      = outmsg.getFragments();
    fragmentCount  = outmsg.getFragmentCount();
    defaultFragLen = outmsg.getPayload() + outmsg.getHeader();
    lastFragLen    = outmsg.getLastPayload() + outmsg.getHeader();

    nextFragment = 0;
    fragLen      = -1;
    fragId       = -1;
    firstFragId  = -1;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // FragmentIterator methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the fragment associated with the given fragment identifier.
   *
   *  @exception IllegalStateException
   *    Raised if the iterator has not been invoked prior to calling
   *    this method.  This is since we cannot know the iterators
   *    fragment identifiers until we have invoked next.
   *  @exception IndexOutOfBoundsException
   *    Raised if the given fragment identifier is not represented by
   *    this iterator.
   *  @exception NoSuchElementException
   *    Thrown if the computed fragment position exceeds the fragment
   *    count of the iterator.  This can never happen.
   */
  public byte[] getFragment(int fragId)
  {
    if (!initialized)
      throw new IllegalStateException("Cannot seek in the iterator before having iterated once");

    int lastFragId = fragmentCount + firstFragId - 1;
    if (fragId > lastFragId || fragId < firstFragId)
      throw new IndexOutOfBoundsException("The fragment identifier ("
        + fragId + ") is not represented by this iterator: " + this);

    nextFragment = fragId - firstFragId;
    if (nextFragment >= fragmentCount)
      throw new NoSuchElementException("There are no more fragments associated with this message.");

    if (fragId == lastFragId)
      fragLen = lastFragLen;
    else
      fragLen = defaultFragLen;
    this.fragId = fragId;

    return fragments[nextFragment];
  }


  /**
   *  Returns the length (including the header) of the active
   *  fragment; that is, the one most recently retrieved through the
   *  <code>next()</code> method. <p>
   *
   *  The last fragment will have a different length than the other
   *  fragments, for which this method will return the default payload
   *  length of a packet.  Also a single fragment counts as a last
   *  fragment, and may thus not use the entire payload length.  This
   *  method always returns the correct length of the active fragment.
   *  <p>
   *
   *  It is not possible to use the fragment.length field, since we
   *  are not copying the data of the original; we are merely returning
   *  its reference.
   */
  public int fragmentLength()
  {
    if (!initialized)
      throw new IllegalStateException("No fragment is active; must invoke next() first");
    return fragLen;
  }


  /**
   *  Returns the fragment identifier of the active fragment.
   */
  public int getFid()
  {
    if (!initialized)
      throw new IllegalStateException("No fragment is active; must invoke next() first");
    return fragId;
  }


  /**
   *  Returns the actul message object assocaited with this fragment
   *  iterator.
   */
  public Msg getMsg()
  {
    if (msg == null)
      throw new NullPointerException("The message is null; iterator is not correctly initialized");
    return msg;
  }


  /**
   *  Returns true if the message stream has more fragments; false is
   *  returned otherwise.
   */
  public boolean hasNext()
  {
    return nextFragment != fragmentCount;
  }


  /**
   *  Returns the next fragment of this message object's stream,
   *  including the marshalled fragment header which includes the message
   *  <code>tag</code> and <code>fragId</code>.  The latter constitute
   *  the packet level header.
   *
   *  @see jgroup.relacs.mss.MssTag
   *  @param broadcast
   *    Indicates if the message fragment should be broadcast, and
   *    thus sets a packet level broadcast field.
   *  @return
   *    A byte array containing the next fragment in the stream,
   *    including the header.
   */
  public byte[] next(boolean broadcast)
  {
    if (nextFragment == fragmentCount)
      throw new NoSuchElementException("There are no more fragments associated with this message.");

    byte tag = msg.getTag();

    /*
     * Set the correct tag for this message fragment.  If there is only
     * one fragment, it will be marked with the tag obtained from the
     * message.  The message tag will always mark the last fragment.  If
     * there are multiple fragments, the first <i>n - 1</i> fragments
     * will be marked with the <code>NOTLASTFRAGMENT</code> tag, while
     * the <i>n</i>-th fragment will be marked with the message tag.
     */
    if (nextFragment + 1 == fragmentCount) {
      fragLen = lastFragLen;
    } else {
      fragLen = defaultFragLen;
      tag = NOTLASTFRAGMENT;
    }

    if (msg instanceof MsgJG) {
      /* The msgSend() method blocks until the sending window is open */
      fragId = msgCntrl.msgSend(this);
    } else {
      /* IAMALIVE, ROUTING, NACK and SYN messages provide their own fragId */
      fragId = msg.getMid();
    }

    if (firstFragId == -1)
      firstFragId = fragId;

    /* The iterator has now been initialized. */
    initialized = true;

    FragmentHeader.marshal(tag, broadcast, fragId, fragments[nextFragment]);

    return fragments[nextFragment++];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[MsgFragmentIterator: nxtFrag=");
    buf.append(nextFragment);
    buf.append(", fragLen=");
    buf.append(fragLen);
    buf.append(", fragId=");
    buf.append(fragId);
    buf.append(", firstFragId=");
    buf.append(firstFragId);
    buf.append(", lastFragId=");
    buf.append((fragmentCount + firstFragId - 1));
    buf.append(", fragCnt=");
    buf.append(fragmentCount);
    buf.append(", lastFragLen=");
    buf.append(lastFragLen);
    buf.append("]");
    return buf.toString();
  }

} // END MsgFragmentIterator
