/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.relacs.types.GroupIndex;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgIamAlive</code> class
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class MsgIamAlive
  implements Msg, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(MsgIamAlive.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  As this message is never stored in a queue, a single message
   *  instance is sufficient.  This allows us to save the cost of
   *  allocating and garbage collecting messages.
   */
  private static final MsgIamAlive msg = new MsgIamAlive();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (dynamic part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Array of flow control entries */
  private FCEntry[] fc;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields (recomputed during unmarshalling)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Message to be sent */
  private transient OutMessage outmsg;

  /** Message received from the Mss */
  private transient InMessage inmsg;

  /** The index of the flow control entry associated with the local host */
  private transient int localFCIndex = UNDEF; 

  /** The local host (where this message resides; sender or receiver) */
  private transient MssHost localHost;

  /** The message sender (obtained from the fragment header) */
  private transient MssHost sender;

  /** The incarnation identifier for the sender (obtained from the fragment header) */
  private transient int incarnationId;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshalling and unmarshalling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshal an IAMALIVE message.
   */
  static MsgIamAlive marshal(FCEntry[] fc)
    throws IOException
  {
    msg.fc     = fc;
    msg.sender = HostTable.getLocalHost();

    /*
     * Here we check if this <code>MsgIamAlive</code> object has already
     * been initialized with an outmsg object, in which case we simply
     * clear the message object and reuse it.  This avoids to create a
     * new outmsg object (allocating additional memory) and there is no
     * need to garbage collect the old one.
     *
     * Note that all <code>MsgIamAlive</code> messages are of the same
     * size, including the flow control data of all hosts.
     */
    int size = GroupIndex.SIZE + fc.length*FCEntry.SIZE;
    if (msg.outmsg == null) {
      msg.outmsg = MsgFactory.get(size);
    }
    if (log.isDebugEnabled()) {
      log.debug("MsgIamAlive.size=" + size);
    }
    msg.outmsg.clear(size);

    GroupIndex.marshal(msg.outmsg, fc.length);
    for (int i = 0; i < fc.length; i++)
      fc[i].writeExternal(msg.outmsg);
    return msg;
  }


  /**
   *  Unmarshal an IAMALIVE message from the given stream.
   */
  static MsgIamAlive unmarshal(InMessage inmsg, FragmentHeader header)
    throws IOException, ClassNotFoundException
  {
    msg.inmsg         = inmsg;
    msg.localHost     = HostTable.getLocalHost();
    msg.sender        = header.getSender();
    msg.incarnationId = header.getIncarnationId();

    int len = GroupIndex.unmarshal(inmsg);
    msg.fc = new FCEntry[len];
    for (int i = 0; i < len; i++) {
      msg.fc[i] = new FCEntry();
      msg.fc[i].readExternal(inmsg);
      if (msg.localFCIndex == UNDEF && msg.fc[i].key.isLocal())
        msg.localFCIndex = i;
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MsgIamAlive specific methods (receiver side only)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Update the flow control data on the receiver side.
   */
  void updateFC(MssDS mssds)
  {
    if (sender.isLocal()) {

      localHost.flush();

    } else {

      for (int i = 0; i < fc.length; i++) {
        MssHost host = mssds.hostLookup(fc[i].key);
        if (host == null) {
          log.error("Unavailable host: " + fc[i].key);
          continue;
        }
        if (log.isDebugEnabled()) {
          log.debug("host=" + host + ", lastMsgRcvd=" + fc[i].lastMsgRcvd
                    + ", cindex=" + sender.getClusterIndex());
        }
        host.getMsgFlow().clusterWindow.set(sender.getClusterIndex(), fc[i].lastMsgRcvd);
      }

    }
  }


  /**
   *  Returns incarnation identifier for the sender of this message.
   */
  public int getIncarnationId()
  {
    return incarnationId;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fragmentation handling (from Msg interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * The fragmentation iterator instance for this message; it can be
   * reused for multiple iterations over this message.
   */
  private transient MsgFragmentIterator fragIter = null;


  /**
   *  Returns a <code>FragmentIterator</code> for this
   *  <code>MsgIamAlive</code> object.  This iterator allows to send the
   *  entire message as multiple fragments of specified size (payload).
   *  At the same time, it marks each fragment with a tag and message
   *  identifier provided through the <code>next()</code> method of the
   *  iterator.
   *
   *  If there is already an iterator for this message, we reuse it.
   */
  public FragmentIterator iterator(MsgCntrl msgCntrl)
  {
    if (fragIter == null)
      fragIter = new MsgFragmentIterator(this, msgCntrl);
    else 
      fragIter.reset(outmsg);
    return fragIter;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Msg interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated with this message.
   */
  public byte getTag()
  {
    return IAMALIVE;
  }


  /**
   *  Returns the message identifier for this message; all IAMALIVE
   *  messages have <code>UNDEF</code> as their message identifier.
   */
  public int getMid()
  {
    return UNDEF;
  }


  /**
   *  Returns the sender of this message.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns false always, since topology messages should never be routed.
   */
  public boolean hasToBeRouted()
  {
    return false;
  }


  /**
   *  Returns true if the sender of this message is the local host.
   */
  boolean isLocal()
  {
    return sender.isLocal();
  }


  /**
   *  Returns the message flow controller for the sender side.
   */
  public MsgFlowSndrSide getMsgFlow()
  {
    return sender.getCluster().getMsgFlow();
  }


  /**
   *  Returns the <code>OutMessage</code> associated with this message.
   */
  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("IAMALIVE, ");
    for (int i=0; i < fc.length; i++)
      buf.append(fc[i]);
    return buf.toString();
  }
  
} // END MsgIamAlive
