/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.core.EndPoint;
import jgroup.relacs.types.GroupIndex;
import jgroup.util.InMessage;
import jgroup.util.MsgFactory;
import jgroup.util.OutMessage;

import org.apache.log4j.Logger;

/**
 *  The <code>MsgRouting</code> class
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public final class MsgRouting
  implements Msg, MssConstants, MssTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static Logger log = Logger.getLogger(MsgRouting.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  As this message is never stored in a queue, a single message
   *  instance is sufficient.  This allows us to save the cost of
   *  allocating and garbage collecting messages.
   */
  private static final MsgRouting msg = new MsgRouting();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message pointers
  ////////////////////////////////////////////////////////////////////////////////////////////

  private transient int fcStartPosition;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (static part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* Tag assigned in upper level (@see jgroup.relacs.mss.MssTag) */
  private byte routingTag;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message Fields (dynamic part)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Routing table */
  private RoutingTable rt;

  /** Array of flow control entries */
  private FCEntry[] fc;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Transient fields (recomputed during unmarshalling)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The MssDS object */
  private transient MssDS mssds;

  /** Message to be sent */
  private transient OutMessage outmsg;

  /** The index of the flow control entry associated with the local host */
  private transient int localFCIndex = UNDEF; 

  /** The message sender (obtained from the fragment header) */
  private transient MssHost sender;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Marshalling and unmarshalling methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Marshal a ROUTING message.
   */
  static MsgRouting marshal(MssDS mssds)
    throws IOException
  {
    msg.routingTag = ROUTING;
    msg.mssds      = mssds;
    msg.sender     = HostTable.getLocalHost();
    msg.rt         = mssds.getRoutingTable();

    /*
     * Compute the max size for this message.
     */
    int size = GroupIndex.SIZE + mssds.size()*FCEntry.SIZE;
    /*
     * Here we check if this <code>MsgRouting</code> object has already
     * been initialized with an outmsg object, in which case we simply
     * clear the message object and reuse it.  This avoids to create a
     * new outmsg object (allocating additional memory) and there is no
     * need to garbage collect the old one.
     */
    if (msg.outmsg == null) {
      int maxsize = size + RoutingTable.getMaxSize(mssds.size(), mssds.numOfClusters());
      if (log.isDebugEnabled()) {
        log.debug("MsgRouting.maxsize=" + maxsize);
      }
      msg.outmsg = MsgFactory.get(maxsize);
    }
    size += msg.rt.getMarshalSize();
    if (log.isDebugEnabled()) {
      log.debug("MsgRouting.size=" + size);
    }
    msg.outmsg.clear(size);

    /* Marshal the routing table */
    msg.rt.writeExternal(msg.outmsg);
    msg.fcStartPosition = msg.outmsg.getPosition();
    /*
     * Note that, we do not initialize the flow control here, since the
     * mss will do so using the <code>setFCData()</code> method below,
     * for each cluster; allowing the mss to reuse the same
     * <code>MsgRouting</code> object for sending to all clusters.
     */
    return msg;
  }


  /**
   *  Unmarshal an FWDROUTING message from the given stream.
   */
  static MsgRouting unmarshal(InMessage in, FragmentHeader header)
    throws IOException, ClassNotFoundException
  {
    /* Obtain stuff from the header */
    msg.routingTag = header.getTag();
    msg.sender     = header.getSender();

    /* Obtain the routing table from the stream */
    msg.rt = new RoutingTable();
    msg.rt.readExternal(in);

    /* Obtain the flow control information table from the stream */
    int len = GroupIndex.unmarshal(in);
    msg.fc = new FCEntry[len];
    for (int i=0; i < len; i++) {
      msg.fc[i] = new FCEntry();
      msg.fc[i].readExternal(in);
      if (msg.localFCIndex == UNDEF && msg.fc[i].key.isLocal())
        msg.localFCIndex = i;
    }
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // MsgRouting specific method
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the largest possible size that a <code>MsgRouting</code>
   *  message may consume, for the given distributed system size.
   */
  public static int getMaxSize(int numOfHosts, int numOfClusters)
  {
    return GroupIndex.SIZE + numOfHosts*FCEntry.SIZE
      + RoutingTable.getMaxSize(numOfHosts, numOfClusters);
  }


  /**
   *  Update the flow control data on the receiver side.
   */
  void updateFC() 
  {
    if (routingTag == FWDROUTING) {

      /* Check if there is a flow control entry for the local host */
      if (localFCIndex != UNDEF) {
        /* Update congestion information (sender side flow control) */
        sender.flush(fc[localFCIndex].lastMsgRcvd);
      }

    } else {

      log.warn("Erroneous routing tag; can only process FWDROUTING: " + routingTag);

    }
  }


  /**
   *  Returns the sender cluster of this message.
   */
  Cluster getCluster()
  {
    return sender.getCluster();
  }


  /**
   *  Returns the topology table associated with the routing table
   *  attached with this message.
   */
  public TopologyEntry[] getTopologyTable()
  {
    return rt.table;
  }


  /**
   *  Negate the cost as reverse poison 
   */
  void splitHorizonOn(EndPoint key) 
  {
    rt.splitHorizonOn(key);
  }


  /**
   *  
   */
  void splitHorizonOff() 
  {
    rt.splitHorizonOff();
  }


  void setFCData(EndPoint destCluster)
  {
    fc = mssds.getClusterFCEntry(destCluster);
    outmsg.seek(fcStartPosition);
    try {
      GroupIndex.marshal(outmsg, fc.length);
      for (int i=0; i < fc.length; i++)
        fc[i].writeExternal(outmsg);
    } catch (IOException e) {
      log.error("setFCData: Could not update flow control information in message stream", e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fragmentation handling (from Msg interface)
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   * The fragmentation iterator instance for this message; it can be
   * reused for multiple iterations over this message.
   */
  private transient MsgFragmentIterator fragIter = null;


  /**
   *  Returns a <code>FragmentIterator</code> for this
   *  <code>MsgRouting</code> object.  This iterator allows to send the
   *  entire message as multiple fragments of specified size (payload).
   *  At the same time, it marks each fragment with a tag and message
   *  identifier provided through the <code>next()</code> method of the
   *  iterator.
   */
  public FragmentIterator iterator(MsgCntrl msgCntrl)
  {
    /* If there is already an iterator for this message, we reuse it. */
    if (fragIter == null)
      fragIter = new MsgFragmentIterator(this, msgCntrl);
    else 
      fragIter.reset(outmsg);
    return fragIter;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Msg interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the tag associated to this message.
   */
  public byte getTag()
  {
    return routingTag;
  }


  /**
   *  Returns the message identifier for this message; all routing
   *  messages have <code>UNDEF</code> as their message identifier.
   */
  public int getMid()
  {
    return UNDEF;
  }


  /**
   *  Returns the sender of this message.
   */
  public MssHost getSender()
  {
    return sender;
  }


  /**
   *  Returns false always, since routing messages should never be routed.
   */
  public boolean hasToBeRouted()
  {
    return false;
  }


  /**
   *  Returns true if the sender of this message is the local host.
   */
  boolean isLocal()
  {
    return sender.isLocal();
  }


  /**
   *  Returns the message flow controller for the sender side.
   */
  public MsgFlowSndrSide getMsgFlow()
  {
    return sender.getCluster().getMsgFlow();
  }


  /**
   *  Returns the <code>OutMessage</code> associated with this message.
   */
  public OutMessage getOutMessage()
  {
    return outmsg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    switch (routingTag) {
    case FWDROUTING:
      buf.append("FWDROUTING, ");
      buf.append(rt.toString());
      for (int i=0; i < fc.length; i++)
        buf.append(fc[i]);
      break;

    case ROUTING:
      buf.append("ROUTING, ");
      buf.append(rt.toString());
      for (int i=0; i < fc.length; i++)
        buf.append(fc[i]);
      break;

    default:
      buf.append("Illegal routing type: " + routingTag);
    }
    return buf.toString();
  }

} // END MsgRouting
