/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketException;

import jgroup.core.EndPoint;

import org.apache.log4j.Logger;
import org.apache.log4j.NDC;

/**
 *  The <code>MulticastNetworkInterface</code> class implements a
 *  multicast version of the network interface.
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @author Marcin Solarski
 *  @since Jgroup 1.2
 */
final class MulticastNetworkInterface
  extends Thread
  implements NetworkInterface, MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(MulticastNetworkInterface.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Upper layer */
  private NIListener niListener;

  /** Multicast internet address */
  private InetAddress address;

  /** Multicast Socket */
  private MulticastSocket group;

  /** Multicast port */
  private int port;

  /** Receive buffer length */
  private int bufferlen;

  /** True when the thread should continue to run */
  private boolean carryOn;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs and starts the multicast network interface layer.
   *
   *  @param name
   *    Thread identifier.
   *  @param niListener
   *    Reference to the upper listener layer through which multicast
   *    receive events are passed as upcall invocations.
   *  @param mcastEndpoint
   *    Multicast group address.
   *  @param bufferlen
   *    Size of receive buffer.
   */
  MulticastNetworkInterface(String name, NIListener niListener, EndPoint mcastEndpoint, int bufferlen, int multicastTTL)
    throws IOException
  {
    /* Initialize thread */
    super(name);
    this.setPriority(NI_PRIORITY);
    this.setDaemon(true);

    if (!mcastEndpoint.isMulticastEndPoint())
      throw new IOException("Endpoint is not a multicast endpoint: " + mcastEndpoint);

    this.niListener = niListener;
    this.bufferlen  = bufferlen;
    this.address    = mcastEndpoint.getAddress();
    this.port       = mcastEndpoint.getPort();
    this.carryOn    = true;

    /* Initialize multicast socket for this endpoint. */
    boolean partitionSimulator = Boolean.getBoolean("jgroup.simulator");
    try {
      if (partitionSimulator)
        group = new UnreliableMulticastSocket(port);
      else
        group = new MulticastSocket(port);
    } catch (SocketException e) {
      throw new IOException("Error initializing multicast socket: " + mcastEndpoint
                            + "\n" + e.getMessage());
    }

    try {
      /* Join the multicast group for this endpoint address. */
      group.joinGroup(address);
    } catch (IOException e) {
      throw new IOException("Unable to join multicast group: " + address);
    }

    if (multicastTTL > 0 && multicastTTL < 17)
      group.setTimeToLive(multicastTTL);
    else
      log.error("Multicast TTL must be in the range [1,16].");
    if (log.isDebugEnabled()) {
      log.debug("Multicast TTL = " + group.getTimeToLive());
      log.debug("Multicast socket initialized: " + mcastEndpoint + " " + bufferlen);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from NetworkInterface
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Javadoc comment inherited from NetworkInterface
  public void send(byte[] buffer, int buflen)
  {
    DatagramPacket packet = new DatagramPacket(buffer, buflen, address, port);

    try {
      group.send(packet);
    } catch (IOException e) {
      log.warn("Error sending packet to ALL", e);
    }
  }


  // Javadoc comment inherited from NetworkInterface
  public void send(EndPoint dest, byte[] buffer, int buflen)
  {
    DatagramPacket packet =
      new DatagramPacket(buffer, buflen, dest.getAddress(), dest.getPort());

    try {
      group.send(packet);
    } catch (IOException e) {
      log.warn("Error sending packet to " + dest, e);
    }
  }


  // Javadoc comment inherited by NetworkInterface
  public void doStart()
  {
    start();
  }


  // Javadoc comment inherited by NetworkInterface
  public void doStop() 
  {
    carryOn = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Clean up system resources used by this class.  This method will be
   *  invoked when there are no more references to this object, and thus
   *  it should be garbage collected.
   */
  protected void finalize() 
  {
    if (group != null) {
      try {
        group.leaveGroup(address);
      } catch (IOException e) {
        log.warn("Unable to cleanly leave the multicast group: " + address, e);
      }
      group.close();      
      group = null;       
    }
    carryOn = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Thread
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Run method of the thread.  It blocks until a
   *  <code>DatagramPacket</code> is received from the network.
   */
  public void run()
  {
    while (carryOn) {
      try {
        byte[] buf = new byte[bufferlen];
        DatagramPacket packet = new DatagramPacket(buf, bufferlen);
        group.receive(packet);
        InetAddress sender = packet.getAddress();
        if (packet.getLength() > 0) {
          if (log.isDebugEnabled()) {
            NDC.push(sender.getHostName());
            log.debug("multicast packet.length=" + packet.getLength()
                + ", sender=" + sender.getHostName()
                + ", mcast-group=" + address.getHostAddress());
          }
          niListener.rnotify(packet);
        } else if (log.isDebugEnabled()) {
          NDC.push(sender.getHostName());
          /*
           * The <code>UnreliableMulticastSocket</code> has set the packet size to zero,
           * which means that we should discard it.
           */
          log.debug("Discarding multicast packet from " + sender.getHostName());
        }
      } catch (IOException e) {
        log.warn("Error receiving packets", e);
      } finally {
        if (log.isDebugEnabled())
          NDC.pop();
      }
    }
    /* The thread was stopped, cleaning up system resources. */
    this.finalize();
  }

} // END MulticastNetworkInterface
