/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;

import jgroup.core.ConfigManager;
import jgroup.core.EndPoint;
import jgroup.relacs.config.DistributedSystemConfig;
import jgroup.util.Util;

import org.apache.log4j.Logger;

/**
 *  The <code>NI</code> class contains the network interface used by the
 *  <code>Mss</code>.
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class NI 
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(NI.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Unicast network interface */
  private NetworkInterface uni;

  /** Multicast network interface */
  private NetworkInterface mni;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Create one instance each of the unicast and multicast network
   *  interfaces.
   */
  public NI(NIListener upperLayer, int bufferlen, int multicastTTL)
    throws IOException
  {
    EndPoint localHost   = DistributedSystemConfig.getLocalHost();
    EndPoint localDomain = DistributedSystemConfig.getLocalDomain();
    uni = new UnicastNetworkInterface("UNI", upperLayer, localHost, bufferlen);
    mni = new MulticastNetworkInterface("MNI", upperLayer, localDomain, bufferlen, multicastTTL);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Send a message to the local cluster.
   */
  public void send(byte[] buffer, int buflen)
  {
    if (log.isDebugEnabled()) {
      StringBuilder sbuf = new StringBuilder("send to local cluster (LAN) multicast address: packet [");
      sbuf.append(buflen);
      sbuf.append(" bytes]");
      if (ConfigManager.logMsgContent) {
        for (int i = 0; i < buflen; i++) {
          sbuf.append(" ");
          sbuf.append(Util.byte2str(buffer[i]));
        }
      }
      log.debug(sbuf.toString());
    }

    mni.send(buffer, buflen);
  }


  /** 
   *  Send a message to a remote host.
   */
  public void send(EndPoint dest, byte[] buffer, int buflen)
  {
    if (log.isDebugEnabled()) {
      StringBuilder sbuf = new StringBuilder();
      if (dest.isMulticastEndPoint())
        sbuf.append("send to local cluster (LAN) multicast address: ");
      else 
        sbuf.append("send to remote cluster leader: ");
      sbuf.append(dest);
      sbuf.append(" packet [");
      sbuf.append(buflen);
      sbuf.append(" bytes]");
      if (ConfigManager.logMsgContent) {
        for (int i = 0; i < buflen; i++) {
          sbuf.append(" ");
          sbuf.append(Util.byte2str(buffer[i]));
        }
      }
      log.debug(sbuf.toString());
    }

    if (dest.isMulticastEndPoint())
      mni.send(dest, buffer, buflen);
    else
      uni.send(dest, buffer, buflen);
  } 


  /**
   *  Starts the unicast and multicast network interfaces.
   */
  public void doStart() 
  {
    uni.doStart();
    mni.doStart();
  }

} // END NI
