/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import org.apache.log4j.Logger;



/**
 *  The <code>SWindow</code> class implements the datastructures for a
 *  sliding window protocol.
 *
 *  @author Salvatore Cammarrata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class SWindow
  implements MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(SWindow.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max width of the sliding window */
  private int maxWindow;

  /** Left edge of the window */
  int lastMsgDlvr;

  /**
   * Array representing the sliding window, in which a true value
   * indicates that the message was received, while a false indicates
   * not yet received.
   */
  private boolean[] win;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  SWindow(int maxWindow, int lastMsgDlvr)
  {
    this.maxWindow = maxWindow;
    this.win = new boolean[maxWindow];
    this.lastMsgDlvr = lastMsgDlvr;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns true if the given message fragment is within the sliding
   *  window range; otherwise false is returned.
   */
  boolean inWindow(int fid)
  {
    return (fid - lastMsgDlvr <= maxWindow);
  }


  /**
   * Returns true if the whole window is unused; that is no messages
   * have been marked as received in the current window space.  If at
   * least one message has been received in this window, a false is
   * returned.
   */
  boolean isAllZero()
  {
    for (int i = 0; i < win.length; i++) {
      if (win[i])
        return false;
    }
    return true;
  }


  /**
   *  Shift the window <code>pos</code> bits to the left.
   */
  int shift(int pos)
  {
    int i;
    for (i = 0; i < win.length - pos ; i++)
      win[i] = win[i+pos];
    for (i = win.length - pos; i < win.length ; i++)
      win[i] = false;
    return (lastMsgDlvr += pos);
  }


  /**
   *  Shift the window one bit to the left.
   */
  int shift()
  {
    return shift(1);
  }


  /**
   *  Shift the window to the left, the number of received (yet
   *  undelivered) message fragments.  Return the number of messages in
   *  FIFO order in this sliding window that has been received, yet has
   *  not been delivered.
   */
  int shiftFIFO()
  {
    int pos = 0;
    while (win[pos])
      pos++;
    shift(pos);
    return pos;
  }


  /**
   *  Set the left edge (last message delivered) of the sliding window
   *  to the given value, and return its previous value.
   */
  int setLastMsgDlvr(int le)
  {
    int ple = lastMsgDlvr;
    lastMsgDlvr = le;
    return ple;
  }


  /**
   *  Return the current left edge (last message delivered) of the
   *  sliding window.
   */
  int getLastMsgDlvr()
  {
    return lastMsgDlvr;
  }


  /**
   *  Increment the current left edge by one.
   */
  int incLastMsgDlvr()
  {
    return (lastMsgDlvr += 1);
  }


  /**
   *  Mark the given fragment as received in the sliding window.
   */
  void set(int fid)
  {
    if (log.isDebugEnabled())
      log.debug("fragId=" + fid + " lastMsgDlvr=" + lastMsgDlvr);
    int pos = fid - lastMsgDlvr - 1;
    if (pos >= win.length || pos < 0)
      log.warn("out of bounds - pos: " + pos);
    else
      win[pos] = true;
  }


  /**
   *  Mark the given fragment as not received in the sliding window.
   */
  void unset(int fid)
  {
    if (log.isDebugEnabled())
      log.debug("fragId=" + fid + " lastMsgDlvr=" + lastMsgDlvr);
    int pos = fid - lastMsgDlvr - 1;
    if (pos >= win.length || pos < 0)
      log.warn("out of bounds - pos: " + pos);
    else
      win[pos] = false;
  }


  /**
   *  Obtain the status of the given fragment.  Returns true if the
   *  fragment has been received; otherwise false.
   */
  boolean lookup(int fid)
  {
    int pos = fid - lastMsgDlvr - 1;
    if (pos >= win.length || pos < 0) {
      log.warn("out of bounds - pos: " + pos);
      return false;
    } else {
      return win[pos];
    }
  }


  /**
   *  Clear the sliding window state.
   */
  void clear(int lastMsgDlvr)
  {
    this.lastMsgDlvr = lastMsgDlvr;
    for (int i = 0 ; i < win.length ; i++)
      win[i] = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append("[SWindow: size=");
    buffer.append(win.length);
    buffer.append(", lastMsgDlvr=");
    buffer.append(lastMsgDlvr);
    buffer.append(", win={");
    for (int i = 0; i < win.length; i++) {
      if (i > 0 && win[i])
        buffer.append(i);
      buffer.append(win[i] ? (" " + i) : "");
    }
    buffer.append("}]");
    return buffer.toString();
  }

} // END SWindow
