/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.mss;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;

import jgroup.core.EndPoint;

import org.apache.log4j.Logger;
import org.apache.log4j.NDC;

/**
 *  The <code>UnicastNetworkInterface</code> class implements a
 *  unicast version of the network interface.
 *
 *  @author Salvatore Cammarata
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
final class UnicastNetworkInterface
  extends Thread
  implements NetworkInterface, MssConstants
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(UnicastNetworkInterface.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Upper layer */
  private NIListener niListener;

  /** UDP Socket */
  private DatagramSocket socket;

  /** Receive buffer length */
  private int bufferlen;

  /** True when the thread should continue to run */
  private boolean carryOn;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs and starts the unicast network interface layer.
   *
   *  @param name
   *    Thread identifier.
   *  @param niListener
   *    Reference to the upper listener layer through which multicast
   *    receive events are passed as upcall invocations.
   *  @param endpoint
   *    Endpoint socket address and port.
   *  @param bufferlen
   *    Size of receive buffer.
   *  @exception IOException
   *    Raised if the unicast socket could not be initialized.
   */
  UnicastNetworkInterface(String name, NIListener niListener, EndPoint endpoint, int bufferlen)
    throws IOException
  {
    /* Initialize thread */
    super(name);
    this.setPriority(NI_PRIORITY);
    this.setDaemon(true);

    this.niListener = niListener;
    this.bufferlen  = bufferlen;
    this.carryOn    = true;

    /* Initialize socket */
    boolean partitionSimulator = Boolean.getBoolean("jgroup.simulator");
    try {
      if (partitionSimulator) {
        socket = new UnreliableDatagramSocket(endpoint.getPort(), endpoint.getAddress());
      } else {
        if (log.isDebugEnabled())
          log.debug("Trying to create a datagram socket at  " + endpoint.getAddress() + " ." +endpoint.getPort() );     
        socket = new DatagramSocket(endpoint.getPort(), endpoint.getAddress());
      }  
    } catch (SocketException e) {
      if (log.isDebugEnabled())
        log.warn("Error initializing unicast socket at " + endpoint, e);
      throw new IOException("Error initializing unicast socket: " + endpoint + "\n" ); 
    }
    if (log.isDebugEnabled())
      log.debug("Unicast socket initialized: " + endpoint + " " + bufferlen);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from NetworkInterface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  This operation is not supported by the unicast network interface,
   *  because a host does not send a message to itself through the
   *  network interface.  If a host is running multiple daemons on
   *  different ports (JVMs), the method below can be used by specifying
   *  the endpoint.  This is possible since the endpoint contains both
   *  the internet address and the port.
   */
  public void send(byte[] buffer, int buflen)
  {
    throw new UnsupportedOperationException();
  }


  // Javadoc comment inherited by NetworkInterface
  public void send(EndPoint dest, byte[] buffer, int buflen)
  {
    DatagramPacket packet =
      new DatagramPacket(buffer, buflen, dest.getAddress(), dest.getPort());

    try {
      socket.send(packet);
    } catch (IOException e) {
      log.warn("Error sending packet to " + dest, e);
    }
  }


  // Javadoc comment inherited by NetworkInterface
  public void doStart()
  {
    start();
  }


  // Javadoc comment inherited by NetworkInterface
  public void doStop() 
  {
    carryOn = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Clean up system resources used by this class.  This method will be
   *  invoked when there are no more references to this object, and thus
   *  it should be garbage collected.
   */
  protected void finalize() 
  {
    carryOn = false;
    if (socket != null) {   
      socket.close();      
      socket = null;       
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Thread
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Run method of the thread. It blocks until a Datagram packet is received
   * through the net.
   */
  public void run()
  {
    while (carryOn) {
      try {
        byte[] buf = new byte[bufferlen];
        DatagramPacket packet = new DatagramPacket(buf, bufferlen);
        socket.receive(packet);
        if (log.isDebugEnabled()) {
          String senderName = packet.getAddress().getHostName();
          NDC.push(senderName);
          log.debug("unicast packet.length=" + packet.getLength() + ", sender=" + senderName);
        }
        niListener.rnotify(packet);
      } catch (IOException e) {
        log.warn("Error receiving packet", e);
      } finally {
        if (log.isDebugEnabled())
          NDC.pop();
      }
    }
    /* The thread was stopped, cleaning up system resources. */
    this.finalize();
  }

} // END UnicastNetworkInterface
