/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.lang.reflect.InvocationHandler;
import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;

import jgroup.core.ConfigManager;
import jgroup.core.ConfigurationException;
import jgroup.core.ExternalGMIService;
import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.Layer;
import jgroup.core.MemberId;
import jgroup.core.MembershipListener;
import jgroup.core.MembershipService;
import jgroup.core.MergingListener;
import jgroup.core.View;
import jgroup.core.protocols.Anycast;
import jgroup.core.protocols.Multicast;
import jgroup.core.registry.BootstrapRegistry;
import jgroup.core.registry.DependableRegistry;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.gm.LeaseInfo;
import jgroup.relacs.rmi.Exporter;
import jgroup.util.Abort;

import org.apache.log4j.Logger;

/**
 *  This class implements the <code>DependableRegistry</code> interface.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.4
 */
public final class RegistryImpl
  implements DependableRegistry, MergingListener, MembershipListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(RegistryImpl.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private ExternalGMIService  egmis;
  private MembershipService   pgms;
  private RegistryTable       registrytable;
  private MemberId            me;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained in the <tt>conf</tt> map are used by Jgroup to
   * configure the group communication system; the distributed system
   * description must be included in this map object.
   *
   * @exception RemoteException
   *   If the registry could not be exported
   * @exception JgroupException
   *   Generic exception raised by Jgroup
   * @exception ConfigurationException
   *   Raised when the distributed system configuration object contains an error
   */
  public RegistryImpl()
    throws RemoteException, JgroupException, ConfigurationException
  {
    this(null, null, null);
  }

  /**
   * Special constructor for creating the dependable registry as part of
   * another group (e.g., the replication manager group).
   * 
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained in the <tt>conf</tt> map are used by Jgroup to
   * configure the group communication system; the distributed system
   * description must be included in this map object.
   *
   * @exception RemoteException
   *   If the registry could not be exported
   * @exception JgroupException
   *   Generic exception raised by Jgroup
   * @exception ConfigurationException
   *   Raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(GroupManager gm)
    throws RemoteException, JgroupException, ConfigurationException
  {
    this(gm, null, null);
    /*
     * Obtain the various layers required by the dependable registry
     * from the group manager stack of the hosting application, and
     * add the dependable registry object as a listener to these layers.
     */
    Layer layer = (Layer) gm.getService("EGMI");
    layer.addListener(this);
    layer = (Layer) gm.getService("SMS");
    layer.addListener(this);
    layer = (Layer) gm.getService("PGMS");
    layer.addListener(this);
  }

  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on the port specified within the configuration
   * object.  Objects contained in the <tt>conf</tt> map are used by Jgroup to
   * configure the group communication system; the distributed system
   * description must be included in this map object.
   *
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param csf
   *   Client-side <code>Socket</code> factory used to make
   *   connections to the registry
   * @param ssf
   *   Server-side <code>ServerSocket</code> factory used to
   *   accept connections to the registry
   * @exception RemoteException
   *   If the registry could not be exported
   * @exception JgroupException
   *   Generic exception raised by Jgroup
   * @exception ConfigurationException
   *   Raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException, ConfigurationException
  {
    this(null, csf, ssf);
  }

  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on the port specified within the configuration
   * object.  Objects contained in the <tt>conf</tt> map are used by Jgroup to
   * configure the group communication system; the distributed system
   * description must be included in this map object.
   *
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param csf
   *   Client-side <code>Socket</code> factory used to make
   *   connections to the registry
   * @param ssf
   *   Server-side <code>ServerSocket</code> factory used to
   *   accept connections to the registry
   * @exception RemoteException
   *   If the registry could not be exported
   * @exception JgroupException
   *   Generic exception raised by Jgroup
   * @exception ConfigurationException
   *   Raised when the distributed system configuration object contains an error
   */
  public RegistryImpl(GroupManager gm, RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException, ConfigurationException
  {
    registrytable = new RegistryTable();
    if (gm == null) {
      gm = GroupManager.getGroupManager(this);
    }
    egmis = (ExternalGMIService) gm.getService(ExternalGMIService.class);
    pgms = (MembershipService) gm.getService(MembershipService.class);
    me = pgms.getMyIdentifier();
    if (log.isDebugEnabled())
      log.debug("Obtained identifier: " + me);
    AppConfig app = AppConfig.getApplication(this);
    int gid = app.getGroupId();
    if (GroupManager.contains(gid)) {
      /* The group manager is for the dependable registry only. */
      pgms.join(gid);
    }
    Registry reg = BootstrapRegistry.getLocalRegistry(csf, ssf);
    Remote handler = egmis.getBootstrapHandler();
    reg.rebind(DEPENDABLE_REGISTRY_NAME, handler);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Main
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Main method of the class
   */
  public static void main(String argv[])
  {
    // Start registry
    try {
      ConfigManager.init();
      DependableRegistry registry = new RegistryImpl();
      log.info("Dependable registry ready...");
      while (true) {
        try { Thread.sleep(0x7fffffffffffffffL);
        } catch(InterruptedException ex) {}
      }
    } catch(Exception exception) {
      Abort.exit("Failed to start the dependable registry", exception, 1);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DependableRegistry
  ////////////////////////////////////////////////////////////////////////////////////////////

  @Multicast public IID bind(String service, RegistryEntry entry, Class serverClass)
    throws RemoteException, AccessException
  {
    return bind(service, entry, serverClass, -1);
  }

  @Multicast public IID bind(String service, RegistryEntry entry, Class serverClass, long leaseTime)
    throws RemoteException, AccessException
  {
    IID iid = egmis.getIdentifier();
    if (log.isDebugEnabled())
      log.debug("bind(): " + service + " " + iid.getVmid());
    checkAccess("Registry.bind");
    BindingList bl = new BindingList(service, entry, serverClass, iid, leaseTime);
    registrytable.add(iid.getVmid(), bl);
    return iid;
  }

  @Multicast public void unbind(IID iid)
    throws RemoteException, NotBoundException, AccessException
  {
    if (log.isDebugEnabled())
      log.debug("Registry.unbind() " + iid);
    checkAccess("Registry.unbind");
    registrytable.remove(iid.getVmid(), iid.getCounter());
  }

  @Anycast public String[] list()
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("Registry.list()");
    String[] services = registrytable.list();
    for (int i = 0; i < services.length; i++) {
      try {
        clearExpiredLeases(services[i]);
      } catch (NotBoundException nbe) {
        log.error("Internal error: " + nbe.getMessage());
      }
    }
    return registrytable.list();
  }

  @Anycast public Remote lookup(String serviceName)
    throws RemoteException, NotBoundException
  {
    clearExpiredLeases(serviceName);
    if (log.isDebugEnabled())
      log.debug("Registry.lookup(): " + serviceName);
    ServiceData service  = registrytable.getService(serviceName);
    RegistryEntry entry = service.getRegistryEntry();
    InvocationHandler handler = entry.createInvocationHandler(serviceName, service.getEntries());
    return Exporter.getProxy(service.getServerClass(), handler);
  }

  @Multicast public void refresh(IID iid)
    throws RemoteException, AccessException, NotBoundException
  {
    /* Updates the timestamp in the BindingList node for the given IID. */
    BindingList bl = registrytable.lookup(iid.getVmid(), iid.getCounter());
    bl.timestamp = System.currentTimeMillis();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void viewChange(View view)
  {
    if (log.isDebugEnabled())
      log.debug("--- RegistryImpl.viewChange: ---" + view);
    registrytable.viewChange(view);
  }

  public void hasLeft() { }

  public void prepareChange() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ShutdownListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void shutdown()
  {
    try {
      if (log.isDebugEnabled())
        log.info("RegistryImpl.shutdown: commenced...");
      BootstrapRegistry.unbind(DEPENDABLE_REGISTRY_NAME);
      pgms.leave();
      if (log.isDebugEnabled())
        log.info("RegistryImpl.shutdown: completed: " + DEPENDABLE_REGISTRY_NAME);
    } catch (Exception e) {
      log.warn("Unable to shutdown RegistryImpl", e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MergingListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  public Object getState(MemberId[] dests)
  {
    if (log.isDebugEnabled()) {
      for (int i=0; i < dests.length; i++)
        log.debug("getState dests["+i+"]: "+dests[i]);
    }
    return registrytable.getState(dests, me);
  }

  public void putState(Object obj, MemberId[] sources)
  {
    if (log.isDebugEnabled()) {
      for (int i=0; i < sources.length; i++)
        log.debug("putState sources["+i+"]: "+sources[i]);
    }
    try {
      registrytable.putState((MergingData) obj, sources);
    } catch (Exception e) {
      log.error("Error when receiving a putState invocation", e);
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Remove all expired leases for the given service.
   *
   *  @param service
   *    The service for which to remove expired leases.
   *  @throws RemoteException
   *  @throws NotBoundException
   */
  private void clearExpiredLeases(String service)
    throws RemoteException, NotBoundException
  {
    /* Gets IID and timestamp for the provided service name */
    LeaseInfo leaseInfo = registrytable.getLookupInf(service);
    /*
     * Check if server has refreshed the lease according to given leastime.
     * If leastime is set to -1, LeaseLayer is not used.
     */
    if (leaseInfo.getLeasetime() != -1) {
      for (int j = 0; j < leaseInfo.size(); j++) {
        if (leaseInfo.leaseExpired(j)) {
          unbind(leaseInfo.getIID(j));
        }
      }
    }
  }

  private static void checkAccess(String s)
    throws AccessException
  {
    // TODO: Currently, bind is always permitted; we need to think if it is necessary to
    // restrict the access to the hosts in the configuration file.
    return;
  }

} // END RegistryImpl
