/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.net.UnknownHostException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.Registry;
import java.rmi.server.RMIClientSocketFactory;
import java.rmi.server.RMIServerSocketFactory;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import jgroup.core.ConfigManager;
import jgroup.core.ConfigurationException;
import jgroup.core.JgroupException;
import jgroup.core.ExternalGMIService.BootstrapInvocationHandler;
import jgroup.core.registry.BootstrapRegistry;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.LookupRegistry;
import jgroup.core.registry.RegistryLocator;
import jgroup.relacs.config.AppConfig;
import jgroup.relacs.config.Host;
import jgroup.relacs.config.HostSet;
import jgroup.relacs.rmi.Exporter;
import jgroup.util.Util;

import org.apache.log4j.Logger;

/**
 *  <code>RelacsRegistryLocator</code> is used to obtain a bootstrap
 *  reference to the Jgroup dependable registry, or to create a Jgroup
 *  dependable registry instance on the local machine.  Every instance
 *  of the Jgroup registry service is based on two remote registries:
 *  the standard local registry and the replicated one.  The standard
 *  local registry is used to retrieve a reference for the replicated
 *  remote registry.  The access port for these registries are
 *  configured in the XML configuration file.
 *
 *  @see <a href="http://java.sun.com/j2se/1.4.1/docs/api/java/rmi/registry/LocateRegistry.html">java.rmi.LocateRegistry</a>
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.3
 */
public final class RelacsRegistryLocator
  implements RegistryLocator
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(RelacsRegistryLocator.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Variable for keeping the current group proxy for the dependable
   *  registry instances.  It is returned when multiple instances query
   *  for a dependable registry within the same JVM.
   */
  private DependableRegistry drProxy = null;

  /**
   *  A list of threads used for polling (checking) for dependable
   *  registry instances within the current distributed system.
   */
  private List<RegistryPollingThread> pollingThreads =
    Collections.synchronizedList(new ArrayList<RegistryPollingThread>(16));

  /**
   *  The set of group proxy handlers for the dependable registry.
   */
  private List<BootstrapInvocationHandler> handlers =
    new ArrayList<BootstrapInvocationHandler>(5);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Will be used to obtain an instance of RegistryLocator.
   */
  public RelacsRegistryLocator() { }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a proxy for the dependable registry service.
   *
   *  @return 
   *    A proxy for the dependable registry service.
   *
   *  @exception RemoteException 
   *    Raised if the proxy could not be obtained.
   *  @exception ConfigurationException
   *    Raised if there is a problem with the distributed system
   *    configuration data.
   */
  public DependableRegistry getRegistry()
    throws JgroupException, RemoteException
  {
    return getRegistry(null);
  }


  /**
   *  Returns a proxy for the dependable registry service formed by all
   *  registry replica instances found within the distributed system
   *  configuration. <p>
   *
   *  Communication with this remote registry will use the supplied
   *  <t>RMIClientSocketFactory csf</t> to create socket connections to
   *  the dependable registry on the distributed system and port. <p>
   *
   *  Note that the method returns the same group reference (variable
   *  <code>drProxy</code>) for the dependable registry upon multiple
   *  invocations of this method.  However, the threads continue to run
   *  also after the first invocation, and thus the group proxy may be
   *  augmented with further dependable registry replicas.  This means
   *  that, although the first (few) use(s) of the returned group proxy
   *  may have accessed only a subset of the available DR replicas,
   *  future invocations on the group proxy will include all available
   *  DR replicas.
   *
   *  @param csf
   *    Client-side socket factory used to make connections to the
   *    dependable registry.  If <code>csf</code> is null, then the
   *    default client-side socket factory will be used.
   *
   *  @return 
   *    A proxy for the dependable registry service
   *
   *  @exception RemoteException 
   *    If a proxy could not be obtained; probably due to unavailability
   *    of any registry replica instances.
   *  @exception ConfigurationException
   *    Raised if there is a problem with the distributed system
   *    configuration data.
   */
  public DependableRegistry getRegistry(RMIClientSocketFactory csf)
    throws JgroupException, RemoteException
  {
    if (drProxy != null) 
      return drProxy;

    /* 
     * Obtains remote references for the dependable registry and inserts
     * them in the handler.  This is achieved by probing each host in
     * the distributed system, using a separate thread for each host.
     * This is to avoid blocking when a host is not responsive.
     */
    HostSet allHosts = ConfigManager.getDistributedSystem().getAllHosts();
    for (Iterator hostIter = allHosts.iterator(); hostIter.hasNext(); ) {
      Host host = (Host) hostIter.next();
      String fqdn = host.getCanonicalHostName();
      /* Create a polling thread for each host in the distributed system. */
      RegistryPollingThread regPoller = new RegistryPollingThread(fqdn, csf);
      Thread t = new Thread(regPoller);
      t.setDaemon(true);
      t.start();
      pollingThreads.add(regPoller);
    }

    /*
     * Here we sleep a bit, to allow (all) registry instances to return
     * their <code>Remote</code> reference.
     */
    AppConfig app = AppConfig.getApplication(RegistryImpl.class);
    int locatorProbeTimeout = app.getIntParam("LocatorProbeTimeout", 1000);
    Util.sleep(locatorProbeTimeout);
    if (handlers.size() < 1) {
      Util.sleep(2*locatorProbeTimeout);
      if (handlers.size() < 1)
        throw new RemoteException("Giving up: No registry instances found.");
    }
    if (log.isDebugEnabled()) {
      log.debug("Number of hosts not responded yet: " + pollingThreads.size()
                + ", registries found thus far: " + handlers.size());
    }

    BootstrapInvocationHandler handler = handlers.get(0);
    for (int i = 1; i < handlers.size(); ++i) {
      handler.merge(handlers.get(i));
    }
    /* Returns the proxy for the registry, using the handler */
    drProxy = (DependableRegistry) Exporter.getProxy(RegistryImpl.class, handler);
    return drProxy;
  }


  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry createRegistry()
    throws JgroupException, RemoteException
  {
    return new RegistryImpl();
  }


  /**
   *  Creates and exports a dependable registry replica on the local
   *  host listening for incoming requests on an the port specified in
   *  the system configuration file. <p>
   *
   *  The new instance of the dependable registry listens to incoming
   *  requests using a <code>ServerSocket</code> created from the
   *  supplied <code>RMIServerSocketFactory</code>.  A client that
   *  receives a reference to this registry will then use a
   *  <code>Socket</code> created from the supplied
   *  <code>RMIClientSocketFactory</code>.
   *
   *  @param csf
   *    Client-side <code>Socket</code> factory used to make connections
   *    to the registry.
   *  @param ssf
   *    Server-side <code>ServerSocket</code> factory used to accept
   *    connections to the registry.
   *
   *  @return 
   *    The newly created dependable registry replica.
   *
   *  @exception RemoteException
   *    Raised if the registry could not be exported.
   *  @exception JgroupException
   *    Raised if a registry instance could not be created.
   *  @exception ConfigurationException
   *    Raised if there was a problem parsing the distributed system
   *    configuration file.
   */
  public DependableRegistry createRegistry(RMIClientSocketFactory csf, RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException
  {
    return new RegistryImpl(csf, ssf);
  }


  /* (non-Javadoc)
   * @see jgroup.core.registry.RegistryLocator#getLookupRegistry()
   */
  public LookupRegistry getLookupRegistry() 
    throws JgroupException, RemoteException
  {
    return (LookupRegistry) getRegistry();
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Probe the specified host for a dependable registry replica, and
   *  returns a remote reference to the dependable registry replica if
   *  such replica exists on the given host; otherwise an exception is
   *  thrown. <p>
   *
   *  Note that the <code>lookup</code> method used within this method
   *  may block if there is a communication problem.
   *
   *  @param hostname
   *    The host to probe for a dependable registry replica.
   *  @param csf
   *    The <code>RMIClientSocketFactory</code> to use.
   *
   *  @return 
   *    The <code>Remote</code> object reference to the registry replica.
   *
   *  @exception RemoteException
   *    Is raised if there was communication problems, such as the host
   *    is not running any bootstrap registry, or other communication
   *    problems.
   *  @exception NotBoundException
   *    Is raised if the bootstrap registry did not contain a reference
   *    to a dependable registry replica.
   */
  private static BootstrapInvocationHandler getSingleRegistry(String hostname, RMIClientSocketFactory csf)
    throws UnknownHostException, RemoteException, NotBoundException
  {
    Registry registry = BootstrapRegistry.getRemoteRegistry(hostname, csf, false);
    return (BootstrapInvocationHandler) registry.lookup(DependableRegistry.DEPENDABLE_REGISTRY_NAME);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private internal class
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  A <code>RegistryPollingThread</code> class is created for each
   *  host on which there may be running a dependable registry instance.
   *  It will query the specified host for its local registry instance
   *  and if there is a communication problem (long delay), the thread
   *  will simply block.  Since each host is being queried by a separate
   *  thread, those threads without communication problems will return
   *  quickly, allowing the system to avoid problems with blocking in
   *  the event of communication problems (due to the lookup problem
   *  with standard Java RMI.) <p>
   *
   *  @author Hein Meling
   *  @since Jgroup 1.2
   */
  private class RegistryPollingThread
    implements Runnable
  {
    private String fqdn;
    private RMIClientSocketFactory csf;

    RegistryPollingThread(String fqdn, RMIClientSocketFactory csf)
    {
      this.fqdn = fqdn;
      this.csf = csf;
    }

    /**
     *  Get the <code>Remote</code> reference for a registry running on
     *  the host associated with this thread instance, and if there was
     *  a registry, insert that <code>Remote</code> reference into the
     *  <code>handler</code> object.  Given that the method
     *  getSingleRegistry does not block, the run method will simply
     *  finish by removing this thread from the polling list and exit.
     */
    public void run()
    {
      try {
        handlers.add(getSingleRegistry(fqdn, csf));
        if (log.isDebugEnabled())
          log.debug(fqdn + ": dependable registry found.");
      } catch (Exception e) {
        if (log.isDebugEnabled())
          log.debug(fqdn + ": dependable registry not found.");
      } finally {
        pollingThreads.remove(this);
      }
    }
  }

} // END RelacsRegistryLocator
