/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.registry;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import jgroup.core.registry.DependableRegistry.RegistryEntry;
import jgroup.relacs.config.AppConfig;


/**
 *  In a registry, this class is used to contain the data associated 
 *  to a service name.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since  Jgroup 0.5
 */
final class ServiceData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** List of interfaces implementented by this object */
  private Class serverClass;

  /** The name of the service */
  private String serviceName;

  /** Type of entries */
  private RegistryEntry entriesType;

  /** List of entries */
  private final List<BindingList> list = new LinkedList<BindingList>();

  /** Boolean for determining if a server class is upgradeEnabled */
  private boolean upgradeEnabled = false;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Builds a new ServiceData object associated with the given service.
   *
   *  @param serverClass
   *    The class of the remote interface associated with this object.
   *  @param serviceName
   *    The name of the associated service.
   * @param entriesType
   *    The type of entries to store
   */
  ServiceData(Class serverClass, String serviceName, RegistryEntry entriesType)
  {
    this.serviceName = serviceName;
    this.serverClass = serverClass;
    this.entriesType = entriesType;

    try {
      AppConfig app = AppConfig.getApplication(serviceName);
      if (app.hasService("Upgrade")) {
        upgradeEnabled = true;
      }
    } catch (NullPointerException e) {
      /*
       * The service name for upgrade service is not defined in apps.xml.
       * Catch it here and do nothing.
       */
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Inserts a new binding related to this service. 
   *  
   *  @param binding the binding to be inserted
   *  @throws RemoteException raised if the new binding is related to
   *    a server whose implementation class is different from those
   *    inserted in this group.
   */
  void add(BindingList binding)
    throws RemoteException
  {
    if (serverClass != null) {
      if (!entriesType.getType().equals(binding.entry.getType())) {
        throw new RemoteException("Service name " + binding.servicename +
          " uses entries of type " + entriesType.getType());
      }

      if (checkCompatibility(binding.serverClass)) {
        list.add(binding);
      } else {
        throw new RemoteException("Service name " + binding.servicename + 
          " is reserved for members compatible with class " + serverClass.getName());
      }
    }
  }

  /**
   *  Returns true if the server class to be bound is
   *  considered compatible with the service interface
   *  represented by this <code>ServiceData</code> object.
   *
   *  @param serverClassToBind
   *    The server class being bound to this service.
   *  @return
   *    True is returned if the server class to be bound is
   *    considered compatible with the service interface
   *    represented by this <code>ServiceData</code> object.
   */
   private boolean checkCompatibility(Class serverClassToBind)
   {
     if (serverClass.equals(serverClassToBind)) {
       return true;
     } else if (upgradeEnabled) {
       // FIXME: add check whether the same application interface is implemented.
       // Solution proposal: check all the interfaces implemented by both classes
       // and compare the sets.  In the current implemenation we check only whether
       // ExternalGMIListener is implemented and assume that the application service
       // is implemented as well.  The latter is done elsewhere.
       return true; 
     } else {
       return false;
     }
   }

  /**
   *  Remove a binding from the list.
   *  
   *  @param the binding to be removed
   */
  void remove(BindingList bl)
  {
    list.remove(bl);
  }

  /**
   *  Returns true if the are no bindings related to this service
   */
  boolean isEmpty()
  {
    return (list.size() == 0);
  }

  /**
   *  Returns the BindingList nodes for the given service 
   *  data object.
   */
  public List getLinkedList()
  {
    return list;
  }

  /**
   * Return entries type
   */
  public RegistryEntry getRegistryEntry() 
  {
    return entriesType;
  }
  
  /**
   * Get iterator to all entries
   * @return <code>Iterator</code> for iterating over all entries
   */
  synchronized Iterator getEntries()
    throws NotBoundException
  {
    final Iterator iter = list.iterator();
    return new Iterator() {
      
      public void remove() { throw new UnsupportedOperationException(); }
      public boolean hasNext() { return iter.hasNext(); }
      
      public Object next() { 
        BindingList b = (BindingList)iter.next();
        return b.entry;
      }
    };
  }

  /**
   * Returns the serverClass.
   * @return Class
   */
  public Class getServerClass()
  {
    return serverClass;
  }

  /**
   * Returns the serviceName.
   * @return String
   */
  public String getServiceName()
  {
    return serviceName;
  }

} // End ServiceData
