/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.simulator;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.ConnectException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import jgroup.core.registry.BootstrapRegistry;
import jgroup.util.log.Eventlogger;
import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;

import org.apache.log4j.Logger;

/**
 *  Implementation of the <code>SocketStatus</code> interface,
 *  allowing an external entity (the simulator) to set the status
 *  of particular links from the JVM hosting this instance of the
 *  SocketStatusImpl.  Note that objects within the local JVM may
 *  obtain a reference to a "singlton" instance of this class, and
 *  thus query the link connectivity map through the
 *  <code>isReliable()</code> method.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.5
 */
public class SocketStatusImpl
  implements SocketStatus
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(SocketStatusImpl.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  private static Random random = new Random();

  /** The local socket status object */
  private static SocketStatusImpl status = null;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Connectivity map */
  private transient Map<InetAddress,Integer> links = new HashMap<InetAddress, Integer>(20);

  /** Connectivity map */
  private transient Map<InetAddress,Integer> committedLinks = null;

  /** SocketStatus objects listening for changes in the reachability patteren */
  private transient List<SocketStatus> listeners = new ArrayList<SocketStatus>(3);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an instance of the <code>SocketStatusImpl</code> class.
   * It is a private constructor to avoid construction from the outside.
   * Use the <code>instance()</code> method to obtain the local single
   * instance of the socket status class.  This construtor is used when
   * we know that there is no other <code>SocketStatus</code> instances
   * on the local machine, and we can safely export and bind this socket
   * status instance.
   */
  private SocketStatusImpl()
    throws RemoteException
  {
    Exporter exporter =
      new BasicJeriExporter(TcpServerEndpoint.getInstance(0), new BasicILFactory());
    Remote socketStatusProxy = exporter.export(this);
    try {
      BootstrapRegistry.bind(SOCKET_STATUS, socketStatusProxy);
      if (log.isDebugEnabled())
        log.debug("Primary SocketStatus exported and bound...");
    } catch (Exception e) {
      log.warn("Failed to initialize the SocketStatusImpl class", e);
      exporter.unexport(true);
    }
  }

  /**
   * This constructor is used when there exists another instance of the
   * <code>SocketStatus</code>, possibly within another JVM on this machine.
   * This instance of <code>SocketStatus</code> will export itself and
   * register with the existing <code>SocketStatus</code> in order to be
   * notified of changes to the reachability.
   * 
   * @param st
   *   Reference to the existing <code>SocketStatus</code> remote object.
   * @throws RemoteException
   */
  private SocketStatusImpl(SocketStatus st)
    throws RemoteException
  {
    Exporter exporter =
      new BasicJeriExporter(TcpServerEndpoint.getInstance(0), new BasicILFactory());
    Remote ssProxy = exporter.export(this);
    st.register((SocketStatus) ssProxy);
    committedLinks = st.getStatus();
    if (log.isDebugEnabled())
      log.debug("Secondary SocketStatus object registered with the primary SocketStatus...");
  }

  /**
   *  Obtain the local instance of the socket status object.
   */
  public static SocketStatusImpl instance()
  {
    if (status != null) {
      return status;
    }
    try {
      try {
        SocketStatus st = (SocketStatus) BootstrapRegistry.lookup(SOCKET_STATUS);
        status = new SocketStatusImpl(st);
      } catch (NotBoundException e) {
        // Since we are the first SocketStatus object available on this host.
        status = new SocketStatusImpl();
      }
    } catch (RemoteException e) {
      log.info("Partition simulator not available", e);
      return null;
    }
    log.info("Partition simulator activated");
    return status;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public access methods for the SocketStatusImpl class
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Check if the link to the given internet address is considered
   *  reliable.
   */
  public boolean isReliable(InetAddress address)
  {
    if (committedLinks == null) {
      // no link probabilities have yet been committed
      return true;
    }
    Integer prob = (Integer) committedLinks.get(address);
    int val = (int) (random.nextFloat() * BASE);
    return (prob == null || val < prob.intValue());
  }


  /**
   *  Check if the link to the given internet host is considered
   *  reliable.
   */
  public boolean isReliable(String hostname)
    throws UnknownHostException
  {
    InetAddress inetAdr = InetAddress.getByName(hostname);
    return isReliable(inetAdr);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Remote methods from the SocketStatus interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* (non-Javadoc)
   * @see jgroup.relacs.simulator.SocketStatus#register(jgroup.relacs.simulator.SocketStatus)
   */
  public void register(SocketStatus sStatus)
    throws RemoteException
  {
    if (!listeners.contains(sStatus))
      listeners.add(sStatus);
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.simulator.SocketStatus#commit()
   */
  public void commit()
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("commit received");
    doCommit();
  }

  private void doCommit()
    throws RemoteException
  {
    committedLinks = new HashMap<InetAddress,Integer>(links);
    for (Iterator<SocketStatus> iter = listeners.iterator(); iter.hasNext();) {
      SocketStatus st = iter.next();
      try {
        st.notifyStatus(committedLinks);
      } catch (ConnectException e) {
        log.warn("Remote SocketStatus object no longer present", e);
        /*
         * Since the remote (machine local) socket status object appears
         * to have crashed, we remove it from the set of listeners.
         */
        iter.remove();
      }
    }
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.simulator.SocketStatus#notifyStatus(java.util.HashMap)
   */
  public void notifyStatus(Map<InetAddress,Integer> committedLinks)
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("notifyStatus received: " + committedLinks);
    this.committedLinks = committedLinks;
    // The following may occur multiple times depending on the number of
    // listeners registered for link status events.
    if (Eventlogger.ENABLED)
      Eventlogger.logEventFlush("LinksCommitted");
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.simulator.SocketStatus#getStatus()
   */
  public Map<InetAddress,Integer> getStatus()
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("getStatus received: " + committedLinks);
    return committedLinks;
  }

  /* (non-Javadoc)
   * @see jgroup.relacs.simulator.SocketStatus#setStatus(java.net.InetAddress, int)
   */
  public void setStatus(InetAddress inet, int prob)
    throws RemoteException
  {
    if (log.isDebugEnabled())
      log.debug("changed the reachability: "+inet+" "+prob);
    if (prob >= BASE)
      links.remove(inet);
    else
      links.put(inet, prob);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    StringBuilder buf = new StringBuilder("[SocketStatusImpl: ");
    buf.append("links.size=" + links.size());
    if (committedLinks != null) {
      buf.append(", committedLinks.size=" + committedLinks.size());
    }
    buf.append("]");
    return buf.toString();
  }

} // END SocketStatusImpl
