/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import jgroup.core.EndPoint;
import jgroup.core.MemberId;
import net.jini.jeri.Endpoint;
import net.jini.jeri.tcp.TcpEndpoint;

/**
 *  Implementation of the <code>jgroup.core.MemberId</code> interface.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.9
 */
final public class MemberIdImpl
  implements MemberId
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6545889514187752437L;

  public static final int SIZE = EndPointImpl.SIZE + IncarnationId.SIZE + LocalId.SIZE;

 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Endpoint of the machine hosting the Jgroup Daemon */
  private EndPoint endpoint;
  
  /** Value used to distinguish different incarnation of the same host */
  private int incarnation;
  
  /** The local identifier for a member associated with a Jgroup daemon */
  private LocalId localId;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization
   */
  public MemberIdImpl() { }


  /**
   *  Builds a new <code>MemberId</code> with the specified arguments.
   */
  public MemberIdImpl(EndPoint endpoint, int incarnation,
    EndPoint localEndPoint, int count)
  {
    this.endpoint = endpoint;
    this.incarnation = incarnation;
    this.localId = new LocalId(count, localEndPoint);
  }


  /**
   *  Builds a new <code>MemberId</code> with the specified arguments.
   */
  public MemberIdImpl(EndPoint endpoint, int incarnation, LocalId localId)
  {
    this.endpoint = endpoint;
    this.incarnation = incarnation;
    this.localId = localId;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MemberId interface
  ////////////////////////////////////////////////////////////////////////////////////////////


  /* (non-Javadoc)
   * @see jgroup.core.MemberId#getEndPoint()
   */
  public EndPoint getEndPoint()
  {
    return endpoint;
  }
  
  /* (non-Javadoc)
   * @see jgroup.core.MemberId#getTcpEndpoint()
   */
  public Endpoint getTcpEndpoint()
  {
    String hostname = getCanonicalHostName();
    int port = getServerPort();
    return TcpEndpoint.getInstance(hostname, port);
  }

  /* (non-Javadoc)
   * @see jgroup.core.MemberId#getLocalId()
   */
  public LocalId getLocalId()
  {
    return localId;
  }

  /* (non-Javadoc)
   * @see jgroup.core.MemberId#getCanonicalHostName()
   */
  public String getCanonicalHostName()
  {
    return localId.getEndpoint().getAddress().getCanonicalHostName();
  }

  /* (non-Javadoc)
   * @see jgroup.core.MemberId#getServerPort(int)
   */
  public int getServerPort()
  {
    /*
     * The port to use is deterministic, and associated directly
     * with the local member, which may be in the same JVM
     * as the Jgroup Daemon (or may not be in the same JVM).
     */
    return localId.getMemberNo() + localId.getEndpoint().getPort();
  }

  /* (non-Javadoc)
   * @see jgroup.core.MemberId#isNewer(jgroup.core.MemberId)
   */
  public boolean isNewer(MemberId obj)
  {
    if (obj == null)
      return false;
    MemberIdImpl id = (MemberIdImpl)obj;
    return (incarnation > id.incarnation
      && localId.equals(id.localId)
      && endpoint.equals(id.endpoint));
  }

  /* (non-Javadoc)
   * @see jgroup.core.MemberId#isNeighbour(jgroup.core.MemberId)
   */
  public boolean isNeighbour(MemberId obj)
  {
    if (obj == null)
      return false;
    MemberIdImpl id = (MemberIdImpl)obj;
    return (incarnation == id.incarnation
      && localId.equals(id.localId)
      && endpoint.equals(id.endpoint));
  }


  /**
   *  Compares this member with the given member for the purpose of sorting.
   *  This is used to define th partial order for all the members in a group.
   *  <p>
   * 
   *  @return
   *     Returns a positive value if and only if this <code>EndPoint</code> is
   *     greater than the given one, or this <code>EndPoint</code> is equal to
   *     the given one and this local id is greater than the given one. <p>
   * 
   *     We return zero if the members have the same id; that is the same EndPoint
   *     and the same localId.  We do not consider the incarnation value. <p>
   * 
   *     Otherwise, we returns a negative value.
   *
   *  @see jgroup.core.MemberId
   */
  public int compareTo(MemberId member)
  {
    if (this == member) {
      return 0;
    } else {
      int epValue = endpoint.compareTo(member.getEndPoint());
      if (epValue == 0) {
        return (localId.getMemberNo() - member.getLocalId().getMemberNo());
      } else {
        return epValue;
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder("[");
    buf.append(endpoint);
    if (!localId.getEndpoint().equals(endpoint)) {
      buf.append(", ");
      buf.append(localId);
    } else {
      buf.append(".");
      buf.append(localId.getMemberNo());
    }
//    buf.append(", ");
//    buf.append(incarnation);
    buf.append("]");
    return buf.toString();
  }


  /**
   *  Returns a hashcode for the <code>MemberId</code>.  Two
   *  <code>MemberId</code>s will have the same hashcode if they are
   *  equal with respect to their content.
   *
   *  @return the hashcode
   */
  public int hashCode()
  {
    return endpoint.hashCode() + incarnation + localId.hashCode();
  }

  /**
   *  Compares two <code>MemberIdImpl</code>s for content equality.
   *
   *  @param obj
   *    The object to compare with
   *  @return 
   *    True if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof MemberIdImpl))
      return false;
    MemberIdImpl id = (MemberIdImpl) obj;
    return (incarnation == id.incarnation
      && localId.equals(id.localId)
      && endpoint.equals(id.endpoint));
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for externalization
  ////////////////////////////////////////////////////////////////////////////////////////////

  
  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
  {
    endpoint = new EndPointImpl();
    endpoint.readExternal(in);
    incarnation = IncarnationId.unmarshal(in);
    localId = new LocalId();
    localId.readExternal(in);
  }

  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    endpoint.writeExternal(out);
    IncarnationId.marshal(out, incarnation);
    localId.writeExternal(out);
  }

} // END MemberIdImpl
