/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.relacs.types;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.net.UnknownHostException;

import jgroup.core.VMID;
import jgroup.util.Abort;
import jgroup.util.Network;

/**
 *  The <code>VMIDImpl</code> class implements the <code>VMID</code>
 *  interface to uniquely identify a Java Virtual Machine.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.7
 */
public final class VMIDImpl
  implements VMID
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = 6243630502385386084L;

  /** IP address of the generating host */
  private InetAddress inetAddress;

  /** Integer representation of the above IP address */
  private int address;

  /** Random number to distinguish concurrent JVMs on the same host */
  private int hash;

  /** Time when this <code>VMID</code> was generated */
  private long time;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Unique virtual machine identifier for the local host */
  private static VMID vmid;

  /*
   *  Create a virtual machine identifier for the local host.
   */
  static {
    try {
      InetAddress inetAddress = Network.getLocalHost();
      long time = System.currentTimeMillis();
      int hash = (new Object()).hashCode();
      vmid = new VMIDImpl(inetAddress, time, hash);
    } catch (Exception e) {
      Abort.exit("Failed to generate an identifier for the virtual machine", e, 1);
    }
  }


  /**
   *  Returns the <code>VMID</code> for the local Java Virtual Machine.
   */
  public static VMID getLocalVMID()
  {
    return vmid;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Public default constructor for externalization.
   */
  public VMIDImpl()
  {
  }

  /**
   *  Private constructor.
   */
  private VMIDImpl(InetAddress inetAddress, long time, int hash)
  {
    this.inetAddress = inetAddress;
    this.time = time;
    this.hash = hash;
    this.address = Network.translate(inetAddress);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the VMID interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the address of the host in which the virtual machine
   *  identified by this <code>VMID</code> is executed.
   */
  public InetAddress getAddress()
    throws UnknownHostException
  {
    /*
     * If the inetAddress is null it means that this object must have
     * been serialized, and thus we only have an integer representation
     * of this JVMs internet address.  Thus, we convert the int
     * representation to an InetAddress representation.
     */
    if (inetAddress == null)
      inetAddress = Network.translate(address);
    return inetAddress;
  }


  /**
   *  Returns the time at which this <code>VMID</code> has been created.
   */
  public long getTime()
  {
    return time;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object.
   */
  public String toString()
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[VMID: ");
    buf.append(Network.translateToString(address));
    buf.append(", ");
    buf.append(time);
    buf.append(", ");
    buf.append(hash);
    buf.append("]");
    return buf.toString();
  }


  /**
   *  Returns a hashcode for the <code>VMIDImpl</code>.  Two
   *  <code>VMIDImpl</code>s will have the same hashcode if they are
   *  equal with respect to their content.
   */
  public int hashCode()
  {
    return address + (int) time + hash;
  }


  /**
   *  Compares two objects for content equality.
   *
   *  @param obj the object to compare with
   *  @return true if these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof VMIDImpl))
      return false;
    VMIDImpl vmid = (VMIDImpl) obj;
    return (hash == vmid.hash && address == vmid.address && time == vmid.time);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Write the content of this VMID on stream <tt>out</tt>.  Note that
   *  the InetAddress is not serialized, because too inefficient.  If
   *  requested, the InetAddress is retrieved through DNS lookup in
   *  method <tt>getAddress</tt>.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(hash);
    out.writeLong(time);
    out.writeInt(address);
  }

  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   * 
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    hash = in.readInt();
    time = in.readLong();
    address = in.readInt();
  }

} // END VMIDImpl
