/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.upgrade;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import jgroup.core.ConfigManager;
import jgroup.core.GroupManager;
import jgroup.core.IID;
import jgroup.core.JgroupException;
import jgroup.core.MembershipService;
import jgroup.core.arm.ExecException;
import jgroup.core.arm.UnknownGroupException;
import jgroup.core.protocols.Multicast;
import jgroup.core.registry.DependableRegistry;
import jgroup.core.registry.RegistryService;
import jgroup.relacs.config.AppConfig;

import org.apache.log4j.Logger;


/**
 *
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class UpgradeManagerImpl
  implements UpgradeManager
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Logger
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Obtain logger for this class */
  private static final Logger log = Logger.getLogger(UpgradeManagerImpl.class);


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final String upgrSrvName = UpgradeService.UPGRADE_SERVICE + "-";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The application information for the upgrade manager application. */
  private AppConfig thisApp;

  /** The dependable registry reference. */
  private DependableRegistry dregistry;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Construct a local <code>UpgradeManagerImpl</code> instance.
   *
   *  @exception JgroupException
   *    Raised if the upgrade manager could not be created for
   *    numerous reasons.
   *  @exception RemoteException
   *    Raised if there was a problem with the dependable registry.
   */
  public UpgradeManagerImpl(DependableRegistry dregistry)
    throws JgroupException, RemoteException
  {
    this.dregistry = dregistry;

    /* Initialize the application name for the upgrade mananger */
    ConfigManager.init();
    thisApp = AppConfig.getApplication(this);
    /*
     * Obtain a group manager for this application and the service proxy
     * for the EGMI service.
     */
    GroupManager gm = GroupManager.getGroupManager(this);
    MembershipService pgms = (MembershipService)  gm.getService(MembershipService.class);
    RegistryService registryService = (RegistryService) gm.getService(RegistryService.class);

    pgms.join(thisApp.getGroupId());

    /* Bind the upgrade manager with the dependable registry */
    IID iid = registryService.bind(thisApp.getRegistryName(), this);
    if (log.isDebugEnabled())
      log.debug("UpgradeManager started");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // UpgradeManager interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upgrade the group for the specified application, with the
   *  applications required redundancy.
   *
   *  @param app
   *    The application details required to upgrade the group.
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  @Multicast public void upgradeGroup(AppConfig app)
    throws RemoteException, UnknownGroupException, ExecException
  {
    if (log.isDebugEnabled())
      log.debug("Upgrading group: " + app);
    String upgradeId = upgrSrvName + app.getGroupId();
    if (log.isDebugEnabled())
      log.debug("UpgradeId: "  + upgradeId);

    try {
      UpgradeService upgrService = (UpgradeService) dregistry.lookup(upgradeId);
      if (log.isDebugEnabled())
        log.debug("Upgrade service proxy: " + upgrService);
      upgrService.upgradeRequest(app);
    } catch (NotBoundException e) {
      throw new UnknownGroupException("Group not bound with the dependable registry", app.getGroupId());
    }
    
    //joan-comment: this is the method that starts the upgrade algorithm.
  }


  /**
   *  Downgrade the group for the given application.
   *
   *  @param app
   *    The application group to remove.
   *  @exception RemoteException
   *    Raised if there was communication problems; only when invoked remotely.
   *  @exception UnknownGroupException
   *    Raised if the specified application has no group in the system.
   */
  @Multicast public void downgradeGroup(AppConfig app)
    throws RemoteException, UnknownGroupException
  {
    throw new UnsupportedOperationException();
  }

} // END UpgradeManagerImpl
