/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 *  Utility class for obtaining the local host.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 1.2
 */
public class Network
{

  /////////////////////////////////////////////////////////////////////////////
  // Static fields/block
  /////////////////////////////////////////////////////////////////////////////

  /** <CODE>InetAddress</CODE> of the local host. */
  private static InetAddress local = null;

  /** Integer version of the local host. */
  private static int packedLocal = 0;

  static {
    try {
      local = InetAddress.getLocalHost();
    } catch (UnknownHostException e) { }
    if (local == null) {
      String name = System.getProperty("jgroup.hostname");
      if (name != null && name.length() > 0) {
        try {
          local = InetAddress.getByName(name);
        } catch (UnknownHostException e) { }
      }
    }
    if (local == null)
      Abort.exit("Unable to detect local host");
    System.out.println("localhost: "+ getLocalHostName());
    packedLocal = translate(local);
  }


  /////////////////////////////////////////////////////////////////////////////
  // Static methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the internet address of the local host.
   */
  public static InetAddress getLocalHost()
  {
    return local;
  }


  /**
   *  Returns the hostname of the local host; <i>may</i> include the
   *  fully qualified domain name.
   */
  public static String getLocalHostName()
  {
    return local.getHostName();
  }


  /**
   *  Returns the machine name of the local host; <i>not</i> including
   *  the fully qualified domain name.
   */
  public static String getLocalMachineName()
  {
    return getMachineName(local.getHostName());
  }


  /**
   *  Returns the machine name of the given host; <i>not</i> including
   *  the fully qualified domain name.
   */
  public static String getMachineName(String hostName)
  {
    int machEndIndex = hostName.indexOf('.');
    if (machEndIndex == -1)
      return hostName;
    else
      return hostName.substring(0, machEndIndex);
  }


  /**
   *  Returns the domain name of the local host; <i>not</i> including
   *  the machine name part.
   */
  public static String getLocalDomainName()
  {
    String hostName = local.getHostName();
    int machEndIndex = hostName.indexOf('.');
    if (machEndIndex == -1)
      return "";
    else
      return hostName.substring(machEndIndex+1);
  }


  /**
   *  Returns true if the specified <code>address</code> is the local
   *  host; false is returned otherwise.
   */
  public static boolean isLocal(InetAddress address)
  {
    return local.equals(address);
  }

  /**
   *  Returns true if the specified <code>packedAddress</code> is the local
   *  host; false is returned otherwise.
   */
  public static boolean isLocal(int packedAddress)
  {
    return packedAddress == packedLocal;
  }


  /**
   *  Returns an integer representation of the specified
   *  <CODE>InetAddress</CODE>.  This is used for efficiency.
   */
  public static int translate(InetAddress address)
  {
    byte[] ad = address.getAddress();
    if (ad.length != 4) 
      throw new IllegalArgumentException("Jgroup accepts only 32-bit IP addresses");
    int packedAddress =
      ((ad[0] << 24) & 0xFF000000) |
      ((ad[1] << 16) & 0xFF0000) |
      ((ad[2] <<  8) & 0xFF00) |
      (ad[3] & 0xFF);
    return packedAddress;
  }


  /**
   *  Returns an <code>InetAddress</code> representation of the
   *  specified integer.  This is used for performance improvements in
   *  serialization.
   *
   *  @param packed
   *     An integer representation of an internet address.
   *  @return 
   *     The corresponding <code>InetAddress</code> representation.
   *  @exception UnknownHostException 
   *     If the specified integer does not correspond to a known host.
   */
  public static InetAddress translate(int packed)
    throws UnknownHostException
  {
    int[] ad = new int[4];
    ad[0] = (packed  >>> 24);
    ad[1] = ((packed >>> 16) & 0x000000FF);
    ad[2] = ((packed >>>  8) & 0x000000FF);
    ad[3] = (packed          & 0x000000FF);
    StringBuilder buf = new StringBuilder();
    buf.append(ad[0]);
    buf.append(".");
    buf.append(ad[1]);
    buf.append(".");
    buf.append(ad[2]);
    buf.append(".");
    buf.append(ad[3]);
    return InetAddress.getByName(buf.toString());
  }


  /**
   *  Returns an integer representation of the specified
   *  <CODE>String</CODE> IP-address.  This is used for efficiency.
   */
  public static int translateToInt(String address)
    throws UnknownHostException
  {
    return translate(InetAddress.getByName(address));
  }


  /**
   *  Returns a <code>String</code> representation of the IP-address for
   *  the specified integer.
   *
   *  @param packed
   *     An integer representation of an internet address.
   *  @return 
   *     The corresponding <code>InetAddress</code> representation.
   */
  public static String translateToString(int packed)
  {
    int[] ad = new int[4];
    ad[0] = (packed  >>> 24);
    ad[1] = ((packed >>> 16) & 0x000000FF);
    ad[2] = ((packed >>>  8) & 0x000000FF);
    ad[3] = (packed          & 0x000000FF);
    StringBuilder buf = new StringBuilder();
    buf.append(ad[0]);
    buf.append(".");
    buf.append(ad[1]);
    buf.append(".");
    buf.append(ad[2]);
    buf.append(".");
    buf.append(ad[3]);
    return buf.toString();
  }

} // END Network
