/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.UTFDataFormatException;

import jgroup.core.ConfigManager;

/**
 *  The <code>OutMessage</code> class is an output stream specially
 *  designed to be used with a transport layer which fragments messages.
 *  Messages are stored in fragments, each fragment being composed by an
 *  header, a payload and a trailer.  The length of these fields may be
 *  configured when creating a new <code>OutMessage</code>. Data written
 *  in the message are stored in the payload field of the fragments.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.8
 */
public final class OutMessage
  extends OutputStream
  implements ObjectOutput, Externalizable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final long serialVersionUID = -6820954817177822595L;

  /** Initial number of fragments composing a message */
  private static final int NUM_BUFFER = 16;     


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Temporary buffer used in writeTemp */
  private static byte[] temp = new byte[8];


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private int       payload;            // Payload size in bytes
  private int       header;             // Header size in bytes
  private int       trailer;            // Trailer size in bytes
  private int       totlen;             // Total length of a fragment
  private int       endpayload;         // Last byte used for payload

  private int       position;           // Current position in the stream [0..byteCount]
  private int       fragment;           // Current fragment number
  private int       offset;             // Current offset in the fragment
  
  private int       fragmentCount;      // Total number of fragments used
  private int       byteCount;          // Total number of bytes written
  private int       maxsize;            // Max size in bytes
  
  private byte[]    buffer;             // Current fragment
  private byte[][]  fragments;          // Byte array fragments

  /** Number of bytes written so far */
  private int written;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Construct a message with header and trailer of length zero. In
   *  this way, OutMessage may be used as a more efficient version of
   *  ByteArrayOutputStream.
   */
  public OutMessage(int payload)
  {
    this(payload, 0, 0);
  }


  /**
   *  Construct a OutMessage starting from an InMessage. In this way, it
   *  possible to forward messages received from other servers without
   *  copying them.
   */
  public OutMessage(InMessage msg)
  {
    // Store and compute fixed limits
    this.payload = msg.getPayload();
    this.header  = msg.getHeader();
    this.trailer = msg.getTrailer();
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = msg.getByteCount();
    fragmentCount = msg.getFragmentCount();
    /*
     * Patch to avoid problems when forwarding/resending a message
     * to a different set of receivers (or different cluster) than
     * the original InMessage had in its trailer.  This means that
     * the MsgJG will want to add flow control data and a different
     * receiver set to the trailer of the newly created OutMessage.
     * The size of these fields may differ (require more capacity)
     * than the original InMessage.  Note that the actual byte count,
     * and the size of the allocated fragments double array will not
     * change until more capacity is required.
     *
     * The original code looked like this:
     *   maxsize = byteCount;
     */
    maxsize = Integer.MAX_VALUE;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    // Create fragments
    fragments = msg.getFragments();
    buffer = fragments[0];
  }


  /**
   *  This constructor builds a message output stream subdivided in
   *  fragments of length <code>header + payload + trailer</code>.
   *  Used for messages whose size is initially unknown.
   *
   *  @exception IndexOutOfBoundsException
   *    Raised if <code>payload</code> is negative or equal to zero;
   *    <code>header</code> is negative; <code>trailer</code> is negative.
   */
  public OutMessage(int payload, int header, int trailer)
  {
    if (payload <= 0 || header < 0 || trailer < 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = payload;
    this.header  = header;
    this.trailer = trailer;
    endpayload = payload + header;
    totlen = endpayload + trailer;

    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 1;
    maxsize = Integer.MAX_VALUE;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    // Create fragments
    fragments = new byte[NUM_BUFFER][];
    fragments[0] = new byte[totlen];
    buffer = fragments[0];
  }


  /**
   *  This constructor builds a message output stream subdivided in
   *  fragments.  Used for messages whose size is known a priori.  If
   *  <code>maxsize</code> is less than <code>payload</code>, the
   *  payload of the message is set to <code>maxsize</code>.  The total
   *  length of each fragments is <code>header + payload +
   *  trailer</code>.
   *
   *  @exception IndexOutOfBoundsException
   *    Raised if <code>payload</code> is negative or equal to zero;
   *    <code>header</code> is negative; <code>trailer</code> is
   *    negative; <code>maxsize</code> is negative or equal to zero.
   */
  public OutMessage(int payload, int header, int trailer, int maxsize)
  {
    if (payload <= 0 || header < 0 || trailer < 0 || maxsize <= 0)
      throw new IndexOutOfBoundsException();

    // Store and compute fixed limits
    this.payload = (maxsize < payload ? maxsize : payload);
    this.header = header;
    this.trailer = trailer;
    endpayload = this.payload + header;
    totlen = endpayload + trailer;

    /*
     * Initialize variables regarding message size; the total number of
     * bytes is already set to maxsize; in this way, in the following
     * we'll skip to check whether the size exceed maxsize or not.
     */
    byteCount = maxsize;
    fragmentCount = ((maxsize-1) / this.payload) + 1;
    this.maxsize = maxsize;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;
    
    // Create fragments
    fragments = new byte[fragmentCount][];
    for (int i=0; i< fragmentCount; i++)
      fragments[i] = new byte[totlen];  
    buffer = fragments[0];
  }


  /**
   *  Clear and reset the <code>OutMessage</code> to the given new
   *  maxsize.  This occurs by reusing the already allocated memory for
   *  the fragments of this <code>OutMessage</code>.
   */
  public void clear(int newmaxsize)
  {
    if (newmaxsize > maxsize)
      throw new IllegalStateException("The newmaxsize cannot be greater than the creation time maxsize");

    /*
     * Initialize variables regarding message size; the total number of
     * byte is already set to maxsize; in this way, in the following
     * we'll skip to check whether the size exceed maxsize or not.
     */
    byteCount = newmaxsize;
    fragmentCount = ((newmaxsize-1) / this.payload) + 1;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;

    buffer = fragments[0];
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Externalizable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Default constructor for externalization.
   */
  public OutMessage() { }


  /**
   *  Restores the content of this object from the marshalled data contained
   *  in the specified input stream.
   *
   *  @param in the stream to be read
   */
  public void readExternal(ObjectInput in)
    throws IOException
  {
    payload       = in.readInt();
    header        = in.readInt();
    trailer       = in.readInt();
    endpayload    = payload + header;
    totlen        = endpayload + trailer;
    fragmentCount = in.readInt();
    byteCount     = in.readInt();
    maxsize       = in.readInt();
    written       = in.readInt();

    // Initialize variables for message position in the stream
    position = 0;
    fragment = 0;
    offset   = header;

    // Reconstruct fragments from stream
    fragments = new byte[fragmentCount][];
    int fcount = fragmentCount - 1;
    for (int i = 0; i < fcount; i++) {
      fragments[i] = new byte[totlen];
      in.readFully(fragments[i]);
    }
    fragments[fcount] = new byte[totlen];
    int lastFragLen = getLastPayload() + header + trailer;
    in.readFully(fragments[fcount], 0, lastFragLen);
    buffer = fragments[0];
//    System.out.println("outmsg READ: " + this.toString(true));
  }


  /**
   *  Marshals the content of this object to the specified output stream.
   * 
   *  @param out the stream to be written
   */
  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    out.writeInt(payload);
    out.writeInt(header);
    out.writeInt(trailer);
    out.writeInt(fragmentCount);
    out.writeInt(byteCount);
    out.writeInt(maxsize);
    out.writeInt(written);

    int fcount = fragmentCount - 1;
    for (int i = 0; i < fcount; i++) {
      out.write(fragments[i]);
    }
    int lastFragLen = getLastPayload() + header + trailer;
    out.write(fragments[fcount], 0, lastFragLen);
//    System.out.println("outmsg WRITE: " + this.toString(true));
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods inherited from OutputStream
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Writes the specified byte to this byte array output stream. 
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <code>ArrayIndexOutOfBoundsException</code> is thrown.
   *
   * @param   b   the byte to be written.
   */
  public void write(int b) 
  {
    /* Increase the number of stored bytes */
    position++;
    if (position > byteCount) {
      byteCount = position; 
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
      if (offset == endpayload) {
        /*
         * We have reached the end of the last fragment; we add a
         * another fragment to store the new byte; fragment, offset and
         * buffer are updated to point to the fist byte of the new
         * fragment.
         */
        fragment++;
        offset = header;
        enlargeSpace(byteCount);
        buffer = fragments[fragment];
      }
    }
    /* We write the byte and increment offset */
    buffer[offset++] = (byte) b;
    written += 1;
  }
  
  
  /**
   * Writes <code>len</code> bytes from the specified byte array 
   * starting at offset <code>off</code> to this msg output stream.
   * <p>
   * If <code>b</code> is <code>null</code>, a 
   * <code>NullPointerException</code> is thrown.
   * <p>
   * If <code>off</code> is negative, or <code>len</code> is negative, or 
   * <code>off+len</code> is greater than the length of the array 
   * <code>b</code>, then an <tt>IndexOutOfBoundsException</tt> is thrown.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param   b     the data.
   * @param   off   the start offset in the data.
   * @param   len   the number of bytes to write.
   */
  public void write(byte[] b, int off, int len) 
  {
    if ((off<0) || (off>b.length) || (len<0) || ((off+len) > b.length)) {
      throw new IndexOutOfBoundsException();
    } else if (len == 0) {
      return;
    } else {
      position += len;
      if (position > byteCount) {
        byteCount = position;
        if (position > maxsize) {
          throw new ArrayIndexOutOfBoundsException();
        }
      }
      if (offset + len <= endpayload) {
        System.arraycopy(b, off, buffer, offset, len);
        offset += len;
      } else {
        enlargeSpace(byteCount);
        int todo = endpayload - offset;
        System.arraycopy(b, off, buffer, offset, todo);
        int remaining = len-todo;
        off += todo;
        fragment++;  
        while (remaining > payload) {                       
          System.arraycopy(b, off, fragments[fragment], header, payload);
          remaining -= payload;
          off += payload;
          fragment++;
        }
        System.arraycopy(b, off, fragments[fragment], header, remaining);
        offset = remaining + header;
        buffer = fragments[fragment];
        
      }
    }
    written += len;
  }

  /**
   * Writes <code>b.length</code> bytes from the specified byte array 
   * to this output stream. The general contract for <code>write(b)</code> 
   * is that it should have exactly the same effect as the call 
   * <code>write(b, 0, b.length)</code>.
   * <p>
   * If the total number of bytes written in the message exceeds the maximum
   * length, a <tt>ArrayIndexOutOfBoundsException</tt> is thrown.
   *
   * @param      b   the data.
   *
   */
  public void write(byte[] b) 
  {
    write(b, 0, b.length);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from DataOutput
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Writes a <code>boolean</code> value to this output stream.
   * If the argument <code>v</code>
   * is <code>true</code>, the value <code>(byte)1</code>
   * is written; if <code>v</code> is <code>false</code>,
   * the  value <code>(byte)0</code> is written.
   * The byte written by this method may
   * be read by the <code>readBoolean</code>
   * method of interface <code>DataInput</code>,
   * which will then return a <code>boolean</code>
   * equal to <code>v</code>.
   *
   * @param      v   the boolean to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeBoolean(boolean v) throws IOException
  {
    write(v ? 1 : 0);
  }
  
  /**
   * Writes to the output stream the eight low-
   * order bits of the argument <code>v</code>.
   * The 24 high-order bits of <code>v</code>
   * are ignored. (This means  that <code>writeByte</code>
   * does exactly the same thing as <code>write</code>
   * for an integer argument.) The byte written
   * by this method may be read by the <code>readByte</code>
   * method of interface <code>DataInput</code>,
   * which will then return a <code>byte</code>
   * equal to <code>(byte)v</code>.
   *
   * @param      v   the byte value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeByte(int v) throws IOException
  {
    write(v);
  }
  
  /**
   * Writes two bytes to the output
   * stream to represent the value of the argument.
   * The byte values to be written, in the  order
   * shown, are: <p>
   * <pre><code>
   * (byte)(0xff &amp; (v &gt;&gt; 8))
   * (byte)(0xff &amp; v)
   * </code> </pre> <p>
   * The bytes written by this method may be
   * read by the <code>readShort</code> method
   * of interface <code>DataInput</code> , which
   * will then return a <code>short</code> equal
   * to <code>(short)v</code>.
   *
   * @param      v   the <code>short</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeShort(int v) throws IOException
  {
    write2(v);
  }
  
  /**
   * Writes a <code>char</code> value, wich
   * is comprised of two bytes, to the
   * output stream.
   * The byte values to be written, in the  order
   * shown, are:
   * <p><pre><code>
   * (byte)(0xff &amp; (v &gt;&gt; 8))
   * (byte)(0xff &amp; v)
   * </code></pre><p>
   * The bytes written by this method may be
   * read by the <code>readChar</code> method
   * of interface <code>DataInput</code> , which
   * will then return a <code>char</code> equal
   * to <code>(char)v</code>.
   *
   * @param      v   the <code>char</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeChar(int v) throws IOException
  {
    write2(v);
  }
  
  /**
   * Writes an <code>int</code> value, which is
   * comprised of four bytes, to the output stream.
   * The byte values to be written, in the  order
   * shown, are:
   * <p><pre><code>
   * (byte)(0xff &amp; (v &gt;&gt; 24))
   * (byte)(0xff &amp; (v &gt;&gt; 16))
   * (byte)(0xff &amp; (v &gt;&gt; &#32; &#32;8))
   * (byte)(0xff &amp; v)
   * </code></pre><p>
   * The bytes written by this method may be read
   * by the <code>readInt</code> method of interface
   * <code>DataInput</code> , which will then
   * return an <code>int</code> equal to <code>v</code>.
   *
   * @param      v   the <code>int</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeInt(int v) throws IOException
  {
    write4(v);
  }
  
  /**
   * Writes an <code>long</code> value, which is
   * comprised of four bytes, to the output stream.
   * The byte values to be written, in the  order
   * shown, are:
   * <p><pre><code>
   * (byte)(0xff &amp; (v &gt;&gt; 48))
   * (byte)(0xff &amp; (v &gt;&gt; 40))
   * (byte)(0xff &amp; (v &gt;&gt; 32))
   * (byte)(0xff &amp; (v &gt;&gt; 24))
   * (byte)(0xff &amp; (v &gt;&gt; 16))
   * (byte)(0xff &amp; (v &gt;&gt;  8))
   * (byte)(0xff &amp; v)
   * </code></pre><p>
   * The bytes written by this method may be
   * read by the <code>readLong</code> method
   * of interface <code>DataInput</code> , which
   * will then return a <code>long</code> equal
   * to <code>v</code>.
   *
   * @param      v   the <code>long</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeLong(long v) throws IOException
  {
    write8(v);
  }
  
  /**
   * Writes a <code>float</code> value,
   * which is comprised of four bytes, to the output stream.
   * It does this as if it first converts this
   * <code>float</code> value to an <code>int</code>
   * in exactly the manner of the <code>Float.floatToIntBits</code>
   * method  and then writes the <code>int</code>
   * value in exactly the manner of the  <code>writeInt</code>
   * method.  The bytes written by this method
   * may be read by the <code>readFloat</code>
   * method of interface <code>DataInput</code>,
   * which will then return a <code>float</code>
   * equal to <code>v</code>.
   *
   * @param      v   the <code>float</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeFloat(float v) throws IOException
  {
    write4(Float.floatToIntBits(v));
  }
  
  /**
   * Writes a <code>double</code> value,
   * which is comprised of eight bytes, to the output stream.
   * It does this as if it first converts this
   * <code>double</code> value to a <code>long</code>
   * in exactly the manner of the <code>Double.doubleToLongBits</code>
   * method  and then writes the <code>long</code>
   * value in exactly the manner of the  <code>writeLong</code>
   * method. The bytes written by this method
   * may be read by the <code>readDouble</code>
   * method of interface <code>DataInput</code>,
   * which will then return a <code>double</code>
   * equal to <code>v</code>.
   *
   * @param      v   the <code>double</code> value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeDouble(double v) throws IOException
  {
    write8(Double.doubleToLongBits(v));
  }
  
  /**
   * Writes a string to the output stream.
   * For every character in the string
   * <code>s</code>,  taken in order, one byte
   * is written to the output stream.  If
   * <code>s</code> is <code>null</code>, a <code>NullPointerException</code>
   * is thrown.<p>  If <code>s.length</code>
   * is zero, then no bytes are written. Otherwise,
   * the character <code>s[0]</code> is written
   * first, then <code>s[1]</code>, and so on;
   * the last character written is <code>s[s.length-1]</code>.
   * For each character, one byte is written,
   * the low-order byte, in exactly the manner
   * of the <code>writeByte</code> method . The
   * high-order eight bits of each character
   * in the string are ignored.
   *
   * @param      s   the string of bytes to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeBytes(String s) throws IOException
  {
    int len = s.length();
    for (int i = 0 ; i < len ; i++) {
      write((byte)s.charAt(i));
    }
  }
  
  /**
   * Writes every character in the string <code>s</code>,
   * to the output stream, in order,
   * two bytes per character. If <code>s</code>
   * is <code>null</code>, a <code>NullPointerException</code>
   * is thrown.  If <code>s.length</code>
   * is zero, then no characters are written.
   * Otherwise, the character <code>s[0]</code>
   * is written first, then <code>s[1]</code>,
   * and so on; the last character written is
   * <code>s[s.length-1]</code>. For each character,
   * two bytes are actually written, high-order
   * byte first, in exactly the manner of the
   * <code>writeChar</code> method.
   *
   * @param      s   the string value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeChars(String s) throws IOException
  {
    int len = s.length();
    for (int i = 0 ; i < len ; i++) {
      int v = s.charAt(i);
      write((v >>> 8) & 0xFF);
      write((v >>> 0) & 0xFF);
    }
  }
  
  /**
   * Writes two bytes of length information
   * to the output stream, followed
   * by the Java modified UTF representation
   * of  every character in the string <code>s</code>.
   * If <code>s</code> is <code>null</code>,
   * a <code>NullPointerException</code> is thrown.
   * Each character in the string <code>s</code>
   * is converted to a group of one, two, or
   * three bytes, depending on the value of the
   * character.<p>
   * If a character <code>c</code>
   * is in the range <code>&#92;u0001</code> through
   * <code>&#92;u007f</code>, it is represented
   * by one byte:<p>
   * <pre>(byte)c </pre>  <p>
   * If a character <code>c</code> is <code>&#92;u0000</code>
   * or is in the range <code>&#92;u0080</code>
   * through <code>&#92;u07ff</code>, then it is
   * represented by two bytes, to be written
   * in the order shown:<p> <pre><code>
   * (byte)(0xc0 | (0x1f &amp; (c &gt;&gt; 6)))
   * (byte)(0x80 | (0x3f &amp; c))
   *  </code></pre>  <p> If a character
   * <code>c</code> is in the range <code>&#92;u0800</code>
   * through <code>uffff</code>, then it is
   * represented by three bytes, to be written
   * in the order shown:<p> <pre><code>
   * (byte)(0xc0 | (0x0f &amp; (c &gt;&gt; 12)))
   * (byte)(0x80 | (0x3f &amp; (c &gt;&gt;  6)))
   * (byte)(0x80 | (0x3f &amp; c))
   *  </code></pre>  <p> First,
   * the total number of bytes needed to represent
   * all the characters of <code>s</code> is
   * calculated. If this number is larger than
   * <code>65535</code>, then a <code>UTFDataFormatError</code>
   * is thrown. Otherwise, this length is written
   * to the output stream in exactly the manner
   * of the <code>writeShort</code> method;
   * after this, the one-, two-, or three-byte
   * representation of each character in the
   * string <code>s</code> is written.<p>  The
   * bytes written by this method may be read
   * by the <code>readUTF</code> method of interface
   * <code>DataInput</code> , which will then
   * return a <code>String</code> equal to <code>s</code>.
   *
   * @param      str   the string value to be written.
   * @exception  IOException  if an I/O error occurs.
   */
  public void writeUTF(String str) throws IOException
  {
    int strlen = str.length();
    int utflen = 0;
    char[] charr = new char[strlen];
    int c, count = 0;
    
    str.getChars(0, strlen, charr, 0);
    
    for (int i = 0; i < strlen; i++) {
      c = charr[i];
      if ((c >= 0x0001) && (c <= 0x007F)) {
        utflen++;
      } else if (c > 0x07FF) {
        utflen += 3;
      } else {
        utflen += 2;
      }
    }
    
    if (utflen > 65535)
      throw new UTFDataFormatException();
    
    byte[] bytearr = new byte[utflen+2];
    bytearr[count++] = (byte) ((utflen >>> 8) & 0xFF);
    bytearr[count++] = (byte) ((utflen >>> 0) & 0xFF);
    for (int i = 0; i < strlen; i++) {
      c = charr[i];
      if ((c >= 0x0001) && (c <= 0x007F)) {
        bytearr[count++] = (byte) c;
      } else if (c > 0x07FF) {
        bytearr[count++] = (byte) (0xE0 | ((c >> 12) & 0x0F));
        bytearr[count++] = (byte) (0x80 | ((c >>  6) & 0x3F));
        bytearr[count++] = (byte) (0x80 | ((c >>  0) & 0x3F));
      } else {
        bytearr[count++] = (byte) (0xC0 | ((c >>  6) & 0x1F));
        bytearr[count++] = (byte) (0x80 | ((c >>  0) & 0x3F));
      }
    }
    write(bytearr);
  }

  /**
   * Returns the current value of the counter <code>written</code>, 
   * the number of bytes written to this data output stream so far.
   * Note: in this version, the counter may overflow.
   *
   * @return  the value of the <code>written</code> field.
   * @see     java.io.DataOutputStream#written
   */
  public final int size() {
    return written;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ObjectOutput
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Write an object to the underlying storage or stream.  The
   * class that implements this interface defines how the object is
   * written.
   *
   * @exception IOException Any of the usual Input/Output related exceptions.
   */
  public void writeObject(Object obj)
    throws IOException
  {
    // workaround since OutMessage don't provide native support for writeObject()
    ObjectOutputStream oos = new ObjectOutputStream(this);
    oos.writeObject(obj);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Move the current position in the stream to the specified
   *  <code>position</code>. <p>
   *
   *  @exception IndexOutOfBoundsException
   *    Raised if the new position in the message exceeds the maximum
   *    length.
   */
  public void seek(int position)
  {
    fragment = position / payload;
    offset = header + position % payload;
    this.position = position;
    /*
     * FIXED HEIN 29/7-02: this didn't enlarge the space, when the
     * position was at the first position in the next fragment.  Thus,
     * it was fixed using >= instead of >.  See also related comment in
     * the enlargeSpace() method below.
     */
    if (position >= byteCount) {
      byteCount = position;
      if (position > maxsize) {
        throw new IndexOutOfBoundsException("New position (" + position
          + ") exceeds maximum length:" + maxsize);
      }
      enlargeSpace(position);
    }
    buffer = fragments[fragment];
  }


  /**
   *  Write a short integer value <code>v</code> into the stream at the
   *  current position.  <p>
   *
   *  @exception ArrayIndexOutOfBoundsException
   *    Raised if the total number of bytes written in the message
   *    exceeds the maximum length.
   */
  public void write2(int v)
  {
    position += 2;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    if (offset + 2 > endpayload) {
      temp[0] = (byte) (v >>> 8);
      temp[1] = (byte) (v >>> 0);
      writeTemp(temp, 2);
    } else {
      buffer[offset++] = (byte) (v >>>  8);
      buffer[offset++] = (byte) (v >>>  0);
    }
    written += 2;
  }


  /**
   *  Write an integer value <code>v</code> into the stream at the
   *  current position.  <p>
   *
   *  @exception ArrayIndexOutOfBoundsException
   *    Raised if the total number of bytes written in the message
   *    exceeds the maximum length.
   */
  public void write4(int v)
  {
    position += 4;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    if (offset + 4 > endpayload) {
      temp[0] = (byte) (v >>> 24);
      temp[1] = (byte) (v >>> 16);
      temp[2] = (byte) (v >>>  8);
      temp[3] = (byte) (v >>>  0);
      writeTemp(temp, 4);
    } else {
      buffer[offset++] = (byte) (v >>>  24);
      buffer[offset++] = (byte) (v >>>  16);
      buffer[offset++] = (byte) (v >>>   8);
      buffer[offset++] = (byte) (v >>>   0);
    }
    written += 4;
  }


  /**
   *  Write a long value <code>v</code> into the stream at the current
   *  position.  <p>
   *
   *  @exception ArrayIndexOutOfBoundsException
   *    Raised if the total number of bytes written in the message
   *    exceeds the maximum length.
   */
  public void write8(long v)
  {
    position += 8;
    if (position > byteCount) {
      byteCount = position;
      if (position > maxsize)
        throw new ArrayIndexOutOfBoundsException();
    }
    int v1 = (int) (v >>> 32);
    int v2 = (int) (v & 0xFFFFFFFFL);
    if (offset + 8 > endpayload) {
      temp[0] = (byte) (v1 >>> 24);
      temp[1] = (byte) (v1 >>> 16);
      temp[2] = (byte) (v1 >>>  8);
      temp[3] = (byte) (v1 >>>  0);
      temp[4] = (byte) (v2 >>> 24);
      temp[5] = (byte) (v2 >>> 16);
      temp[6] = (byte) (v2 >>>  8);
      temp[7] = (byte) (v2 >>>  0);
      writeTemp(temp, 8);
    } else {
      buffer[offset++] = (byte) (v1 >>> 24);
      buffer[offset++] = (byte) (v1 >>> 16);
      buffer[offset++] = (byte) (v1 >>> 8);
      buffer[offset++] = (byte) (v1 >>> 0);
      buffer[offset++] = (byte) (v2 >>> 24);
      buffer[offset++] = (byte) (v2 >>> 16);
      buffer[offset++] = (byte) (v2 >>> 8);
      buffer[offset++] = (byte) (v2 >>> 0);
    }
    written += 8;
  }

  /**
   *  Write the content of another <code>OutMessage</code> into the
   *  current message. <p>
   *
   *  @exception ArrayIndexOutOfBoundsException
   *    Raised if the total number of bytes written in the message
   *    exceeds the maximum length.
   */
  public void write(OutMessage msg) 
  {
    byte[][] fragments = msg.getFragments();
    int length = msg.getFragmentCount();
    for (int i=0; i < length-1; i++)
      write(fragments[i], msg.getHeader(), msg.getPayload());
    write(fragments[length-1], msg.getHeader(), msg.getLastPayload());
  }
  
  /**
   *  Reset the byte array
   */
  public void reset()
  {
    // Reinitialize fragments
    buffer = fragments[0];
    
    // Initialize variables regarding message size
    byteCount = 0;
    fragmentCount = 1;
    maxsize = Integer.MAX_VALUE;

    // Initialize variables regarding message position in the stream
    position = 0;
    fragment = 0;
    offset = header;
  }

  /**
   * Compute the checksum of the bytes stored in the message
   *
   * @return the computed checksum.
   */
  public int computeChecksum()
  {
    byte check = 0;

    for (int j = 0; j < fragmentCount-1; j++)
      for (int i = header; i < endpayload; i++)
        check ^= fragments[j][i];
    byte[] buffer = fragments[fragmentCount-1];
    int end = getLastPayload() + header;
    for (int i = header; i < end; i++)
      check ^= buffer[i];

    return check;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the current position in this message.
   */
  public int getPosition()
  {
    return position;
  }

  /**
   *  Return the length of the payload for this message.
   */
  public int getPayload()
  {
    return payload;
  }

  /**
   *  Return the length of the header for this message.
   */
  public int getHeader()
  {
    return header;
  }

  /**
   *  Return the length of the trailer for this message.
   */
  public int getTrailer()
  {
    return trailer;
  }

  /**
   *  Returns the total number of bytes stored in this message.
   */
  public int getByteCount()
  {
    return byteCount;
  }

  /**
   *  Returns the number of fragments composing the message.
   */
  public int getFragmentCount()
  {
    return fragmentCount;
  }

  /**
   *  Returns the length of the payload of the last fragment composing
   *  the message.
   */
  public int getLastPayload()
  {
    if (byteCount % payload == 0)
      return payload;
    else
      return (byteCount % payload);
  }
  
  /**
   *  Returns an array of byte arrays which contains the fragments
   *  composing the message.  The length of fragment array may be
   *  greater than the number of fragments actually allocated; the
   *  actual number of fragments is obtained by invoking method
   *  <code>getFragmentCount</code>.
   */
  public byte[][] getFragments()
  {
    return fragments;
  }

  /**
   *  Converts the whole <code>OutMessage</code> into a byte array.
   */
  public byte[] toByteArray()
  {
    byte[] ret = new byte[byteCount];
    if (byteCount > 0) {
      for (int i=0; i < fragmentCount-1; i++) {
        System.arraycopy(fragments[i], 0, ret, i*payload, payload);
      }
      System.arraycopy(fragments[fragmentCount-1], 0, ret, 
        (fragmentCount-1)*payload, getLastPayload());
    }
    return ret;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Copy <code>length</code> bytes from array <code>temp</code>; this
   *  method is invoked when methods <code>write2</code>,
   *  <code>write4</code>, or <code>write8</code> should write part of
   *  their data in the current fragment and the rest in the following
   *  fragment.
   */
  private void writeTemp(byte[] temp, int length)
  {
    enlargeSpace(byteCount);
    for (int i = 0; i < length; i++) {
      if (offset == endpayload) {
        fragment++;
        buffer = fragments[fragment];
        offset = header;
      }
      buffer[offset++] = temp[i];
    }
  }


  /**
   *  Allocate new byte fragments to guarantee that there is enough
   *  space to store <code>newsize</code> bytes.
   */
  private void enlargeSpace(int newsize)
  {
    /*
     * FIXED HEIN 29/7-02: computing the number of fragments was
     * incorrect when the newsize was equal to the payload.  This caused
     * a problem when doing seek(payload).  The seek() would compute
     * that it was in the first position of the next fragment, but
     * computing nfragment to one less fragment than seek() expected,
     * and so it didn't enlarge the space after all.  So when doing a
     * write() from the new current position, we would get a
     * NullPointerException since it was trying to access a buffer[]
     * that was not initialized.
     *
     * The old code was this:
     *    int nfragment = ((newsize-1) / payload) + 1;
     *
     * In anycase, this does not lead to many extra memory allocations,
     * since the case when newsize equals to payload is very rare.
     */

    /* Total number of fragments needed to store <code>newsize</code> bytes. */
    int nfragment = (newsize / payload) + 1;
    if (nfragment > fragments.length) {
      /* Double the fragments array */
      byte[][] tempbuffer = new byte[nfragment*2][];
      System.arraycopy(fragments, 0, tempbuffer, 0, fragments.length);
      fragments = tempbuffer;
    }
    for (int i = fragmentCount; i < nfragment; i++)
      fragments[i] = new byte[totlen];
    if (fragmentCount < nfragment)
      fragmentCount = nfragment;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Object methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of this object (used only for
   *  debugging)
   */
  public String toString(boolean all)
  {
    StringBuilder buf = new StringBuilder();
    buf.append("[OutMessage: byteCnt=");
    buf.append(byteCount);
    buf.append(", payload=");
    buf.append(payload);
    buf.append(", header=");
    buf.append(header);
    buf.append(", trailer=");
    buf.append(trailer);
    buf.append(", totlen=");
    buf.append(totlen);
    buf.append(", endpayload=");
    buf.append(endpayload);
    buf.append(", position=");
    buf.append(position);
    buf.append(", fragment=");
    buf.append(fragment);
    buf.append(", offset=");
    buf.append(offset);
    buf.append(", fragmentCount=");
    buf.append(fragmentCount);
    buf.append(", maxsize=");
    buf.append(maxsize);
    buf.append("]");

    if (ConfigManager.logMsgContent) {
      buf.append(": ");
    } else {
      // Return without logging message content
      return buf.toString();
    }

    if (all) {
      for (int i = 0; i < fragmentCount-1; i++) {
        for (int j = header; j < endpayload; j++) {
          buf.append(" ");
          buf.append(Util.byte2str(fragments[i][j]));
        }
      }
      int stop = byteCount - (fragmentCount-1)*payload + header;
      for (int j = header; j < stop; j++) {
        buf.append(" ");
        buf.append(Util.byte2str(fragments[fragmentCount-1][j]));
      }
    } else {
      for (int i = 0, j = header; j < endpayload; j++) {
        buf.append(" ");
        buf.append(Util.byte2str(fragments[i][j]));
      }
    }
    return buf.toString();
  }


  /**
   *  Returns a string representation of this object
   */
  public String toString()
  {
    return toString(false);
  }

} // END OutMessage
