/*
 * Copyright (c) 1998-2002 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.util;

import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *  The <code>Queue</code> class implements a queue with last-in,
 *  first-out semantics.  The queue is implemented as a linked list of
 *  buckets, each of them contains a fixed number of objects.  Appended
 *  objects are inserted in the last bucket (if full, a new bucket is
 *  allocated), while removed object are extracted from the first
 *  bucket.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @since Jgroup 0.1
 */
public final class Queue
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* The default bucket size */
  private final static int QUEUE_BUCKET_SIZE = 32;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** First bucket in the queue */
  private Bucket head;

  /** Last bucket in the queue */
  private Bucket tail;

  /** Number of object in a single bucket */
  private int bucketSize;

  /** Current object index in the first bucket */
  private int first = 0;

  /** Current object index in the last bucket */
  private int last = 0;

  /** Number of elements in the queue */
  private int ncount = 0;

  /**
   * The number of times this Queue has been structurally modified.
   * Structural modifications are those that change the number of
   * elements in the queue.  This field is used to make the iterator
   * fail-fast, without risk of rendering the datastructure in an
   * inconsistent state.  (See ConcurrentModificationException).
   */
  private transient int modCount = 0;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Allocates a queue with bucket size equal to <code>bucketSize</code>.
   *
   *  @param bucketSize
   *    The size of internal buckets.
   */
  public Queue(int bucketSize)
  {
    if (bucketSize < 1) 
      throw new IllegalArgumentException("Bucket size must be greater than zero");
    this.bucketSize = bucketSize;
    head = new Bucket(bucketSize);
    tail = head;
  }


  /**
   *  Allocates a queue with default bucket size.
   */
  public Queue()
  {
    this(QUEUE_BUCKET_SIZE);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Append <code>object</code> to the end of the queue.
   *
   *  @param object
   *    The object to be inserted.
   */
  public void insert(Object object)
  {
    tail.objects[last] = object;
    last++;
    if (last == bucketSize) {
      if (tail.next == null)
        tail.next = new Bucket(bucketSize);
      tail = tail.next;
      last = 0;
    }
    modCount++;
    ncount++;
  }


  /**
   *  Returns the size of the queue.
   */
  public int size()
  {
    return ncount;
  }


  /**
   *  Returns true if the queue is empty.
   */
  public boolean isEmpty()
  {
    return (ncount == 0);
  }


  /**
   *  Returns the first element of the queue; null if the queue is
   *  empty.
   */
  public Object getFirst()
  {
    if (ncount == 0)
      return null;

    return head.objects[first];
  }


  /**
   *  Removes the first element of the queue.
   *
   *  @return the first element of the queue; null if the queue is empty.
   */
  public Object removeFirst()
  {
    if (ncount == 0)
      return null;
    modCount++;

    Object ret = head.objects[first];
    head.objects[first] = null;
    ncount--;
    first++;
    if (first == bucketSize) {
      Bucket temp = tail.next;
      tail.next = head;                 // This avoid useless allocations
      head = head.next;
      tail.next.next = temp;
      first = 0;
    }
    return ret;
  }


  /**
   *  Add the elements of the addend queue at the end of this queue.
   *
   *  @param addend
   *    The queue to be added to the end of this queue.
   */
  public void add(Queue addend)
  {
    modCount++;
    
    Object obj;
    while ((obj = addend.removeFirst()) != null)
      insert(obj);
  }


  /**
   *  Removes all the elements from the queue.
   */
  public void clear()
  {
    modCount++;
    for (int i = first; i < bucketSize; i++)
      head.objects[i] = null;
    head.next = null;
    tail = head;
    first  = 0;
    last   = 0;
    ncount = 0;
  }


  /**
   *  Copies the element of the queue in an array of objects, in the
   *  queue order.
   */
//   public Object[] toArray()
//   {
//     Object[] ret = new Object[ncount];
//     Bucket scan = head;
//     int pos = first;
//     for (int i = 0; i < ncount; i++) {
//       ret[i] = scan.objects[pos];
//       pos++;
//       if (pos == bucketSize) {
//         pos = 0;
//         scan = scan.next;
//       }
//     }
//     return ret;
//   }


  /**
   *  Returns an iterator over the queue content.
   */
  public Iterator iterator()
  {
    return new QueueIterator();
  }


  public String toString()
  {
    StringBuilder buf = new StringBuilder("[Queue: ");
    for (Iterator qi = iterator(); qi.hasNext();) {
      buf.append(qi.next());
      if (qi.hasNext()) 
        buf.append(", ");
    }
    buf.append("]");
    return buf.toString();
  }


  /**
   *  Provides an iterator over the elements in the queue.
   *
   *  @author Hein Meling
   *  @since Jgroup 1.2
   */
  private class QueueIterator
    implements Iterator
  {
    private Bucket scan = head;
    private int pos = first;
    private int nextIndex = 0;

    /**
     * The modCount value that the iterator believes that the backing
     * Queue should have.  If this expectation is violated, the iterator
     * has detected concurrent modification.
     */
    private int expectedModCount = modCount;


    public boolean hasNext()
    {
      return nextIndex != ncount;
    }


    public Object next()
    {
      if (modCount != expectedModCount)
        throw new ConcurrentModificationException();
      if (nextIndex == ncount)
        throw new NoSuchElementException();

      Object obj = scan.objects[pos++];
      if (pos == bucketSize) {
        pos = 0;
        scan = scan.next;
      }
      nextIndex++;
      return obj;
    }


    public void remove() 
    {
      throw new UnsupportedOperationException();
    }

  } // END QueueIterator


  /**
   *  The <code>Bucket</code> class represents the container class used
   *  in queue to avoid expensive allocations and to save memory.
   *
   *  @author Alberto Montresor
   *  @since Jgroup 0.1
   */
  private class Bucket {

    //////////////////////////////////////////////////////////////////////////////////////////
    // Fields
    //////////////////////////////////////////////////////////////////////////////////////////

    /** Stored objects */
    Object[] objects;

    /** Next element */
    Bucket next;


    //////////////////////////////////////////////////////////////////////////////////////////
    // Constructor
    //////////////////////////////////////////////////////////////////////////////////////////

    /**
     *  Constructs a QueueBucket with the specified size
     *
     *  @param bucketSize
     *    Number of elements in a bucket.
     */
    Bucket(int bucketSize)
    {
      this.objects = new Object[bucketSize];
      this.next = null;
    }

  } // END Bucket

} // END Queue

